'
' EDM32.INC, PB/DLL custom editor control, written by Borje Hagsten.
' Build to be fast and feature-rich, plus easy to use and program, with
' many built-in functions and configuration messages. Primary use - code
' editor, since it only handles fixed width fonts and has no automatic
' word wrap. I've been using it in my own, highly customized code editor
' for a couple of years now and it hasn't failed or crashed on me yet,
' so I think it's quite safe. But as always, use at own responsibility..
'------------------------------------------------------------------------------
' Oct 15, 2004  Changed to use %DEFAULT_GUI_FONT in FUNCTION MakeScrollInfoLabel.
'               Added missing readonly handling for VK_TAB under SUB ctmChar.
'               Added exported SUB InitEDM32EDITClass, which enables registering
'               the control and the use CONTROL ADD "EDM32EDIT" in DDT dialogs.
'                (CreateEdm32edit can still be used directly as before)
'               Added missing SetCapture to ctmMouseLButtonDblClk call.
'               Removed KillTimer before SetTimer in ctmMouseMove call and
'                 instead make sure @ed.hTimer = 0 before creation. Gives
'                 smoother timer scroll when/if mouse is moved above/below
'                 the editor with left button pressed down (in select-state).
'               Removed edmWrapText for TXT format under edmLoadFile.
'               Added missing handling of Shift+Delete to FUNCTION ctmKeyDown
'               after tip from Vladimir Kapustin.
'               Made sure to Push/Pop Eax around pointer asssignment in
'               SUB edmTxtUpDateLineEnds after tip from Roberto Bianchi.
'               Thanks fellows!

' Apr 15, 2003  To fix a bug if first keyword is alone in it's index (A, B, etc)
'                 I have removed "IF @q2 THEN" -block around edmArrayScan in all
'                 ctmPaint.. routines, and under %EDM_ISKEYWORD message.
'               Fixed a small error in edmTxtFindNext under StartHere label
'                 that caused last character to not become included in search
'                 for single character. Fix: "IF pos <= LEN(txt) - sln + 1 THEN"
'               Changed Page Up/Down under ctmKeyDown (%VK_PRIOR/%VK_NEXT) to
'                 page size - 1, so top or bottom line stays in sight.
'               Fixed a minor error in edmUndoDoUndo/Redo which could cause caret to
'                be placed outside a line's max length at Undo/Redo Replace all under
'
' Apr, 2003     Added crude HTML code syntax coloring as ed.FileType = 3. Not
'                 perfect, but enough good to be helpful at editing of simple
'                 HTML. See LoadHTMLdata, ctmPaintHtml and edmLoadFile routines.
'               Added %EDM_SETFILETYPE to enable direct switch between syntax
'                 color parsing schemes.
'
' Feb, 2003     Fixed a parantheses error with edmUCASE in edmTxtReplaceAll after
'               tip from Pierre Bellisle - thanks Pierre, well spotted!
' Jan 14, 2003: Changed to DWORD for handles in some places and
'               now use GetWindowLong(hWnd, %GWL_HINSTANCE) to
'               get proper instance handle at creation. Added
'               lParam check to %WM_GETDLGCODE in VListProc.
'               Added edmSetFontEditor(Fixedsys ..) to CreateEdm32edit,
'               to ensure it always starts with a fixed width font.
' Nov, 2002.   Added keywords for PB/WIN 7.0, replaced BEEP with MessageBeep
'              and added own color for PBForms statement to syntax coloring.
'              Thumb scroll info label now shows top-bottom visible line numbers.
'              Changed selection in left margin so it now selects whole line,
'              including line feed. Fixed scroll wheel redraw problem.
' May 4, 2002. Added enumeration of character set for edmMakeFont. Made some
'              minor corrections and fixes to undo/redo pos for Replace All
'              and other multiple actions. Added %EDM_SETAUTOINDENT message,
'              to enable setting auto indent on/off. Replaced a bunch of
'              CHR$(x) with matching equates (like $TAB, $CRLF, etc.).
'              Added LCASE and MCASE options for syntax highlight.
'
' January 21, 2002. updated and corrected to work with both PB/DLL 6.0 and 6.1
' October 9, changed some in way to handle @ed.hFont, so user-set font is
' properly deleted at exit. Old code could cause small memory leak, but no more.
' September 26 - added edmFileExist to make sure edmLoadFile loads an existing file..
' September 18 - added %EDM_ISUNTITLED message to help keep track of utitled
' files. Very useful in MDI editors. Also fixed a bug in WM_SETFONT that could
' cause a system font to be deleted in NT systems.
'------------------------------------------------------------------------------
' Free of charge, released into Public Domain in July 2001.
' Sample editor, EDM32 demo, included in zip file. Place EDM32.INC in your
' PBWin70(or Pbdll60/61)/Winapi -folder, among the other PB include files and look
' at the sample editor to learn basics of how to use/program the EDM32 editor control.
' Also, since full source code is provided, look through the messages and provided
' comments below to learn more about it. Feel free to add functionality and/or
' modify the code as you whish. :)
'------------------------------------------------------------------------------
' Some of many features:
'   Asm Text Case handling routines, edmUCASE, edmLCASE edmMCASE, edmTCASE
'   Asm Word Count routine, edmWordCount
'   Bookmarks, toggling on/off, moving between, etc.
'   Built-in file loading routine
'   Built-in font handling via edmSetFontEditor and edmMakeFont
'   Built-in keyboard shortcuts for basic edit commands
'   Built-in line handling routines, like sorting, (un)commenting, in/outdenting
'   Built-in Search and Replace all -functions.
'   Built-in Sub/Function finder routine for PB code
'   Dynamic memory manager(s), handles "unlimited" line sizes..
'   Fully configurable via messages, for color settings, etc.
'   Line number label on thumbtrack scroll (can be turned off)
'   Syntax color parser for PB/DLL code files (.BAS/.INC)
'   Syntax color parser for PB resource files (.RC)
'   Tab-key indents/outdents when several lines are selected
'   Unlimited file sizes (memory dependent, of course)
'   Unlimited Undo/Redo  (also memory dependent)
'------------------------------------------------------------------------------
' Known "issues"
'   Many "Replace all" operations on a large file eats memory, since
'   entire text is saved in Undo buffer for this command. Will rewrite
'   this some day, but for now this will have to do.
'------------------------------------------------------------------------------
' How To:  (a few "how to's", not covered in sample program.)
'   Get entire contents into a string:           txt = edmTxtGetSelText(hEdit, 0)
'   Get selected text into a string:             txt = edmTxtGetSelText(hEdit, 1)
'   Get a line into a string (first line = 0):   txt = edmTxtGetLine(hEdit, LineNum)
'   Select a line (line feeds not included):     Sendmessage hEdit, %EDM_SELECTLINE, 5, 1
'   Get length of line (line feed not included): ln = SendMessage(hEdit, %EDM_GETLINELEN, LineNum, 0)
'   reset undo buffer (some like this on save):  SendMessage hEdit %EDM_UNDOCLEAR, 0, 0
'   change text color (here to blue):            SendMessage hEdit %EDM_SETCOLORTXT, RGB(0,0,255), 1
'   find next occurrence of search text:         pos = edmTxtFindNext(hEdit, SearchTxt, sDirection, sMatchCase, sWWordsOnly, 1)
'   replace all occurrences of search text:      cnt = edmTxtReplaceAll(hEdit, SearchTxt, ReplaceTxt, MatchCase, WWordsOnly, 0)
'   count all occurrences of search text:        cnt = edmTxtReplaceAll(hEdit, SearchTxt, "", MatchCase, WWordsOnly, 1)
'

'#RESOURCE "EDM32.PBR"                'resource file, containing right-hand cursor. Best placed in main code.

'------------------------------------------------------------------------------
'FILE RELATED MESSAGES
%EDM_LOADFILE        = %WM_USER + 100   'wParam = pointer to File name string/asciiz, lParam is focus flag
%EDM_GETFILETYPE     = %WM_USER + 107   'returns @ed.FileType
%EDM_ISUNTITLED      = %WM_USER + 108   'sets/returns untitled file status. wParam sets flag if lParam <> 0
%EDM_SETFILETYPE     = %WM_USER + 109   'wParam sets @ed.FileType, 2=BAS, 3=HTML, 4=RC, lParam = redraw

'EDIT RELATED MESSAGES
%EDM_UNDO            = %WM_USER + 120  'undo edit operation
%EDM_REDO            = %WM_USER + 121  'redo undo action
%EDM_CANUNDO         = %WM_USER + 122  'returns number of possible undo actions
%EDM_CANREDO         = %WM_USER + 123  'returns number of possible redo actions
%EDM_UNDOCLEAR       = %WM_USER + 124  'empties undo buffer via UndoInit
%EDM_COPY            = %WM_USER + 131  'copies selected or whole text (if no selection exists)
%EDM_PASTE           = %WM_USER + 132  'adjusts and pastes clipboard contents at carets place in text
%EDM_DELETE          = %WM_USER + 133  'wParam <>0, delete all, else delete selected text
%EDM_GETTEXT         = %WM_USER + 140  'get all text. wParam must be pointer to string, lParam allocated length
%EDM_GETTEXTLEN      = %WM_USER + 141  'returns length of text in editor
%EDM_GETLINELEN      = %WM_USER + 142  'returns length of wParam line in editor (not counting line feeds)

%EDM_SETSEL          = %WM_USER + 150  'wParam < 0, reset sel. lParam < 0, sel all. Else wParam = start and lParam = end of selection.
%EDM_SELECTLINE      = %WM_USER + 151  'wParam < 0, select current line - else wParam is line number, lParam is redraw flag
%EDM_RESETSEL        = %WM_USER + 152  'reset selection, lParam holds redraw flag
%EDM_REPLACESEL      = %WM_USER + 153  'wParam is pointer to replace text, lParam is redraw flag
%EDM_REFRESH         = %WM_USER + 160  'lParam = setfocus flag
%EDM_SETCHANGED      = %WM_USER + 161  'lParam sets changed-flag
%EDM_GETCHANGED      = %WM_USER + 162  'returns changed-flag

'BOOKMARK MESSAGES
%EDM_SETBMARK        = %WM_USER + 170  'if lParam is zero, bookmark current line, else lParam is line number
%EDM_RESETBMARKSALL  = %WM_USER + 171  'resets all bookmarks
%EDM_SETBMCOLOR      = %WM_USER + 172  'wParam holds new color, lParam is redraw flag
%EDM_NEXTBMARK       = %WM_USER + 173  'go to next bookmark, looking from line number lParam - if zero, looking from caret's line
%EDM_PREVBMARK       = %WM_USER + 174  'go to previous bookmark, looking from line number lParam - if zero, looking from caret's line
%EDM_GETBMARKCOUNT   = %WM_USER + 175  'returns number of set bookmarks
%EDM_GETBMARK        = %WM_USER + 176  'returns bookmark status for line number given in lParam

'SETTINGS MESSAGES
%EDM_SETCOLORGRIDLINE     = %WM_USER + 182 'wParam sets gridline color
%EDM_SETCOLORSYNTAXASM    = %WM_USER + 183 'wParam sets syntax color for inline Asm code - default RED
%EDM_SETCOLORSYNTAXCODE   = %WM_USER + 184 'wParam sets syntax color for PB code - default BLUE
%EDM_SETCOLORSYNTAXREM    = %WM_USER + 185 'wParam sets syntax color for uncommented code (rem) - default GREEN
%EDM_SETCOLORSYNTAXSTRING = %WM_USER + 186 'wParam sets syntax color for strings - default RED
%EDM_SETCOLORPAGE         = %WM_USER + 190 'wParam sets window color - default %COLOR_WINDOW, lParam is redraw flag
%EDM_SETCOLORTXT          = %WM_USER + 191 'wParam sets text color - default %COLOR_WINDOWTEXT, lParam is redraw flag
%EDM_SETCOLORTXTBK        = %WM_USER + 192 'wParam sets text background color - default %COLOR_WINDOW, lParam is redraw flag
%EDM_SETGRID              = %WM_USER + 200 'wParam sets grid-lines on/off, 0 = OFF, else ON
%EDM_SETINSERTMODE        = %WM_USER + 202 'toggles Insert/Overwrite mode
%EDM_SETKBSHORTCUTS       = %WM_USER + 204 'wParam sets internal keyboad shortcuts on/off
%EDM_SETKWCASE            = %WM_USER + 206 'wParam sets KeyWord case status, 0 = ignore case, else auto UCASE
%EDM_SETSHOWNONPRINT      = %WM_USER + 210 'wParam sets show-status of non-visual characters, 0 = SHOW, else HIDE
%EDM_SETSYNTAXONOFF       = %WM_USER + 212 'wParam sets syntax coloring on/off, 0 = OFF, else ON
%EDM_SETWORDWRAP          = %WM_USER + 214 'wParam sets WordWrap On/Off
%EDM_SHOWSCROLLINFO       = %WM_USER + 216 'wParam sets thumbtrack scrollinfo label on/off
%EDM_SETAUTOINDENT        = %WM_USER + 218 'wParam sets auto indent on/off

%EDM_GETCOLORGRIDLINE     = %WM_USER + 232 'returns gridline color
%EDM_GETCOLORSYNTAXASM    = %WM_USER + 233 'returns syntax color for inline Asm code - default RED
%EDM_GETCOLORSYNTAXCODE   = %WM_USER + 234 'returns syntax color for PB code - default BLUE
%EDM_GETCOLORSYNTAXREM    = %WM_USER + 235 'returns syntax color for uncommented code (rem) - default GREEN
%EDM_GETCOLORSYNTAXSTRING = %WM_USER + 236 'returns syntax color for strings - default RED
%EDM_GETCOLORPAGE         = %WM_USER + 240 'returns window color - default %COLOR_WINDOW
%EDM_GETCOLORTXT          = %WM_USER + 241 'returns text color - default %COLOR_WINDOWTEXT
%EDM_GETCOLORTXTBK        = %WM_USER + 242 'returns text background color - default %COLOR_WINDOW
%EDM_GETGRID              = %WM_USER + 250 'returns flag for grid-lines on/off status
%EDM_GETINSERTMODE        = %WM_USER + 252 'returns flag for Insert/Overwrite mode
%EDM_GETKWCASE            = %WM_USER + 256 'returns KeyWord case status, auto UCASE or not..
%EDM_GETSHOWNONPRINT      = %WM_USER + 260 'returns flag for show-status of non-visual characters
%EDM_GETSYNTAXONOFF       = %WM_USER + 262 'returns flag for syntax coloring on/off
%EDM_GETWORDWRAP          = %WM_USER + 264 'returns flag for WordWrap On/Off

'CUSTOM FUNCTION MESSAGES
%EDM_UCASE           = %WM_USER + 300  ' make selection UPPER CASE
%EDM_MCASE           = %WM_USER + 301  ' make selection Mixed Case
%EDM_LCASE           = %WM_USER + 302  ' make selection lower case
%EDM_TCASE           = %WM_USER + 303  ' make selection tOGGLED cASE
%EDM_SORTLINES       = %WM_USER + 310  ' wParam = 0-Asc. 1-Asc, ignore case. 2-Desc. 3-Desc. ignore case
%EDM_COMMENT         = %WM_USER + 312  ' lParam = 0-comment out. 1-remove rem out. 2-Indent. 3-outdent.
%EDM_ISKEYWORD       = %WM_USER + 314  ' word as strptr in lParam. Returns %TRUE if word is keyword
%EDM_SETREADONLY     = %WM_USER + 320  ' wParam sets ReadOnly mode, <> 0 = ReadOnly
%EDM_GETREADONLY     = %WM_USER + 321  ' returns ReadOnly status

'INFORMATION MESSAGES
%EDM_GETCARETPOS     = %WM_USER + 400  ' returns linear caret position
%EDM_GETCARETX       = %WM_USER + 401  ' returns caret's X-position on line
%EDM_GETCARETY       = %WM_USER + 402  ' returns caret's Y-position (line)
%EDM_SETCARETPOS     = %WM_USER + 404  ' wParam sets linear position in text
%EDM_SETCARETX       = %WM_USER + 405  ' wParam sets caret's X-position on line
%EDM_SETCARETY       = %WM_USER + 406  ' wParam sets caret's Y-position (line)

%EDM_GETSELPOSSTART  = %WM_USER + 410  ' returns linear selStart position
%EDM_GETSELPOSEND    = %WM_USER + 411  ' returns linear selEnd position
%EDM_GETSELSTARTX    = %WM_USER + 412  ' returns selstart's X-position on line
%EDM_GETSELSTARTY    = %WM_USER + 413  ' returns selstart's Y-position (line)
%EDM_GETSELENDX      = %WM_USER + 414  ' returns selend's X-position on line
%EDM_GETSELENDY      = %WM_USER + 415  ' returns selend's Y-position (line)
%EDM_GETSELLENGTH    = %WM_USER + 416  ' returns selection's length

%EDM_SETSELPOSSTART  = %WM_USER + 420   ' wParam sets linear selstart position in text
%EDM_SETSELPOSEND    = %WM_USER + 421   ' wParam sets linear selend position in text
%EDM_SETSELSTARTX    = %WM_USER + 422   ' wParam sets selstart's X-position on line
%EDM_SETSELSTARTY    = %WM_USER + 423   ' wParam sets selstart's Y-position (line)
%EDM_SETSELENDX      = %WM_USER + 424   ' wParam sets selend's X-position on line
%EDM_SETSELENDY      = %WM_USER + 425   ' wParam sets selend's Y-position (line)

%EDM_GETTOPLINE      = %WM_USER + 450   ' returns top line (First visible line)
%EDM_LINELENWINDOW   = %WM_USER + 451   ' returns max number of characters a window can show on a line
%EDM_LINESPERWINDOW  = %WM_USER + 452   ' returns max number of full lines a window can show
%EDM_GETLINECOUNT    = %WM_USER + 453   ' returns number of lines in text
%EDM_GETWORDCOUNT    = %WM_USER + 454   ' returns number of words in text
%EDM_GETHSCROLL      = %WM_USER + 456   ' returns horizontal scrollbar's current value
%EDM_GETHSCROLLMAX   = %WM_USER + 457   ' returns horizontal scrollbar's max value
%EDM_GETCHARHEIGHT   = %WM_USER + 460   ' returns character height = line height for selected font
%EDM_GETCHARWIDTH    = %WM_USER + 461   ' returns character width for selected font

%EDM_SETTOPLINE      = %WM_USER + 480   ' wParam sets top line (First visible line). lParam is redraw flag
%EDM_SETHSCROLL      = %WM_USER + 482   ' wParam sets horizontal scroll position
%EDM_GOTOLINE        = %WM_USER + 483   ' wParam is line number, lParam is redraw flag

%EDM_SETTABSIZE      = %WM_USER + 484   ' wParam sets Tab width (number of spaces that replaces a tab)
%EDM_GETTABSIZE      = %WM_USER + 485   ' returns Tab width (number of spaces that replaces a tab)
%EDM_GETLEFTMARGIN   = %WM_USER + 487   ' returns width of left side's selection/bookmark field (margin)
%EDM_SETLEFTMARGIN   = %WM_USER + 488   ' sets width: wParam = width, lParam = redraw flag
%EDM_GETAUTOINDENT   = %WM_USER + 489   ' returns auto indent on/off

%EDM_GETCAPSLOCK     = %WM_USER + 500   ' returns status for CapsLock key (0-off, 1-active)
%EDM_GETNUMLOCK      = %WM_USER + 501   ' returns status for NumLock key (0-off, 1-active)
%EDM_GETSCROLLOCK    = %WM_USER + 502   ' returns status for ScrollLock key (0-off, 1-active)

'INTERNAL CONTROL MESSAGES
%EDM_CHECKBOUNDS     = %WM_USER + 520   ' internal use, checks bounds for TopLine (first visible line)
%EDM_CARETHANDLER    = %WM_USER + 530   ' internal use, checks bounds for caret and sets scrollbar, etc.

'NOTIFICATION MESSAGES
%EDM_SETINFO         = %WM_USER + 600   'sent to parent - notification message on all caret movements, etc

'------------------------------------------------------------------------------
'Following two UDT's can be freely customized to meet extended needs
'------------------------------------------------------------------------------
Type UNDODATA             'User-defined type for undo/redo actions
  uAction         As Long ' for type of undo/redo action, 1 = insert text, 2 = delete text
  uActionCount    As Long ' for number of undo/redo actions to be done in one take
  uActionStartX   As Long ' for x startpos of undo/redo action
  uActionStartY   As Long ' for y startpos of undo/redo action
  uActionEndX     As Long ' for x endpos of undo/redo action
  uActionEndY     As Long ' for y endpos of undo/redo action
  uActionLen      As Long ' for endpos of undo/redo action
  uTxtBlockStart  As Long ' for startpos in undo/redo text memory block
  uTxtBlockEnd    As Long ' for length of text in undo/redo text memory block
  uDescription    As Long ' for description of undo/redo action
End Type

Type EDM32DATA                   'User-defined type for control-specific data
  hWnd               As Dword     'control's handle
  id                 As Long      'control's id
  hInst              As Dword     'control's instance handle
  hParent            As Dword     'control's parent's handle
  isUntitled         As Long      'flag for untitled (new) status
  memDC              As Dword     'memDC used in WM_PAINT
  hBit               As Dword     'memDC's bitmap
  CursArrow          As Dword     'handle of normal arrow cursor
  CursArrowRigth     As Dword     'handle of custom right-arrow cursor
  CursIBeam          As Dword     'handle of normal beam cursor
  CursDragDrop       As Dword     'handle of drag&drop cursor - not implemented yet..
  autoIndent         As Long      'auto indent mode flag
  cInsOvr            As Long      'Insert or Overwrite..
  dClicked           As Long      'for different ways to handle double-click
  charHeight         As Long      'character height is same as Line height
  charWidth          As Long      'character width
  confineCaret       As Long      'if to confine caret to line's legth or not - not implemented yet
  wnWidth            As Long      'editor's width (window width)
  wnHeight           As Long      'editor's height (window height)
  TopLine            As Long      'current first visible line
  OldTopLine         As Long      'previous first visible line
  lMargin            As Long      'Left margin, equals width of left selection/bookmark field.
  xPos               As Long      'xPos = @ed.lMargin + @ed.charWidth * (- @ed.iHscrollPos)
  OldHscrollPos      As Long      'previous horizontal scroll pos, used in WM_PAINT
  iHscrollPos        As Long      'current horizontal scroll pos
  iHscrollMax        As Long      'max horizontal scroll pos
  hTimer             As Dword     'timer for scroll when mouse has capture above/below control
  hScrollLabel       As Dword     'handle for scroll info label
  showScrollInfo     As Long      'flag for scroll info label
  infoLeadText       As Asciiz * 10 'lead text for scroll info label
  xCaretOld          As Long      'caret's previous x-pos
  xCaret             As Long      'caret's current x-pos
  yCaret             As Long      'caret's current y-pos (line pos)
  SelStartX          As Long      'selection's start x-pos
  SelStartY          As Long      'selection's start y-pos (line)
  SelEndX            As Long      'selection's end x-pos
  SelEndY            As Long      'selection's end y-pos (line)
  SelClickedY        As Long      'selection's absolute start y-pos (line)
  dSelStartX         As Long      'selection's start x-pos on double-clicked word
  dSelStartY         As Long      'selection's start y-pos on double-clicked word
  dSelEndX           As Long      'selection's end x-pos on double-clicked word
  SelLen             As Long      'selection length
  ColorPage          As Long      'editor's page color
  ColorTxtBk         As Long      'text background color
  ColorTxt           As Long      'text color
  ColorSyntaxAsm     As Long      'syntax color for inline Asm code - default RED
  ColorSyntaxCode    As Long      'syntax color for PB code - default BLUE
  ColorSyntaxPBForms As Long      'syntax color for PBForms statement line - default BROWN
  ColorSyntaxRem     As Long      'syntax color for uncommented code (rem) - default GREEN
  ColorSyntaxString  As Long      'syntax color for strings - default RED
  ColorGridLine      As Long      'gridline color
  hFont              As Dword     'editor font handle
  hBrushBMark        As Dword
  hBrushScrlBar      As Dword
  Changed            As Long      'flag for changed or not (for save needs)
  FileType           As Long      'file type value
  kbShortCuts        As Long      'flag for built-in keyboard shortcuts (in WM_CHAR)
  ReadOnly           As Long      'flag for readonly mode
  LineUpdate         As Long      'internal flag for WM_PAINT.
  ScrollMode         As Long      'flag for ScrollLock key
  ShowGridLines      As Long      'flag for show/hide gridlines
  ShowNonPrintChars  As Long      'flag for show/hide non-printable characters
  ShowSyntaxColor    As Long      'flag for syntax coloring on/off
  WordWrap           As Long      'flag for static wordWrap when loading a file
  heapbMarks         As Dword     'handle to heap, bMarks memory block
  heapTxt            As Dword     'handle to heap, txt memory block
  heapLineEnd        As Dword     'handle to heap, LineEnd memory block
  heapUndo           As Dword     'handle to heap, Undo structure memory block
  heapUndotxt        As Dword     'handle to heap, Undo text memory block
  bMarks             As Dword     'pointer to bookmarks memory block, Longs, 4 byte boundary
  LineEnd            As Dword     'pointer to line end memory block, Longs, 4 byte boundary
  txt                As Dword     'pointer to main text memory block (dynamic block length)
  txtLen             As Long      'holds text length, main text
  UBtxBuf            As Long      'holds number of lines in main text (Ubound txt)
  kwCase             As Long      'if all keywords shall be converted to UCASE or not on syntax parsing
  pData              As Dword     'pointer to PB keywords array, dynamic strings, fixed block length
  pDataEndPos        As Dword     'pointer to PB keywords end points in memory block
  pStartCount        As Dword     'pointer to PB keywords index, Startcount (A, B, C, etc,)
  pEndCount          As Dword     'pointer to PB keywords index, Endcount
  hData              As Dword     'pointer to HTML keywords array, dynamic strings, fixed block length
  hDataEndPos        As Dword     'pointer to HTML keywords end points in memory block
  hStartCount        As Dword     'pointer to HTML keywords index, Startcount (A, B, C, etc,)
  hEndCount          As Dword     'pointer to HTML keywords index, Endcount
  rData              As Dword     'pointer to RC keywords array, dynamic strings, fixed block length
  rDataEndPos        As Dword     'pointer to RC keywords end points in memory block
  rStartCount        As Dword     'pointer to RC keywords index, Startcount (A, B, C, etc,)
  rEndCount          As Dword     'pointer to RC keywords index, Endcount
  tabSize            As Long      'default tab size is set to 4 space characters
  undoCount          As Long      'counter for undo/redo actions
  undoPos            As Long      'position for undo/redo action
  undotxt            As Dword     'pointer to undo/redo text
  undo               As UNDODATA Ptr ' pointer to undo/redo data
  fName              As Asciiz * %MAX_PATH  'this enables control to store loaded file's path and name
End Type

'------------------------------------------------------------------------------
Declare Function CreateEdm32edit(ByVal hWnd As Dword, ByVal CtrlId As Long, _
                                 ByVal vLeft As Long, ByVal vTop As Long, _
                                 ByVal vWidth As Long, ByVal vHeight As Long, _
                                 ByVal wStyle As Long, ByVal wStyleEx As Long, _
                                 ByVal DDTstyle As Long) As Dword

Declare Function ctlWndProc(ByVal hWnd   As Dword, ByVal wMsg   As Dword,_
                            ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmCaretHandler(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmChar(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmCommand(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmCreate(ByVal hWnd As Dword, ByVal lParam As Long) As Long
Declare Function ctmCtlColorStatic(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmDestroy(ByVal hWnd As Dword) As Long
Declare Function ctmHscroll(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmKeyDown(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmKeyUp(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmKillFocus(ByVal hWnd As Dword, ByVal wParam As Long) As Long
Declare Function ctmMouseLButtonDblClk(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmMouseLButtonDown(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmMouseLButtonUp(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmMouseMove(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmMouseRButtonDown(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmMouseRButtonDblClk(ByVal hWnd   As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmMouseWheel(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmPaint(ByVal hWnd As Dword, ByVal wParam As Long) As Long
Declare Function ctmSetFocus(ByVal hWnd As Dword, ByVal wParam As Long) As Long
Declare Function ctmSetFont(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmSize(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmTimer(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function ctmVscroll(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long

Declare Function edmArrayScan(ByVal txt As String, ByVal mb As Long, ByVal Idx As Long, _
                              ByVal dStart As Long, ByVal dEnd As Long) As Long
Declare Function edmBookMarked(ByVal hWnd As Dword, ByVal y As Long) As Long
Declare Function edmBookMarkGetCount(ByVal hWnd As Dword) As Long
Declare Function edmBookMarkGetNext(ByVal hWnd As Dword, ByVal y As Long) As Long
Declare Function edmBookMarkGetPrev(ByVal hWnd As Dword, ByVal y As Long) As Long
Declare Function edmBookMarksToList(ByVal hWnd As Dword, ByVal lsthWnd As Long) As Long
Declare Function edmClipboardGetText () As String
Declare Function edmClipboardPutText (ByVal txt As String) As Long
Declare Function edmFullLineProc(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function edmGetFileName(ByVal hWnd As Dword) As String
Declare Function edmGetWordLimits(ByVal hWnd As Dword, ByVal StartOrEnd As Long, ed As EDM32DATA) As Long
Declare Function edmLoadFile(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function edmMakeFont(ByVal FontName As String, ByVal PointSize As Long, ByVal fBold As Long, _
                             ByVal fItalic As Long, ByVal fUnderline As Long) As Dword
Declare Function edmSetFontEditor(ByVal hWnd As Dword, ByVal FontName As String, ByVal PointSize As Long, _
                                  ByVal fBold As Long, ByVal fItalic As Long, ByVal fUnderline As Long) As Dword
Declare Function edmTxtFindCRLF(ByVal s As String) As String
Declare Function edmTxtFindNext(ByVal hWnd As Dword, ByVal s As String, ByVal dr As Long, _
                                ByVal mc As Long, ByVal wo As Long, ByVal eRedraw As Long) As Long
Declare Function edmTxtReplaceAll(ByVal hWnd As Dword, ByVal s As String, ByVal r As String, _
                                  ByVal mc As Long, ByVal wo As Long, CountOnly As Long) As Long
Declare Function edmTxtGetLine(ByVal hWnd As Dword, ByVal y As Long) As String
Declare Function edmTxtGetLineLen(ed As EDM32DATA, ByVal y As Long) As Long
Declare Function edmTxtGetLineStart(ed As EDM32DATA, ByVal y As Long) As Long
Declare Function edmTxtGetMaxLineLen(ByVal hWnd As Dword) As Long
Declare Function edmTxtGetPos(ByVal hWnd As Dword, ByVal x As Long, ByVal y As Long) As Long
Declare Function edmTxtGetSelLength(ed As EDM32DATA) As Long
Declare Function edmTxtGetSelText(ByVal hWnd As Dword, ByVal selFlag As Long) As String
Declare Function edmTxtGetFileSummary(ByVal hWnd As Dword) As String
Declare Function edmTxtInit(ed As EDM32DATA) As Dword
Declare Function edmUndoGetDescription(ByVal hWnd As Dword, ByVal UndoOrRedo As Long) As String
Declare Function edmUndoGetString(ByVal iType As Long) As String
Declare Function edmUndoInit(ed As EDM32DATA) As Dword
Declare Function edmUCASER(ByVal st As String) As String
Declare Function edmWordCount(a$) As Long
Declare Function edmWrapText(ByVal txt As String, ByVal nw As Long, ByVal strip As Long) As String
Declare Function LoadPBdata(ed As EDM32DATA) As Long
Declare Function LoadHTMLdata(ed As EDM32DATA) As Long
Declare Function LoadRCdata(ed As EDM32DATA) As Long
Declare Function MakeScrlBitMap(ByVal hDlg As Dword) As Dword
Declare Function MakeScrollInfoLabel(ByVal hParent As Dword) As Dword

Declare Sub edmBookMarkDelete(ed As EDM32DATA, ByVal Pos As Long, ByVal Count As Long)
Declare Sub edmBookMarkInitialize(ByVal hWnd As Dword)
Declare Sub edmBookMarkInsert(ed As EDM32DATA, ByVal Pos As Long, ByVal Count As Long)
Declare Sub edmBookMarkResetAll(ByVal hWnd As Dword)
Declare Sub edmBookMarkSet(ByVal hWnd As Dword, ByVal y As Long, ByVal bmVal As Long)
Declare Sub edmCheckXYbounds(ed As EDM32DATA)
Declare Sub edmLCASE(a$)
Declare Sub edmMCASE(a$)
Declare Sub edmTCASE(a$)
Declare Sub edmUCASE(a$)
Declare Sub edmGetSubs(ByVal hWnd As Dword, ByVal hCombo As Dword, ByVal hList As Dword)
Declare Sub ctmPaintBasic(ByVal hWnd As Dword, ByVal Buf As String, ByVal yLine As Long)
Declare Sub ctmPaintHtml(ByVal hWnd As Dword, ByVal Buf As String, ByVal yLine As Long)
Declare Sub ctmPaintRc(ByVal hWnd As Dword, ByVal Buf As String, ByVal yLine As Long)
Declare Sub ctmPaintSelection(ByVal hWnd As Dword)
Declare Sub edmSelectSub(ByVal hWnd As Dword, KeepSelect As Long, ed As EDM32DATA)
Declare Sub edmSelectWord(ByVal hWnd As Dword, KeepSelect As Long, ed As EDM32DATA)
Declare Sub edmSetFileName(ByVal hWnd As Dword, ByVal fName As String)
Declare Sub edmSetScrollData(ByVal hWnd As Dword, ByVal vhPage As Long, ByVal vhMax As Long, _
                             ByVal vhPos As Long, ByVal vhBar As Long)
Declare Sub edmSetText(ByVal hWnd As Dword, ByVal txt As String)
Declare Sub edmTxtAddToLineEnds(ed As EDM32DATA, ByVal addThis As Long)
Declare Sub edmTxtDeleteSel(ed As EDM32DATA, ByVal selFlag As Long, ByVal uFlag As Long, ByVal iType As Long)
Declare Sub edmTxtGetXYpos(ed As EDM32DATA, ByVal Pos As Long, x As Long, y As Long)
Declare Sub edmTxtLineEndResize(ed As EDM32DATA, ByVal ii As Long)
Declare Sub edmTxtPutString(ed As EDM32DATA, ByVal s As String, ByVal x As Long, ByVal y As Long, _
                            ByVal mFlag As Long, ByVal uFlag As Long, ByVal iType As Long)
Declare Sub edmTxtResizeBlocks(ed As EDM32DATA, ByVal nSize As Long)
Declare Sub edmTxtUpDateLineEnds(ed As EDM32DATA, ByVal xp As Long, ByVal lp As Long)
Declare Sub edmUndoAdd(ed As EDM32DATA, ByVal act As Long, ByVal iType As Long, ByVal txt As String, _
                       ByVal x1 As Long, ByVal y1 As Long, ByVal x2 As Long, ByVal y2 As Long)
Declare Sub edmUndoDoRedo(ed As EDM32DATA)
Declare Sub edmUndoDoUndo(ed As EDM32DATA)
Declare Sub edmUndoRedoFillList(ByVal lsthWnd As Dword, ByVal edhWnd As Dword, ByVal UndoOrRedo As Long)

'
' Register the EDM32EDIT class
'--------------------------------------------------------------------
Sub InitEDM32EDITClass Alias "InitEDM32EDITClass" Export
  Local wc As WNDCLASSEX, szClassName As Asciiz * 20

  szClassName      = "EDM32EDIT"
  wc.cbSize        = SizeOf(WNDCLASSEX)
  wc.lpszClassName = VarPtr(szClassName)
  wc.lpfnWndProc   = CodePtr(ctlWndProc)
  wc.hInstance     = GetModulehandle("")
  wc.STYLE         = %CS_DBLCLKS Or %CS_GLOBALCLASS
  wc.cbWndExtra    = 4  ' for pointer to private TYPE data

  RegisterClassEx wc 'register new class

End Sub

'
' Create control
'--------------------------------------------------------------------
Function CreateEdm32edit Alias "CreateEdm32edit" _
                        (ByVal hParent  As Dword, ByVal CtrlId   As Long, _
                         ByVal cLeft    As Long,  ByVal cTop     As Long, _
                         ByVal cWidth   As Long,  ByVal cHeight  As Long, _
                         ByVal wStyle   As Long,  ByVal wStyleEx As Long, _
                         ByVal DDTstyle As Long)  Export As Dword
  Local hEdm As Dword

  InitEDM32EDITClass 'initiate the EDM32EDIT window class

  If DDTstyle Then
     Dialog Units hParent, cLeft, cTop To Pixels cLeft, cTop
     Dialog Units hParent, cWidth, cHeight To Pixels cWidth, cHeight
  End If

  hEdm = CreateWindowEx(wStyleEx, "EDM32EDIT", "", wStyle, _
                        cLeft, cTop, cWidth, cHeight, _
                        hParent, CtrlId, GetWindowLong(hParent, %GWL_HINSTANCE), ByVal 0)

  If hEdm Then
     edmSetFontEditor hEdm, "Fixedsys", 9, 0, 0, 0 'set initial fixed width font
  End If

  Function = hEdm 'return the handle
End Function

'
' Editor's main "window procedure", center for all message handling
'
Function ctlWndProc(ByVal hWnd   As Dword, ByVal wMsg  As Dword,_
                    ByVal wParam As Long, ByVal lParam As Long) As Long

  Local edmI As Long, edmJ As Long, ed As EDM32DATA Ptr, edmTxt As String
  If hWnd And wMsg <> %WM_CREATE Then ed = GetWindowLong(hWnd, 0)

  Select Case wMsg
  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ' CONTROL MESSAGES
  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     Case %WM_CREATE         : Function = ctmCreate(hWnd, lParam)      'received while in CreateWindow
     Case %WM_CTLCOLORSTATIC : Function = ctmCtlColorStatic(hWnd, wParam, lParam)
     Case %WM_COMMAND        : Function = ctmCommand(hWnd, wParam, lParam)
     Case %WM_DESTROY        : Function = ctmDestroy(hWnd)
     Case %WM_GETFONT        : Function = @ed.hFont
     Case %WM_KILLFOCUS      : Function = ctmKillFocus(hWnd, wParam)   'hide and destroy the caret
     Case %WM_SETFOCUS       : Function = ctmSetFocus(hWnd, wParam)    'create and show the caret
     Case %WM_SETFONT        : Function = ctmSetFont(hWnd, wParam, lParam)
     Case %WM_SIZE           : Function = ctmSize(hWnd, wParam, lParam)
     Case %WM_TIMER          : Function = ctmTimer(hWnd, wParam, lParam)
     Case %WM_SYSCOLORCHANGE
        If @ed.hBrushScrlBar Then DeleteObject @ed.hBrushScrlBar
        @ed.hBrushScrlBar = MakeScrlBitMap(hWnd)

  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ' KEYBOARD MESSAGES
  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     Case %WM_CHAR       : Function = ctmChar(hWnd, wParam, lParam)
     Case %WM_GETDLGCODE : If lParam Then Function = %DLGC_WANTALLKEYS 'control processes all keys by itself
     Case %WM_KEYDOWN    : Function = ctmKeyDown(hWnd, wParam, lParam)
     Case %WM_KEYUP      : Function = ctmKeyUp(hWnd, wParam, lParam)
     Case %WM_HSCROLL    : Function = ctmHscroll(hWnd, wParam, lParam)
     Case %WM_VSCROLL    : Function = ctmVscroll(hWnd, wParam, lParam)

  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ' MOUSE MESSAGES
  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     Case %WM_LBUTTONDBLCLK : Function = ctmMouseLButtonDblClk(hWnd, wParam, lParam)
     Case %WM_LBUTTONDOWN   : Function = ctmMouseLButtonDown(hWnd, wParam, lParam)
     Case %WM_LBUTTONUP     : Function = ctmMouseLButtonUp(hWnd, wParam, lParam)
     Case %WM_MOUSEMOVE     : Function = ctmMouseMove(hWnd, wParam, lParam)
     Case %WM_MOUSEWHEEL    : Function = ctmMouseWheel(hWnd, wParam, lParam)
     Case %WM_RBUTTONDOWN   : Function = ctmMouseRButtonDown(hWnd, wParam, lParam)
     Case %WM_RBUTTONDBLCLK : Function = ctmMouseRButtonDblClk(hWnd, wParam, lParam)

  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ' DRAWING MESSAGES
  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     Case %WM_ERASEBKGND    : Function = 1 'Don't erase background
     Case %WM_PAINT         : Function = ctmPaint(hWnd, wParam)

  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ' CONTROL SPECIFIC MESSAGES
  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  'FILE RELATED
     Case %EDM_LOADFILE     : Function = edmLoadFile(hWnd, wParam, lParam)
     Case %EDM_GETFILETYPE  : Function = @ed.FileType
     Case %EDM_ISUNTITLED
        If lParam Then @ed.isUntitled = wParam         'lParam decides get/set
        Function = @ed.isUntitled                      'return new file (untitled) status
     Case %EDM_SETFILETYPE
        @ed.FileType = wParam
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, -1

  'EDIT RELATED
     Case %WM_CUT
        If @ed.ReadOnly = 1 Then Exit Function         'no action if ReadOnly
        edmClipboardPutText edmTxtGetSelText(hWnd, 1)  'copy selection directly to clipboard
        SendMessage hWnd, %EDM_DELETE, 0, -1           'and delete it
        If  @ed.undoCount Then @ed.@undo[@ed.undoCount - 1].uDescription = 10 'set undo/redo description text

     Case %WM_COPY, %EDM_COPY                          '%WM_COPY for compatibility reasons
        If @ed.SelLen Then
           edmClipboardPutText edmTxtGetSelText(hWnd, 1)
        ElseIf wMsg = %EDM_COPY Then                    '%EDM_COPY, for copying all text
           edmClipboardPutText edmTxtGetSelText(hWnd, 0)
        End If

     Case %WM_PASTE, %EDM_PASTE                         '%WM_PASTE for compatibility reasons
        If @ed.ReadOnly = 1 Then Exit Function          'no action if ReadOnly

        edmTxt = edmClipboardGetText                        'place text in string
        If InStr(edmTxt, Chr$(0)) Then Replace Chr$(0) With Chr$(32) In edmTxt  ' replace ev. zeros with Spaces
        If InStr(edmTxt, $Tab) Then Replace $Tab With Space$(@ed.tabSize) In edmTxt ' replace eventual tabs..
        If InStr(edmTxt, Chr$(13, 13, 10)) Then                  'weird MS FrontPage format..?
           Replace Chr$(13, 13, 10) With $CrLf In edmTxt  'convert
        ElseIf InStr(edmTxt, $CrLf) = 0 Then              'check if unusual line feeds has been used
           If InStr(edmTxt, $Lf) Then                       'UNIX text-format
              Replace $Lf With $CrLf In edmTxt       'convert
           ElseIf InStr(edmTxt, $Cr) Then                   'MAC text-format
              Replace $Cr With $CrLf In edmTxt       'convert
           End If
        End If

        If @ed.SelLen Then                                                 'if something is selected
           SendMessage hWnd, %EDM_DELETE, 0, 0                             'delete it
           edmTxtPutString @ed, edmTxt, @ed.xCaret, @ed.yCaret, -1, -1, 13 'and put string
           If @ed.undoPos Then @ed.@undo[@ed.undoPos - 1].uActionCount = 1 'tell undo of double action
        Else                                                               'else
           edmTxtPutString @ed, edmTxt, @ed.xCaret, @ed.yCaret, -1, -1, 5  'just put string
        End If
        If Len(edmTxt) = 1 Then Incr @ed.xCaret                               'then simulate result of keypress
        @ed.LineUpdate = 0
        SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0
        SendMessage hWnd, %EDM_REFRESH, 0, -1

     Case %WM_CLEAR, %EDM_DELETE
        If wParam And wMsg = %EDM_DELETE Then               'Delete all, only for %EDM_DELETE message
           edmBookMarkInitialize hWnd                       'remove all bookmarks
           @ed.SelStartX = 0 : @ed.SelStartY = 0            'and select all text
           @ed.SelEndX = edmTxtGetLineLen(@ed, @ed.UBtxBuf)
           @ed.SelEndY = @ed.UBtxBuf
        Else                                                'Delete selection
           If edmTxtGetSelLength(@ed) = @ed.txtLen Then     'all selected
              edmBookMarkInitialize hWnd                    'remove all bookmarks
           ElseIf @ed.SelStartY <> @ed.SelEndY Then         'several lines selected, remove eventual bookmarks
              edmBookMarkDelete @ed, Min&(@ed.SelStartY, @ed.SelEndY), Abs(@ed.SelEndY - @ed.SelStartY)
           End If
        End If
        edmTxtDeleteSel @ed, 0, -1, 12                 'delete
        SendMessage hWnd, %EDM_RESETSEL, 0, 0          'reset selection status
        SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0       'check bounds for @ed.TopLine
        If lParam Or wMsg = %WM_CLEAR Then
           '@ed.LineUpdate = 0
           SendMessage hWnd, %EDM_REFRESH, 0, -1       'redraw
        End If

     Case %EM_SETSEL, %EDM_SETSEL  '%EDM_SETSEL is custom, if to select without repainting
        If wParam < 0 Then                               'if wParam is negative, reset selection
           SendMessage hWnd, %EDM_RESETSEL, 0, 0
           Exit Function
        End If
        If lParam < 0 Then                               'if lParam is negative, select all
           @ed.SelStartX = 0 : @ed.SelStartY = 0
           @ed.SelEndX = edmTxtGetLineLen(@ed, @ed.UBtxBuf) : @ed.SelEndY = @ed.UBtxBuf
        Else                                             'else wParam is start and lParam is end of selection
           edmTxtGetXYpos @ed, wParam, @ed.SelStartX, @ed.SelStartY
           edmTxtGetXYpos @ed, lParam, @ed.SelEndX, @ed.SelEndY
           @ed.xCaret = @ed.SelEndX : @ed.yCaret = @ed.SelEndY
        End If
        @ed.SelLen = edmTxtGetSelLength(@ed)
        If wMsg = %EM_SETSEL Then SendMessage hWnd, %EDM_REFRESH, 0, -1

     Case %EDM_SELECTLINE
        If wParam < 0 Then                                'if wParam is negative, select current line
           @ed.SelStartY = @ed.yCaret
        Else                                              'else lParam holds line number
           @ed.SelStartY = Min&(@ed.UBtxBuf, Max&(0, wParam))
        End If
        @ed.SelEndY = @ed.SelStartY                       'one line only, so SelEndY = SelStartY
        @ed.SelStartX = 0                                 'whole line - starts from zero
        @ed.SelEndX   = edmTxtGetLineLen(@ed, @ed.SelStartY) 'and ends at line length
        @ed.SelLen    = edmTxtGetSelLength(@ed)              'calculate selLen
        @ed.xCaret    = @ed.SelEndX : @ed.yCaret = @ed.SelEndY 'set Caret to end of selection (line)
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, -1   'if lParam, repaint control

     Case %EDM_RESETSEL  'reset selection, lParam holds redraw flag
        @ed.SelStartX = @ed.xCaret : @ed.SelEndX = @ed.xCaret
        @ed.SelStartY = @ed.yCaret : @ed.SelEndY = @ed.yCaret : @ed.SelLen = 0
        If wParam Then
           InvalidateRect hwnd, ByVal %NULL, 0 : UpdateWindow hwnd
        End If

     Case %EDM_REPLACESEL                                      'replace selection
        If wParam Then                                         'wParam is pointer to replace text
           If @ed.ReadOnly = 1 Then Exit Select                'no action if ReadOnly

           Local zTxtPtr As Asciiz Ptr
           zTxtPtr = wParam                                    'pick up replace text via its string pointer
           If @ed.SelLen Then                                  'if anything selected
              SendMessage hWnd, %EDM_DELETE, 0, 0              'delete selection
              If Len(@zTxtPtr) Then                            'if replace text has length
                 edmTxtPutString @ed, @zTxtPtr, @ed.xCaret, @ed.yCaret, -1, -1, 22  'put replace text
                 If @ed.undoPos Then @ed.@undo[@ed.undoPos - 1].uActionCount = 1    'tell undo of multiple action
              End If
           End If
           SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0             'check bounds, for TopLine
           If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, -1 'lParam holds redraw flag
        End If

     Case %EDM_GETTEXT                                 'get all text from editor
        edmTxt = edmTxtGetSelText(hWnd, 0)             'place all text in string
        If wParam And lParam = Len(edmTxt) Then        'wParam must be pointer to receiving string, lParam allocated length
           Poke$ wParam, edmTxt                        'poke txt into receiving string
        End If

     Case %EDM_GETINSERTMODE  : Function = @ed.cInsOvr
     Case %EDM_SETINSERTMODE  : SendMessage hWnd, %WM_KEYDOWN, %VK_INSERT, 0
     Case %EDM_GETCHANGED     : Function = @ed.Changed
     Case %EDM_SETCHANGED     : @ed.Changed = wParam

  'INTERNAL
     Case %EDM_CARETHANDLER : Function = ctmCaretHandler(hWnd, wParam, lParam)

     Case %EDM_CHECKBOUNDS  'internal message, check bounds for first visible line (TopLine)
        If @ed.yCaret > @ed.TopLine + @ed.wnHeight - 1 Then
           @ed.TopLine = Min&(@ed.yCaret - @ed.wnHeight + 1, @ed.UBtxBuf - @ed.wnHeight + 1)
           @ed.LineUpdate = 0
        ElseIf @ed.yCaret < @ed.TopLine Then
           @ed.TopLine = Max&(0, @ed.yCaret)
           @ed.LineUpdate = 0
        End If

        If @ed.TopLine > @ed.UBtxBuf - @ed.wnHeight + 1 Then
           @ed.TopLine = Max&(0, @ed.UBtxBuf - @ed.wnHeight + 1)
        End If

     Case %EDM_REFRESH      'redraw editor
        If @ed.xCaret < @ed.iHscrollPos - 1 Or _
              @ed.SelStartX < @ed.iHscrollPos - 1 Or _
              @ed.SelEndX < @ed.iHscrollPos - 1 Then
           @ed.iHscrollPos = Max&(0, @ed.xCaret - 10)
           @ed.LineUpdate = 0                          'redraw whole window - all visible lines
        End If

        If @ed.xCaret > @ed.iHscrollPos + @ed.wnWidth Then
           @ed.iHscrollPos = @ed.xCaret - @ed.wnWidth + 10
           @ed.LineUpdate = 0
        End If

        If @ed.xCaret > @ed.iHscrollPos + @ed.wnWidth Then @ed.iHscrollPos = @ed.xCaret - @ed.wnWidth + 10
        @ed.xPos = @ed.lMargin + @ed.charWidth * (- @ed.iHscrollPos)
        InvalidateRect hwnd, ByVal %NULL, 0 : UpdateWindow hwnd
        SetCaretPos @ed.xPos + @ed.xCaret * @ed.charWidth, (@ed.yCaret - @ed.TopLine) * @ed.charHeight
        edmSetScrollData hwnd, @ed.wnWidth / 3, @ed.iHscrollMax, @ed.iHscrollPos, %SB_HORZ
        edmSetScrollData hwnd, @ed.wnHeight, Max&(0, @ed.UBtxBuf), @ed.TopLine, %SB_VERT
        SendMessage GetParent(hWnd), %EDM_SETINFO, 0, 1
        If lParam Then SetFocus hWnd

  'CUSTOM
     Case %EDM_ISKEYWORD
        Local zp As Asciiz Ptr, q As Dword Ptr, q2 As Dword Ptr
        zp = lParam

        edmI = Max&(0, Asc(@zp) - 64)
        If edmI < 30 Then
           q2 = @ed.pEndCount + edmI * 4
           q = @ed.pStartCount + edmI * 4
           Function = edmArrayScan(@zp, @ed.pData, @ed.pDataEndPos, Max&(0, @q), Max&(0, @q2))
        End If

     Case %EDM_SETREADONLY
        @ed.ReadOnly = wParam
        If @ed.ReadOnly Then
           HideCaret hWnd
           DestroyCaret
           @ed.CursIBeam = LoadCursor(%NULL, ByVal %IDC_ARROW)
        Else
           @ed.CursIBeam = LoadCursor(%NULL, ByVal %IDC_IBEAM)
        End If

     Case %EDM_GETREADONLY    : Function = @ed.ReadOnly
     Case %EDM_GETTEXTLEN     : Function = @ed.txtLen
     Case %EDM_GETLINELEN     : Function = edmTxtGetLineLen(@ed, Min&(Max&(0, wParam - 1), @ed.UBtxBuf))
     Case %EDM_GETWORDWRAP    : Function = @ed.WordWrap
     Case %EDM_SETWORDWRAP    : @ed.WordWrap = wParam
     Case %EDM_SHOWSCROLLINFO : @ed.showScrollInfo = wParam
     Case %EDM_GETCARETX      : Function = @ed.xCaret
     Case %EDM_GETCARETY      : Function = @ed.yCaret
     Case %EDM_GETCARETPOS    : Function = edmTxtGetPos(hWnd, @ed.xCaret, @ed.yCaret)
     Case %EDM_SETCARETPOS
        edmTxtGetXYpos @ed, wParam, @ed.xCaret, @ed.yCaret
        SendMessage hWnd, %EDM_RESETSEL, 0, 0
        SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0

     Case %EDM_SETCARETX
        If wParam >= 0 Then
           @ed.xCaret = wParam
        Else
           @ed.xCaret = edmTxtGetLineLen(@ed, @ed.UBtxBuf)
        End If

     Case %EDM_SETCARETY
        If wParam >= 0 Then
           @ed.yCaret = wParam
        Else
           @ed.yCaret = @ed.UBtxBuf
        End If

     Case %EDM_GETSELPOSSTART : Function = edmTxtGetPos(hWnd, @ed.SelStartX, @ed.SelStartY)
     Case %EDM_GETSELPOSEND   : Function = edmTxtGetPos(hWnd, @ed.SelEndX, @ed.SelEndY)
     Case %EDM_GETSELSTARTX   : Function = @ed.SelStartX
     Case %EDM_GETSELSTARTY   : Function = @ed.SelStartY
     Case %EDM_GETSELENDX     : Function = @ed.SelEndX
     Case %EDM_GETSELENDY     : Function = @ed.SelEndY
     Case %EDM_GETSELLENGTH   : Function = edmTxtGetSelLength(@ed)

     Case %EDM_SETSELPOSSTART
        edmTxtGetXYpos @ed, wParam, @ed.SelStartX, @ed.SelStartY
        SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0

     Case %EDM_SETSELPOSEND
        edmTxtGetXYpos @ed, wParam, @ed.SelEndX, @ed.SelEndY
        SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0

     Case %EDM_SETSELSTARTX   : @ed.SelStartX = wParam
     Case %EDM_SETSELSTARTY   : @ed.SelStartY = wParam
     Case %EDM_SETSELENDX     : @ed.SelEndX = wParam
     Case %EDM_SETSELENDY     : @ed.SelEndY = wParam
     Case %EDM_GETCHARHEIGHT  : Function = @ed.charHeight
     Case %EDM_GETCHARWIDTH   : Function = @ed.charWidth
     Case %EDM_LINELENWINDOW  : Function = @ed.wnWidth
     Case %EDM_LINESPERWINDOW : Function = @ed.wnHeight
     Case %EDM_GETTOPLINE     : Function = @ed.TopLine
     Case %EDM_SETTOPLINE
        @ed.TopLine = Max&(0, wParam)
        SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, -1

     Case %EDM_SETHSCROLL
        @ed.iHscrollPos = wParam
        SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, -1

     Case %EDM_GOTOLINE
        If wParam > 0 Then     'wParam = line number
           @ed.yCaret = Min&(wParam - 1, @ed.UBtxBuf)
        ElseIf wParam = 0 Then 'next line
           @ed.yCaret = Min&(@ed.yCaret + 1, @ed.UBtxBuf)
        ElseIf wParam < 0 Then 'previous line
           @ed.yCaret = Max&(@ed.yCaret - 1, 0)
        Else
           SetFocus hWnd
           Exit Function
        End If
        @ed.xCaret = 0
        @ed.TopLine = Max(0, @ed.yCaret - 3)
        @ed.SelStartX = @ed.xCaret : @ed.SelEndX = @ed.xCaret
        @ed.SelStartY = @ed.yCaret : @ed.SelEndY = @ed.yCaret : @ed.SelLen = 0
        SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, -1
        Function = @ed.yCaret

     Case %EDM_GETLINECOUNT   : Function = @ed.UBtxBuf + 1
     Case %EDM_GETHSCROLL     : Function = @ed.iHscrollPos
     Case %EDM_GETHSCROLLMAX  : Function = @ed.iHscrollMax
     Case %EDM_SETHSCROLL     : @ed.iHscrollPos = wParam
     Case %EDM_SETKBSHORTCUTS : @ed.kbShortCuts = wParam

     Case %EDM_GETCAPSLOCK  : Function = (GetKeyState(%VK_CAPITAL) And &H1) 'returns 1 for on, 0 for off
     Case %EDM_GETNUMLOCK   : Function = (GetKeyState(%VK_NUMLOCK) And &H1)
     Case %EDM_GETSCROLLOCK : Function = (GetKeyState(%VK_SCROLL)  And &H1)

     Case %EDM_UCASE, %EDM_MCASE, %EDM_LCASE, %EDM_TCASE    'Case functions
        If @ed.SelLen = 0 Or @ed.ReadOnly = 1 Then Exit Function
        edmTxt = edmTxtGetSelText(hWnd, lParam)               'place all text in string
        Select Case wMsg
           Case %EDM_UCASE : edmUCASE edmTxt : edmI = 6 'Make selection UPPER CASE
           Case %EDM_MCASE : edmMCASE edmTxt : edmI = 7 'Make selection Mixed Case
           Case %EDM_LCASE : edmLCASE edmTxt : edmI = 8 'Make selection lower case
           Case %EDM_TCASE : edmTCASE edmTxt : edmI = 9 'Make selection Toggle Case
        End Select
        SendMessage hWnd, %EDM_DELETE, 0, 0
        edmTxtPutString @ed, edmTxt, @ed.xCaret, @ed.yCaret, -1, -1, edmI
        If @ed.undoPos Then @ed.@undo[@ed.undoPos - 1].uActionCount = 1
        SendMessage hWnd, %EDM_REFRESH, 0, -1

     Case %EDM_COMMENT   : edmFullLineProc hWnd, wParam, lParam
     Case %EDM_SORTLINES : edmFullLineProc hWnd, wParam, 4

     'BOOKMARK MESSAGES
     Case %EDM_SETBMARK                          'set bookmark
        lParam = Max&(0, lParam)                 'make sure lParam isn't negative
        If lParam = 0 Then lParam = @ed.yCaret   'if lParam is zero, bookmark current line
        edmBookMarkSet hWnd, lParam, Not edmBookMarked(hWnd, @ed.yCaret)  'set bookmark
        SendMessage hWnd, %EDM_REFRESH, 0, -1                             'and redraw control

     Case %EDM_NEXTBMARK                         'go to next bookmark
        If lParam = 0 Then lParam = @ed.yCaret   'start looking from line number lParam - if zero, caret's line
        edmI = edmBookMarkGetNext(hWnd, lParam)
        If edmI = 0 And edmBookMarkGetCount(hWnd) Then
           edmI = edmBookMarkGetNext(hWnd, -1)
           If edmI Then MessageBeep &HFFFFFFFF
        End If
        If edmI Then
           @ed.xCaret = 0
           @ed.yCaret = edmI - 1
           @ed.TopLine = Min&(Max&(0, @ed.yCaret - 0), Max&(0, @ed.UBtxBuf - @ed.wnHeight + 1))
           SendMessage hWnd, %EDM_REFRESH, 0, -1
        End If

     Case %EDM_PREVBMARK                         'go to previous bookmark
        If lParam = 0 Then lParam = @ed.yCaret
        edmI = edmBookMarkGetPrev(hWnd, lParam)
        If edmI = 0 And edmBookMarkGetCount(hWnd) Then
           edmI = edmBookMarkGetPrev(hWnd, @ed.UBtxBuf + 1)
           If edmI Then MessageBeep &HFFFFFFFF
        End If
        If edmI Then
           @ed.xCaret = 0
           @ed.yCaret = edmI - 1
           @ed.TopLine = Min&(Max&(0, @ed.yCaret - 0), Max&(0, @ed.UBtxBuf - @ed.wnHeight + 1))
           SendMessage hWnd, %EDM_REFRESH, 0, -1
        End If

     Case %EDM_RESETBMARKSALL
        edmBookMarkResetAll hWnd                'reset all bookmarks
        SendMessage hWnd, %EDM_REFRESH, 0, -1   'and redraw control

     Case %EDM_GETBMARKCOUNT : Function = edmBookMarkGetCount(hWnd)
     Case %EDM_GETBMARK      : Function = edmBookMarked(hWnd, lParam) 'returns bookmark status for line number lParam
     Case %EDM_SETBMCOLOR
        If @ed.hBrushBMark Then DeleteObject @ed.hBrushBMark
        @ed.hBrushBMark = CreateSolidBrush(wParam)
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, -1

     'UNDO/REDO MESSAGES
     Case %EM_UNDO, %WM_UNDO : edmUndoDoUndo @ed
     Case %EDM_UNDO
        edmUndoDoUndo @ed
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, -1

     Case %EDM_REDO
        edmUndoDoRedo @ed
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, -1

     Case %EM_CANUNDO  : Function = @ed.undoPos > 0
     Case %EDM_CANUNDO
        edmJ = 0
        If @ed.undoPos > 0 Then
           For edmI = @ed.undoPos - 1 To 0 Step - 1
              Incr edmJ
              If @ed.@undo[edmI].uActionCount > 0 Then edmI = edmI - @ed.@undo[edmI].uActionCount
           Next
        End If
        Function = edmJ

     Case %EDM_CANREDO
        edmJ = 0
        If @ed.undoPos < @ed.undoCount Then
           For edmI = @ed.undoPos To @ed.undoCount - 1
              Incr edmJ
              If @ed.@undo[edmI].uActionCount > 0 Then edmI = edmI + @ed.@undo[edmI].uActionCount
           Next
        End If
        Function = edmJ

     Case %EDM_UNDOCLEAR : edmUndoInit @ed 'clear undo buffer

     Case %EDM_GETWORDCOUNT
        edmTxt = edmTxtGetSelText(hWnd, lParam)
        Function = edmWordCount(edmTxt)

     'VISUAL MESSAGES
     Case %EDM_GETGRID              ' Get Gridlines on/off - default OFF
        Function = @ed.ShowGridLines

     Case %EDM_SETGRID              ' Set Gridlines on/off - default OFF
        @ed.ShowGridLines = wParam  'wParam = status, 0 = OFF, else ON
        SendMessage hWnd, %EDM_REFRESH, 0, -1

     Case %EDM_GETLEFTMARGIN : Function = @ed.lMargin

     Case %EDM_SETLEFTMARGIN
        @ed.lMargin = wParam
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, 0

     Case %EDM_GETTABSIZE           ' Get tab size - default 4 spaces
        Function = @ed.tabSize

     Case %EDM_SETTABSIZE           ' Set tab size - default 4 spaces (also controls auto indent)
        @ed.tabSize = wParam
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, 0

     Case %EDM_GETAUTOINDENT
        Function = @ed.autoIndent

     Case %EDM_SETAUTOINDENT 'wParam sets auto indent on /off (tab=indent / shift+tab=outdent
        @ed.autoIndent = wParam

     Case %EDM_GETSYNTAXONOFF       ' Get syntax coloring on/off - default ON
        Function = @ed.ShowSyntaxColor

     Case %EDM_SETSYNTAXONOFF       ' Set syntax coloring on/off - default ON
        @ed.ShowSyntaxColor = wParam
        If lParam = 0 Then SendMessage hWnd, %EDM_REFRESH, 0, -1


     Case %EDM_GETSHOWNONPRINT      ' Get settings for showing non-visual characters - default OFF
        Function = @ed.ShowNonPrintChars

     Case %EDM_SETSHOWNONPRINT      ' Set settings for showing non-visual characters - default OFF
        @ed.ShowNonPrintChars = wParam
        If lParam = 0 Then SendMessage hWnd, %EDM_REFRESH, 0, -1

     'SET SYNTAX COLOR SETTINGS
     Case %EDM_SETKWCASE            ' Set settings for KeyWord case - default UCASE
        @ed.kwCase = wParam
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, 0

     Case %EDM_SETCOLORSYNTAXASM    ' Set syntax color for inline Asm code - default RED
        @ed.ColorSyntaxAsm = wParam
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, 0

     Case %EDM_SETCOLORSYNTAXCODE   ' Set syntax color for PB code - default BLUE
        @ed.ColorSyntaxCode = wParam
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, 0

     Case %EDM_SETCOLORSYNTAXREM    ' Set syntax color for uncommented code (rem) - default GREEN
        @ed.ColorSyntaxRem = wParam
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, 0

     Case %EDM_SETCOLORSYNTAXSTRING ' Set syntax color for strings - default RED
        @ed.ColorSyntaxString = wParam
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, 0

     Case %EDM_SETCOLORPAGE         ' Set window color - default %COLOR_WINDOW
        @ed.ColorPage = wParam
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, 0

     Case %EDM_SETCOLORTXT          ' Set text color - default %COLOR_WINDOWTEXT
        @ed.ColorTxt = wParam
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, 0

     Case %EDM_SETCOLORTXTBK        ' Set text background color - default %COLOR_WINDOW
        @ed.ColorTxtBk = wParam
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, 0

     Case %EDM_SETCOLORGRIDLINE ' Set gridline color
        @ed.ColorGridLine = wParam
        If lParam Then SendMessage hWnd, %EDM_REFRESH, 0, 0

     'GET SYNTAX COLOR SETTINGS
     Case %EDM_GETKWCASE            ' Get settings for KeyWord case - default UCASE
        Function = @ed.kwCase

     Case %EDM_GETCOLORSYNTAXASM    ' Get syntax color for inline Asm code - default RED
        Function = @ed.ColorSyntaxAsm

     Case %EDM_GETCOLORSYNTAXCODE   ' Get syntax color for PB code - default BLUE
        Function = @ed.ColorSyntaxCode

     Case %EDM_GETCOLORSYNTAXREM    ' Get syntax color for uncommented code (rem) - default GREEN
        Function = @ed.ColorSyntaxRem

     Case %EDM_GETCOLORSYNTAXSTRING ' Get syntax color for strings - default RED
        Function = @ed.ColorSyntaxString

     Case %EDM_GETCOLORPAGE         ' Get window color - default %COLOR_WINDOW
        Function = @ed.ColorPage

     Case %EDM_GETCOLORTXT          ' Get text color - default %COLOR_WINDOWTEXT
        Function = @ed.ColorTxt

     Case %EDM_GETCOLORTXTBK        ' Get text background color - default %COLOR_WINDOW
        Function = @ed.ColorTxtBk

     Case %EDM_GETCOLORGRIDLINE     ' Get gridline color
        Function = @ed.ColorSyntaxString

     Case Else : Function = DefWindowProc(hWnd, wMsg, wParam, lParam) 'default message processing
  End Select

End Function

'
' WM_CREATE message handler
'
Function ctmCreate(ByVal hWnd As Dword, ByVal lParam As Long) As Long
  Local ed As EDM32DATA Ptr

  ed = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, SizeOf(@ed))
  If ed Then
     SetWindowLong hWnd, 0, ed  'Store the pointer for later use
  Else
     'MessageBox hWnd, "Could not allocate memory for Editor control" & CHR$(0), _
     '                 "Error message" & CHR$(0), %MB_OK
     Function = -1 : Exit Function  'Return -1 to break the action
  End If

  @ed.heapTxt     = HeapCreate(0, 0&, 0&)
  @ed.heapLineEnd = HeapCreate(0, 0&, 0&)
  @ed.heapbMarks  = HeapCreate(0, 0&, 0&)
  @ed.heapUndo    = HeapCreate(0, 0&, 0&)
  @ed.heapUndotxt = HeapCreate(0, 0&, 0&)
  edmUndoInit @ed

  @ed.hParent     = GetWindowLong(hWnd, %GWL_HWNDPARENT)
  @ed.hWnd        = hWnd
  @ed.id          = GetWindowLong(hWnd, %GWL_ID)
  @ed.hInst       = GetWindowLong(hWnd, %GWL_HINSTANCE)

  If IsFalse(edmTxtInit(@ed)) Then  'edmTxtInit() initializes memory for the edit control
     SendMessage @ed.hParent, %WM_COMMAND, MakLng(@ed.id, %EN_ERRSPACE), hwnd
     Function = -1 : Exit Function  'Return -1 to break the action on failure
  End If

  @ed.ColorPage          = GetSysColor(%COLOR_WINDOW)      'page (window) color
  @ed.ColorTxtBk         = @ed.ColorPage                   'text background color
  @ed.ColorTxt           = GetSysColor(%COLOR_WINDOWTEXT)  'textcolor
  @ed.ColorSyntaxAsm     = &HFF         'syntax color for inline Asm code - default RED
  @ed.ColorSyntaxCode    = &HFF0000     'syntax color for PB code - default BLUE
  @ed.ColorSyntaxPBForms = &H8000C0     'syntax color for PBForms statement
  @ed.ColorSyntaxRem     = &H008000     'syntax color for uncommented code (rem) - default GREEN
  @ed.ColorSyntaxString  = &HFF         'syntax color for strings - default RED
  @ed.ColorGridLine      = GetSysColor(%COLOR_3DFACE) 'RGB(239, 239, 239)  'gridline color, default very light gray
  @ed.confineCaret       = 1            'keep caret to line's length
  @ed.CursArrow          = LoadCursor(@ed.hInst, "CursorRight")
  @ed.CursIBeam          = LoadCursor(%NULL, ByVal %IDC_IBEAM)
  If @ed.CursArrow = 0 Then
     @ed.CursArrow = LoadCursor(%NULL, ByVal %IDC_ARROW) 'in case resource cursor failed..
  End If
  @ed.FileType          = 2
  @ed.hBrushBMark       = CreateSolidBrush(RGB(255, 0, 0))
  @ed.hBrushScrlBar     = MakeScrlBitMap(hWnd)
  @ed.infoLeadText      = "Line"
  @ed.autoIndent        = 1
  @ed.kwCase            = 1 'default = upper case
  @ed.lMargin           = 15
  @ed.ScrollMode        = (GetKeyState(%VK_SCROLL) And &H1)
  @ed.showScrollInfo    = 1
  @ed.ShowSyntaxColor   = 1
  @ed.tabSize           = 4
  @ed.UBtxBuf           = 0
  @ed.xCaret            = 0
  @ed.yCaret            = 0

  LoadPBdata   ByVal ed
  LoadHTMLdata ByVal ed
  LoadRCdata   ByVal ed

End Function

'
' %EDM_CARETHANDLER message handler
'
Function ctmCaretHandler(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)

  @ed.xCaret = Max&(0, wParam)
  @ed.yCaret = Min&(Max&(0, lParam), @ed.UBtxBuf)
  If @ed.confineCaret Then
     @ed.xCaret = Min&(@ed.xCaret, edmTxtGetLineLen(@ed, @ed.yCaret))
     @ed.xCaretOld = Max&(@ed.xCaretOld, @ed.xCaret)
  End If

  If @ed.xCaret < @ed.iHscrollPos Then
     @ed.iHscrollPos = Max&(0, @ed.xCaret - 10)
     @ed.xPos = @ed.lMargin + @ed.charWidth * (- @ed.iHscrollPos)
     InvalidateRect hwnd, ByVal %NULL, 0 : UpdateWindow hwnd
  ElseIf @ed.xCaret > @ed.iHscrollPos + @ed.wnWidth Then
     @ed.iHscrollPos = @ed.xCaret - @ed.wnWidth + 10
     @ed.xPos = @ed.lMargin + @ed.charWidth * (- @ed.iHscrollPos)
     InvalidateRect hwnd, ByVal %NULL, 0 : UpdateWindow hwnd
  End If

  If @ed.hTimer Then
     @ed.SelEndX = @ed.xCaret : @ed.SelEndY = @ed.yCaret
     @ed.SelLen = edmTxtGetSelLength(@ed)
     If @ed.yCaret > @ed.TopLine + @ed.wnHeight - 1 Then @ed.TopLine = @ed.yCaret - @ed.wnHeight + 1
     If @ed.yCaret < @ed.TopLine Then @ed.TopLine = @ed.yCaret
     InvalidateRect hwnd, ByVal %NULL, 0 : UpdateWindow hwnd
  ElseIf (GetKeyState(%VK_SHIFT) And &H8000) <> 0 Or @ed.dClicked = 3 Then 'Selection mode
     @ed.SelEndX = @ed.xCaret : @ed.SelEndY = @ed.yCaret
     @ed.SelLen = edmTxtGetSelLength(@ed)
     If @ed.yCaret > @ed.TopLine + @ed.wnHeight - 1 Then @ed.TopLine = @ed.yCaret - @ed.wnHeight + 1
     If @ed.yCaret < @ed.TopLine Then @ed.TopLine = @ed.yCaret
     If @ed.confineCaret = 0 Then
         If @ed.SelStartX > edmTxtGetLineLen(@ed, @ed.SelStartY) Then
            @ed.SelStartX = edmTxtGetLineLen(@ed, @ed.SelStartY)
         End If
         If @ed.SelEndX > edmTxtGetLineLen(@ed, @ed.SelEndY) Then
            @ed.SelEndX = edmTxtGetLineLen(@ed, @ed.SelEndY)
         End If
     End If
     InvalidateRect hwnd, ByVal %NULL, 0 : UpdateWindow hwnd

  ElseIf @ed.SelLen Then
     If @ed.yCaret > @ed.TopLine + @ed.wnHeight - 1 Then @ed.TopLine = @ed.yCaret - @ed.wnHeight + 1
     If @ed.yCaret < @ed.TopLine Then @ed.TopLine = @ed.yCaret
     SendMessage hWnd, %EDM_RESETSEL, -1, 0

  Else
     SendMessage hWnd, %EDM_RESETSEL, 0, 0
     If @ed.yCaret < @ed.TopLine Or @ed.yCaret > @ed.TopLine + @ed.wnHeight - 1 Then
        If @ed.yCaret > @ed.TopLine + @ed.wnHeight - 1 Then @ed.TopLine = @ed.yCaret - @ed.wnHeight + 1
        If @ed.yCaret < @ed.TopLine Then @ed.TopLine = @ed.yCaret
        InvalidateRect hwnd, ByVal %NULL, 0 : UpdateWindow hwnd
     End If
  End If

  edmSetScrollData hwnd, @ed.wnWidth / 3, @ed.iHscrollMax, @ed.iHscrollPos, %SB_HORZ
  edmSetScrollData hwnd, @ed.wnHeight, Max&(0, @ed.UBtxBuf), @ed.TopLine, %SB_VERT

  SetCaretPos @ed.xPos + @ed.xCaret * @ed.charWidth, (@ed.yCaret - @ed.TopLine) * @ed.charHeight
  SendMessage GetParent(hWnd), %EDM_SETINFO, 0, 0
  @ed.LineUpdate = 0

End Function

'
' %WM_CHAR message handler
'
Function ctmChar(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
  Local ed As EDM32DATA Ptr, ch As String, Ac As Long, Letter As Byte Ptr
  ed = GetWindowLong(hWnd, 0)
  @ed.LineUpdate = 0

  If @ed.kbShortCuts And (GetKeyState(%VK_CONTROL) And &H8000) <> 0 Then 'use built-in keyboard shortcuts
     Select Case wparam
        Case 1  'Ctrl + A  - Select all
           SendMessage hWnd, %EM_SETSEL, 0, -1

        Case 3  'Ctrl + C - copy
           SendMessage hWnd, %EDM_COPY, 0, 0
           Exit Function

        Case 7  'Ctrl + G - gridlines on/off
           @ed.ShowGridLines = (@ed.ShowGridLines = 0)
           SendMessage hWnd, %EDM_REFRESH, 0, -1

        Case 8  'Ctrl + H - syntax color on/off
           @ed.ShowSyntaxColor = (@ed.ShowSyntaxColor = 0)
           SendMessage hWnd, %EDM_REFRESH, 0, -1

        Case 12  'Ctrl + L - LCASE
           SendMessage hWnd, %EDM_LCASE, 0, -1

        Case 13  'Ctrl + M - MCASE
           SendMessage hWnd, %EDM_MCASE, 0, -1

        Case 20  'Ctrl + T - TCASE
           SendMessage hWnd, %EDM_TCASE, 0, -1

        Case 21  'Ctrl + U - UCASE
           SendMessage hWnd, %EDM_UCASE, 0, -1

        Case 22 'Ctrl + V - paste
           SendMessage hWnd, %WM_PASTE, 0, 0
           Exit Function

        Case 24 'Ctrl + X - cut
           SendMessage hWnd, %WM_CUT, 0, 0
           SendMessage hWnd, %EDM_REFRESH, 0, -1
           Function = 0 : Exit Function

        Case 25 'Ctrl + Y  - Redo
           SendMessage hWnd, %EDM_REDO, 0, 0
           SendMessage hWnd, %EDM_REFRESH, 0, -1
           Exit Function

        Case 26 'Ctrl + Z  - Undo
           SendMessage hWnd, %WM_UNDO, 0, 0
           SendMessage hWnd, %EDM_REFRESH, 0, -1
           Exit Function

        Case 127 'Ctrl + Backspace, ignore here, processed in WM_KEYDOWN
           Exit Function

     End Select

  ElseIf (GetKeyState(%VK_CONTROL) And &H8000) <> 0 Then
     If wParam = 127 Then Exit Function 'Ctrl + Backspace, ignore here, processed in WM_KEYDOWN
  End If

  '********************************************************************
  ' Input text to proper ARRAY element and show it
  '********************************************************************
  Select Case wParam
     Case %VK_TAB
        If @ed.ReadOnly = 1 Then Exit Function
        If @ed.SelLen Then
           ch = Space$(@ed.tabSize) & edmTxtGetSelText(hWnd, 1)

           If InStr(ch, $CrLf) Then                       'multiple lines
              If (GetKeyState(%VK_SHIFT) And &H8000) <> 0  Then 'Move selected lines to the left
                 edmFullLineProc hWnd, 0, 3
              Else                                        'Move selected lines to the right
                 edmFullLineProc hWnd, 0, 2
              End If
           Else                                           'single line with something selected
              Ac = @ed.tabSize - (@ed.xCaret Mod @ed.tabSize)
              ch = Space$(Ac)
              SendMessage hWnd, %EDM_DELETE, 0, -1
              edmTxtPutString @ed, ch, @ed.xCaret, @ed.yCaret, -1, -1, 3
              If Ac = 1 Then @ed.xCaret = @ed.xCaret + Ac 'then must mimic ordinary character input
              If @ed.undoPos Then @ed.@undo[@ed.undoPos - 1].uActionCount = 1
           End If

        Else
           Ac = @ed.tabSize - (@ed.xCaret Mod @ed.tabSize)
           If (GetKeyState(%VK_SHIFT) And &H8000) = 0  Then 'Normal Tab, expand with spaces
              ch = Space$(Ac)
              edmTxtPutString @ed, ch, @ed.xCaret, @ed.yCaret, -1, -1, 3
              If Ac = 1 Then @ed.xCaret = @ed.xCaret + Ac    'then must mimic ordinary character input

           Else 'Shift Tab, remove tab number of spaces if possible
              If @ed.xCaret >= @ed.tabSize Then
                 If Ac = @ed.tabSize Then Ac = 0
                 ch = edmTxtGetLine(hWnd, @ed.yCaret)
                 ch = Mid$(ch, @ed.xCaret - (@ed.tabSize - Ac) + 1, (@ed.tabSize - Ac))
                 If Len(Trim$(ch)) = 0 Then
                    @ed.SelStartX = @ed.xCaret - (@ed.tabSize - Ac)
                    @ed.SelEndX = @ed.xCaret
                    SendMessage hWnd, %EDM_DELETE, 0, -1
                 End If

              End If
           End If
        End If

     Case %VK_RETURN       'line feed
        If @ed.ReadOnly = 1 Then Exit Function

        Local ch2 As String
        ch = edmTxtGetLine(hWnd, @ed.yCaret)

        Letter = StrPtr(ch) : Ac = 0
        Do While @Letter = 32
           Incr Ac : Incr Letter
        Loop

        If Ac = 0 Or @ed.xCaret = 0 Or @ed.autoIndent = 0 Then
           ch2 = $CrLf
        Else
           If @ed.xCaret < Ac Then Ac = @ed.xCaret
           ch = Trim$(edmUCASER(ch)) : ch2 = Space$(Ac)
           If (Left$(ch, 3) = "IF " And Right$(ch, 5) = " THEN") Or _
                 Left$(ch, 4) = "ELSE" Or _
                 Left$(ch, 7) = "SELECT " Or _
                 Left$(ch, 5) = "CASE " Or _
                 Left$(ch, 4) = "FOR " Or _
                 Left$(ch, 3) = "DO " Or _
                 Left$(ch, 3) = "DO" + $Cr Or _
                 Left$(ch, 6) = "WHILE " Or _
                 Left$(ch, 6) = "WHILE" + $Cr Then
              ch2 = $CrLf + ch2 + Space$(@ed.tabSize)
           Else
              ch2 = $CrLf + ch2
           End If
        End If
        edmTxtPutString @ed, ch2, @ed.xCaret, @ed.yCaret, -1, -1, 2
        SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0

     Case %VK_ESCAPE
        SendMessage @ed.hParent, %WM_COMMAND, MakLng(%IDCANCEL, %BN_CLICKED), hwnd
        Exit Function

     Case 32 To 255 'all ANSI characters
        If @ed.ReadOnly = 1 Then Exit Function
        If @ed.SelLen Then
           SendMessage hWnd, %EDM_DELETE, 0, -1                                 'delete selection
           edmTxtPutString @ed, Chr$(wParam), @ed.xCaret, @ed.yCaret, -1, -1, 4 'insert character
           If @ed.undoPos Then @ed.@undo[@ed.undoPos - 1].uActionCount = 1      'tell undo to do two operations
        Else
           If @ed.cInsOvr Then
              If @ed.xCaret < edmTxtGetLineLen(@ed, @ed.yCaret) Then edmTxtDeleteSel @ed, 1, -1, 12
              edmTxtPutString @ed, Chr$(wParam), @ed.xCaret, @ed.yCaret, 0, -1, 4
              If @ed.undoPos Then @ed.@undo[@ed.undoPos - 1].uActionCount = 1
           Else
              edmTxtPutString @ed, Chr$(wParam), @ed.xCaret, @ed.yCaret, 0, -1, 1
           End If
        End If
        Incr @ed.xCaret
        SendMessage hWnd, %EDM_RESETSEL, 0, 0

     Case Else : Exit Function
  End Select

  If @ed.SelStartX <> @ed.SelEndX Or @ed.SelStartY <> @ed.SelEndY Then
     If wParam <> %VK_TAB Then SendMessage hWnd, %EDM_RESETSEL, 0, 0
  End If
  If @ed.yCaret < 0 Then @ed.yCaret = 0
  If @ed.xCaret < 0 Then @ed.xCaret = 0
  If @ed.yCaret > @ed.UBtxBuf Then @ed.yCaret = @ed.UBtxBuf
  If @ed.confineCaret Then
     If @ed.xCaret > edmTxtGetLineLen(@ed, @ed.yCaret) Then @ed.xCaret = edmTxtGetLineLen(@ed, @ed.yCaret)
     @ed.xCaretOld = @ed.xCaret
  End If
  SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0
  SendMessage hWnd, %EDM_REFRESH, 0, -1

End Function

'
' WM_COMMAND message handler
'
Function ctmCommand(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
  Local ed As EDM32DATA Ptr
  ed = GetWindowLong(hWnd, 0)
  SendMessage @ed.hParent, %WM_COMMAND, wParam, lParam
End Function

'
' %WM_CTLCOLORSTATIC message handler, for scroll-info label
'
Function ctmCtlColorStatic(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
  SetBkColor wParam, GetSysColor(%COLOR_INFOBK)
  SetTextColor wParam, GetSysColor(%COLOR_INFOTEXT)
  Function = GetSysColorBrush(%COLOR_INFOBK)
End Function

'
' WM_DESTROY message handler
'
Function ctmDestroy(ByVal hWnd As Dword) As Long
  Local ed As EDM32DATA Ptr
  ed = GetWindowLong(hWnd, 0) 'Get stored data and clean up memory

  If @ed.hBrushScrlBar  Then DeleteObject @ed.hBrushScrlBar
  If @ed.hBrushBMark    Then DeleteObject @ed.hBrushBMark
  If @ed.hFont          Then DeleteObject @ed.hFont
  If @ed.hbit           Then DeleteObject SelectObject(@ed.memDC, @ed.hbit)
  If @ed.memDC          Then DeleteDC @ed.memDC

  If @ed.pData          Then Call HeapFree(GetProcessHeap(), 0, ByVal @ed.pData)
  If @ed.pDataEndPos    Then Call HeapFree(GetProcessHeap(), 0, ByVal @ed.pDataEndPos)
  If @ed.pStartCount    Then Call HeapFree(GetProcessHeap(), 0, ByVal @ed.pStartCount)
  If @ed.pEndCount      Then Call HeapFree(GetProcessHeap(), 0, ByVal @ed.pEndCount)

  If @ed.rData          Then Call HeapFree(GetProcessHeap(), 0, ByVal @ed.rData)
  If @ed.rDataEndPos    Then Call HeapFree(GetProcessHeap(), 0, ByVal @ed.rDataEndPos)
  If @ed.rStartCount    Then Call HeapFree(GetProcessHeap(), 0, ByVal @ed.rStartCount)
  If @ed.rEndCount      Then Call HeapFree(GetProcessHeap(), 0, ByVal @ed.rEndCount)

  If @ed.txt            Then Call HeapFree(@ed.heapTxt, 0&, ByVal @ed.txt)
  If @ed.heapTxt        Then Call HeapDestroy(@ed.heapTxt)
  If @ed.LineEnd        Then Call HeapFree(@ed.heapLineEnd, 0&, ByVal @ed.LineEnd)
  If @ed.heapLineEnd    Then Call HeapDestroy(@ed.heapLineEnd)

  If @ed.undo           Then Call HeapFree(@ed.heapUndo, 0&, ByVal @ed.undo)
  If @ed.heapUndo       Then Call HeapDestroy(@ed.heapUndo)
  If @ed.undotxt        Then Call HeapFree(@ed.heapUndotxt, 0&, ByVal @ed.undotxt)
  If @ed.heapUndotxt    Then Call HeapDestroy(@ed.heapUndotxt)

  If @ed.bMarks         Then Call HeapFree(@ed.heapbMarks, 0&, ByVal @ed.bMarks)
  If @ed.heapbMarks     Then Call HeapDestroy(@ed.heapbMarks)
  If ed                 Then Call HeapFree(GetProcessHeap(), 0, ByVal ed)
End Function

'
' WM_SETFONT message handler
'
Function ctmSetFont(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
  If wParam = GetStockObject(%DEFAULT_GUI_FONT) Or _
     wParam = GetStockObject(%DEVICE_DEFAULT_FONT) Then Exit Function 'ignore Windows attempt to set any of these

  Local hDC As Dword, hFont As Dword, ed As EDM32DATA Ptr, wRect As RECT, tm As TEXTMETRIC
  ed = GetWindowLong(hWnd, 0)

  If wParam Then
     hDC = GetDC (hWnd)                          'get control's Device Context
       hFont = SelectObject(hDC, wParam)         'select new font into control's DC, old is returned
       GetTextMetrics hDC, tm                    'get font parameters
       SelectObject hDC, hFont                   'select old font back into control's DC
     ReleaseDC hWnd, hDC                  'release control's DC

     @ed.charWidth  = Max&(3, tm.tmAveCharWidth) 'Average character width
     @ed.charHeight = Max&(6, tm.tmHeight) '  + tm.tmExternalLeading   Line spacing (line height)

     GetClientRect hWnd, wRect
     SendMessage hWnd, %WM_SIZE, 0, MakLng(wRect.nRight, wRect.nBottom)
  End If
End Function

'
' %WM_SETFOCUS message handler
'
Function ctmSetFocus(ByVal hWnd As Dword, ByVal wParam As Long) As Long
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)

  If @ed.xCaret < @ed.iHscrollPos Then @ed.iHscrollPos = Max&(0, @ed.xCaret - 10)
  If @ed.xCaret > @ed.iHscrollPos + @ed.wnWidth Then @ed.iHscrollPos = @ed.xCaret - @ed.wnWidth + 10
  @ed.xPos = @ed.lMargin + @ed.charWidth * (- @ed.iHscrollPos)
  If @ed.ReadOnly = 0 Then
     If @ed.cInsOvr Then
        CreateCaret hWnd, %NULL, @ed.charWidth, @ed.charHeight
     Else
        CreateCaret hWnd, %NULL, 2, @ed.charHeight
     End If
  End If
  SetCaretPos @ed.xPos + @ed.xCaret * @ed.charWidth, (@ed.yCaret - @ed.TopLine) * @ed.charHeight
  If @ed.ReadOnly = 0 Then ShowCaret hWnd
  edmSetScrollData hwnd, @ed.wnWidth / 3, @ed.iHscrollMax, @ed.iHscrollPos, %SB_HORZ
  edmSetScrollData hwnd, @ed.wnHeight, Max&(0, @ed.UBtxBuf), @ed.TopLine, %SB_VERT
  SendMessage GetParent(hWnd), %EDM_SETINFO, 0, 1

  Local nm As NMHDR
  nm.hwndFrom = hWnd
  nm.idFrom   = @ed.id
  nm.code     = (%NM_FIRST-7) '%NM_SETFOCUS
  SendMessage @ed.hParent, %WM_NOTIFY, @ed.id , ByVal VarPtr(nm)

End Function

'************************************************************************
' edmFileExist - make sure a file or folder exists
'************************************************************************
Function edmFileExist(ByVal sfName As String) As Long
  Local hRes As Dword, tWFD As WIN32_FIND_DATA

  If Asc(sfName, -1) = 92 Then                'if a path with trailing backslash
     sfName = Left$(sfName, Len(sfName) - 1)  'remove trailing backslash
  End If

  hRes = FindFirstFile(ByVal StrPtr(sfName), tWFD)
  If hRes <> %INVALID_HANDLE_VALUE Then
     Function = %TRUE
     FindClose hRes
  End If
End Function

'
' %EDM_LOADFILE message handler
' wParam is pointer to file name, or if zero, all buffers are reset (for NEW command)
' lParam is focus flag
'
Function edmLoadFile(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
  Local ed As EDM32DATA Ptr, hFile As Dword, azpFName As Asciiz Ptr, txt As String
  ed = GetWindowLong(hWnd, 0)
  azpFName = wParam

  If IsFalse(edmTxtInit(@ed)) Then 'refresh/clear control
     SendMessage @ed.hParent, %WM_COMMAND, MakLng(@ed.id, %EN_ERRSPACE), hwnd
     Exit Function  'break action on failure
  End If

  If Len(@azpFName) Then
     If edmFileExist(@azpFName) = 0 Then
        Function = -1 : Exit Function     'return -1 on failure
     End If

     hFile = FreeFile
     Open @azpFName For Binary As hFile Len = 8192
        If Err Then
           Function = -1 : Exit Function 'return -1 on failure
        Else
           @ed.fName = @azpFName
        End If
        Get$ hFile, Lof(hFile), txt
     Close hFile

     If Len(txt) Then
        If InStr(txt, Chr$(0)) Then                        'binary file - adjust for better view..
           SendMessage hWnd, %EDM_SETREADONLY, 1, 0                           'set readonly flag
           If InStr(txt, Chr$(0, 0)) Then Replace Chr$(0, 0) With "  " In txt 'replace eventual double zeros with space
           If InStr(txt, Chr$(0)) Then txt = Remove$(txt, Chr$(0))            'remove the rest to make double-byte resource strings, etc. readable..
        End If

        Replace $CrLf With $Cr In txt   'preparation stage to ensure proper line feeds later on..
        Replace $Lf With $Cr In txt     'in case some odd CHR$(10) still is there..

        Select Case Right$(edmUCASER(@azpFName), 4)
           Case ".TXT"         : @ed.FileType = 1          'ordinary text file
           Case ".BAS", ".INC" : @ed.FileType = 2          'PB (VB?) code file
           Case ".HTM", "HTML" : @ed.FileType = 3          'HTML - crude, but works fairly well..
'              REPLACE "<BR>" WITH "<BR>" + $CR IN txt
'              REPLACE "<br>" WITH "<br>" + $CR IN txt
'              REPLACE "</"   WITH $CR + "</"   IN txt
'              REPLACE "<pre>"  WITH "<pre>"  + $CR IN txt
'              REPLACE "</pre>" WITH "</pre>" + $CR IN txt
'              REPLACE "<PRE>"  WITH "<PRE>"  + $CR IN txt
'              REPLACE "</PRE>" WITH "</PRE>" + $CR IN txt
'              REPLACE "<P>"  WITH "<P>" + CHR$(13, 13) IN txt
'              REPLACE "<p>"  WITH "<p>" + CHR$(13, 13) IN txt
'              REPLACE "<A "  WITH CHR$(13, 60, 65, 32) IN txt
'              REPLACE "<a "  WITH CHR$(13, 60, 97, 32) IN txt
'              REPLACE "><"   WITH CHR$(62, 13, 60) IN txt

           Case Else
              If Right$(edmUCASER(@azpFName), 3) = ".BI" Then     'PB code file
                 @ed.FileType = 2
              ElseIf Right$(edmUCASER(@azpFName), 3) = ".RC" Then 'PB resource file
                 @ed.FileType = 4
              ElseIf Right$(edmUCASER(@azpFName), 3) = ".RC" Then 'PB resource file
                 @ed.FileType = 4
              Else
                 @ed.FileType = 1                          'else it's a text file..?
'                 txt = edmWrapText(txt, @ed.wnWidth, 1)
              End If

        End Select

        If InStr(txt, $Tab) Then Replace $Tab With Space$(@ed.tabSize) In txt 'replace eventual tabs..
        Replace $Cr With $CrLf In txt            'restore line feeds again

        edmTxtPutString @ed, txt, 0, 0, 0, 0, 0  'place file's contents into editor
     End If

  End If

  @ed.xCaretOld = 0 : @ed.xCaret = 0             'reset these
  @ed.yCaret = 0 : @ed.Changed = 0
  edmUndoInit @ed                                'initialize undo buffer
  SendMessage hWnd, %EDM_RESETSEL, 0, 0          'reset selection variables
  SendMessage hWnd, %EDM_REFRESH, 0, lParam      'refresh control - if lParam, also set focus to it
  Function = Len(txt)                            'return length of text on success

End Function

'
' %WM_KEYDOWN message handler
'
Function ctmKeyDown(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
  Local lnLen As Long, tyCaret As Long, txCaret As Long, ed As EDM32DATA Ptr
  ed = GetWindowLong(hWnd, 0)
  @ed.LineUpdate = 0

  If @ed.kbShortCuts Then                                     'If to use internal keyboard shortcuts
     If wParam = %VK_F2 Then                                  'handle F2-key for bookmarks
        If (GetKeyState(%VK_CONTROL) And &H8000) <> 0 Then    'Ctrl+F2
           SendMessage hWnd, %EDM_SETBMARK, 0, 0              'toggle bookmark on/off
        ElseIf (GetKeyState(%VK_SHIFT) And &H8000) <> 0 Then  'Shift+F2
           SendMessage hWnd, %EDM_PREVBMARK, 0, 0             'go to previous bookmark
        ElseIf (GetKeyState(%VK_SHIFT) And &H8000) = 0 Then   'F2
           SendMessage hWnd, %EDM_NEXTBMARK, 0, 0             'go to next bookmark
        End If
        Exit Function
     End If
  End If

  Select Case wParam
     Case %VK_SHIFT, %VK_CONTROL, %VK_CAPITAL, %VK_TAB
        If wParam = %VK_CAPITAL Then SendMessage GetParent(hWnd), %EDM_SETINFO, 0, 1 'Caps lock..
        Exit Function

     Case %VK_BACK                                           'Backspace
        If (GetKeyState(%VK_CONTROL) And &H8000) <> 0 Then   'if Ctrl + BackSpace pressed (delete word to left of caret)
           If @ed.SelLen Then                                'if something is selected
              SendMessage hWnd, %EDM_DELETE, 0, -1           'delete it and refresh
              Exit Select                                    'do no more - exit
           Else
              If @ed.xCaret = 0 And @ed.yCaret  = 0 Then     'First line - pos 0, can do nothing
                 MessageBeep &HFFFFFFFF                      'give signal
                 If edmTxtGetLineLen(@ed, @ed.yCaret) = 0 Then  'if empty fist line is bookmarked, remove bookmark
                    If edmBookMarked(hWnd, @ed.yCaret) Then edmBookMarkSet hWnd, @ed.yCaret, 0
                 End If
                 Exit Select                                 'and exit
              End If
              @ed.SelEndY = @ed.yCaret                       'set SelEndY to current line
              @ed.SelStartX = edmGetWordLimits(hWnd, -1, @ed) 'set SelStartX to left word's limit
              @ed.SelEndX = @ed.xCaret                       'set SelEndX to result of edmGetWordLimits
              @ed.SelStartY = @ed.yCaret                     'set SelStartY to result of edmGetWordLimits
              @ed.SelLen = edmTxtGetSelLength(@ed)              'calculate selection's length
              SendMessage hWnd, %EDM_DELETE, 0, -1           'delete selection (word)
              Exit Select                                    'exit
           End If
        End If

        If @ed.SelLen Then                                   'BackSpace only pressed - if selection exist
           SendMessage hWnd, %EDM_DELETE, 0, -1              'delete selection
        Else                                                 'else
           If @ed.xCaret > 0 Then                            'if xCaret is above zero
              Decr @ed.xCaret : @ed.SelStartX = @ed.xCaret   'decrease xCaret and set SelStartX
              edmTxtDeleteSel @ed, 1, -1, 11                 'delete selection (one character selected)
           Else                                              'else, if xCaret is zero
              If @ed.yCaret > 0 Then                         'and line above zero
                 lnLen = edmTxtGetLineLen(@ed, @ed.yCaret)
                 If edmBookMarked(hWnd, @ed.yCaret) Then edmBookMarkSet hWnd, @ed.yCaret, 0
                 edmBookMarkDelete @ed, @ed.yCaret, 1
                 edmTxtDeleteSel @ed, 2, -1, 11
                 Decr @ed.yCaret
                 @ed.xCaret = edmTxtGetLineLen(@ed, @ed.yCaret) - lnLen
              Else
                 MessageBeep &HFFFFFFFF
                 If edmTxtGetLineLen(@ed, @ed.yCaret) = 0 Then
                    If edmBookMarked(hWnd, @ed.yCaret) Then edmBookMarkSet hWnd, @ed.yCaret, 0
                 End If
                 Exit Function
              End If
           End If
           SendMessage hWnd, %EDM_RESETSEL, 0, 0
        End If
        SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0

  Case %VK_DELETE
     If @ed.SelLen Then                               'if anything is selected
        If (GetKeyState(%VK_SHIFT)  And &H8000) Then  'handle Shift+Delete = Cut
            SendMessage hWnd, %WM_CUT, 0, 0
        Else
            SendMessage hWnd, %EDM_DELETE, 0, -1      'else it's plain Delete
        End If

     Else
        If edmTxtGetLineLen(@ed, @ed.yCaret) Then        'If line has text
           If @ed.xCaret < edmTxtGetLineLen(@ed, @ed.yCaret) Then 'and xCaret is before line end
              edmTxtDeleteSel @ed, 1, -1, 12
           Else                                       'if xcaret is at end of line
              If @ed.yCaret < @ed.UBtxBuf Then        'If line has text and yCaret is below last line
                 edmBookMarkDelete @ed, @ed.yCaret + 1, 1
                 edmTxtDeleteSel @ed, 1, -1, 12
                 SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0
                 InvalidateRect hwnd, ByVal %NULL, 0 : UpdateWindow hwnd
              Else                                    'else nothing more to delete
                 MessageBeep &HFFFFFFFF
                 If edmBookMarked(hWnd, @ed.yCaret) Then edmBookMarkSet hWnd, @ed.yCaret, 0
              End If
           End If
        Else                                          'else if line has no text
           If @ed.yCaret < @ed.UBtxBuf Then           'and yCaret is below last line
              edmBookMarkDelete @ed, @ed.yCaret, 1
              edmTxtDeleteSel @ed, 1, -1, 12
              SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0
              InvalidateRect hwnd, ByVal %NULL, 0 : UpdateWindow hwnd
           Else
              MessageBeep &HFFFFFFFF                   'else nothing more to delete
              If edmBookMarked(hWnd, @ed.yCaret) Then edmBookMarkSet hWnd, @ed.yCaret, 0
           End If
        End If
     End If

  Case %VK_INSERT
     If (GetKeyState(%VK_CONTROL) And &H8000) <> 0 Then   'Ctrl + Insert = copy
        SendMessage hWnd, %EDM_COPY, 0, 0
        Exit Function
     ElseIf (GetKeyState(%VK_SHIFT) And &H8000) <> 0 Then 'Shift + Insert = paste
        If @ed.ReadOnly Then Exit Function
        SendMessage hWnd, %EDM_PASTE, 0, 0
        Exit Function
     Else
        If @ed.ReadOnly Then Exit Function
        @ed.cInsOvr = Not @ed.cInsOvr
        If @ed.cInsOvr Then
           CreateCaret hWnd, %NULL, @ed.charWidth, @ed.charHeight
        Else
           CreateCaret hWnd, %NULL, 2, @ed.charHeight
        End If
        ShowCaret hWnd
     End If

  Case %VK_HOME
     If @ed.ScrollMode Then
        SendMessage hWnd, %WM_VSCROLL, MakLng(%SB_TOP,0), 0
        Exit Function
     End If

     If (GetKeyState(%VK_SHIFT) And &H8000) = 0 Then
        If @ed.SelLen Then SendMessage hWnd, %EDM_RESETSEL, -1, 0
     End If
     If @ed.yCaret = 0 And @ed.xCaret = 0 Then Exit Function

     If (GetKeyState(%VK_CONTROL) And &H8000) <> 0 Then
        txCaret = 0 : tyCaret = 0
     Else
        txCaret = 0 : tyCaret = @ed.yCaret
     End If
     If @ed.confineCaret Then @ed.xCaretOld = txCaret

     SendMessage hWnd, %EDM_CARETHANDLER, txCaret, tyCaret
     Exit Function

  Case %VK_END
     If @ed.ScrollMode Then
        SendMessage hWnd, %WM_VSCROLL, MakLng(%SB_BOTTOM, 0), 0
        Exit Function
     End If

     If (GetKeyState(%VK_SHIFT) And &H8000) = 0 Then
        If @ed.SelLen Then SendMessage hWnd, %EDM_RESETSEL, -1, 0
     End If
     If @ed.yCaret = @ed.UBtxBuf And @ed.xCaret = edmTxtGetLineLen(@ed, @ed.yCaret) Then Exit Function

     If (GetKeyState(%VK_CONTROL) And &H8000) <> 0 Then
        tyCaret = @ed.UBtxBuf : txCaret = edmTxtGetLineLen(@ed, tyCaret)
     Else
        tyCaret = @ed.yCaret : txCaret = edmTxtGetLineLen(@ed, @ed.yCaret)
     End If

     SendMessage hWnd, %EDM_CARETHANDLER, txCaret, tyCaret
     Exit Function

  Case %VK_PRIOR 'PgUp
     If @ed.ScrollMode Then
        SendMessage hWnd, %WM_VSCROLL, MakLng(%SB_PAGEUP, 0), 0
        Exit Function
     End If

     If (GetKeyState(%VK_SHIFT) And &H8000) = 0 Then
        If @ed.SelLen Then SendMessage hWnd, %EDM_RESETSEL, -1, 0
     End If

     If @ed.yCaret < @ed.TopLine Then @ed.yCaret = @ed.TopLine + 1
     If @ed.yCaret > @ed.TopLine + @ed.wnHeight Then
        @ed.yCaret = @ed.TopLine + @ed.wnHeight
     End If
     If @ed.yCaret = 0 Then Exit Function

     If (GetKeyState(%VK_CONTROL) And &H8000) <> 0 Then
        txCaret = @ed.xCaret : tyCaret = 0
     Else
        txCaret     = @ed.xCaret
        tyCaret     = Max&(0, @ed.yCaret - @ed.wnHeight + 1)
        @ed.TopLine = Max&(0, @ed.TopLine - @ed.wnHeight + 1)
     End If

     If @ed.confineCaret Then
        If txCaret < @ed.xCaretOld Then
           txCaret = Min&(@ed.xCaretOld, edmTxtGetLineLen(@ed, tyCaret))
        End If
     End If

     SendMessage hWnd, %EDM_CARETHANDLER, txCaret, tyCaret
     If (GetKeyState(%VK_SHIFT) And &H8000) = 0 Then
        InvalidateRect hwnd, ByVal %NULL, 0 : UpdateWindow hwnd
     End If
     Exit Function

  Case %VK_NEXT  'PgDn
     If @ed.ScrollMode Then
        SendMessage hWnd, %WM_VSCROLL, MakLng(%SB_PAGEDOWN, 0), 0
        Exit Function
     End If

     If (GetKeyState(%VK_SHIFT) And &H8000) = 0 Then
        If @ed.SelLen Then SendMessage hWnd, %EDM_RESETSEL, -1, 0
     End If

     If @ed.yCaret < @ed.TopLine Then @ed.yCaret = @ed.TopLine
     If @ed.yCaret > @ed.TopLine + @ed.wnHeight - 1 Then
        @ed.yCaret = @ed.TopLine + @ed.wnHeight - 1
     End If
     If @ed.yCaret = @ed.UBtxBuf Then Exit Function

     If (GetKeyState(%VK_CONTROL) And &H8000) <> 0 Then
        txCaret = @ed.xCaret : tyCaret = @ed.UBtxBuf
     Else
        tyCaret     = Min&(@ed.UBtxBuf, @ed.yCaret + @ed.wnHeight - 1)
        @ed.TopLine = Min&(Max&(0, @ed.UBtxBuf - @ed.wnHeight + 1), @ed.TopLine + @ed.wnHeight - 1)
     End If

     If @ed.confineCaret Then
        If txCaret < @ed.xCaretOld Then
           txCaret = Min&(@ed.xCaretOld, edmTxtGetLineLen(@ed, tyCaret))
        End If
     End If

     SendMessage hWnd, %EDM_CARETHANDLER, txCaret, tyCaret
     If (GetKeyState(%VK_SHIFT) And &H8000) = 0 Then
        InvalidateRect hwnd, ByVal %NULL, 0 : UpdateWindow hwnd
     End If
     Exit Function

  Case %VK_LEFT
     If @ed.ScrollMode Then
        SendMessage hWnd, %WM_HSCROLL, MakLng(%SB_LINELEFT, 0), 0
        Exit Function
     End If

     If (GetKeyState(%VK_SHIFT) And &H8000) = 0 Then
        If @ed.SelLen Then SendMessage hWnd, %EDM_RESETSEL, -1, 0
     End If
     If @ed.yCaret = 0 And @ed.xCaret = 0 Then Exit Function

     txCaret = @ed.xCaret - 1 : tyCaret = @ed.yCaret
     If txCaret < 0 And tyCaret > 0 Then
        Decr tyCaret
        txCaret = edmTxtGetLineLen(@ed, tyCaret)
     End If

     If (GetKeyState(%VK_CONTROL) And &H8000) <> 0 Then
        txCaret = edmGetWordLimits(hWnd, -1, @ed)
     End If

     If @ed.confineCaret Then @ed.xCaretOld = Max&(0, txCaret)
     SendMessage hWnd, %EDM_CARETHANDLER, txCaret, tyCaret
     Exit Function

  Case %VK_RIGHT
     If @ed.ScrollMode Then
        SendMessage hWnd, %WM_HSCROLL, MakLng(%SB_LINERIGHT, 0), 0
        Exit Function
     End If

     If (GetKeyState(%VK_SHIFT) And &H8000) = 0 Then
        If @ed.SelLen Then SendMessage hWnd, %EDM_RESETSEL, -1, 0
     End If
     If @ed.yCaret = @ed.UBtxBuf And @ed.xCaret = edmTxtGetLineLen(@ed, @ed.yCaret) Then Exit Function

     txCaret = @ed.xCaret + 1 : tyCaret = @ed.yCaret
     If @ed.confineCaret Then
        If txCaret > edmTxtGetLineLen(@ed, tyCaret) And tyCaret < @ed.UBtxBuf Then
           Incr tyCaret
           txCaret = 0
        End If
     End If

     If (GetKeyState(%VK_CONTROL) And &H8000) <> 0 Then
        If @ed.confineCaret = 0 Then
           If txCaret > edmTxtGetLineLen(@ed, tyCaret) And tyCaret < @ed.UBtxBuf Then
              Incr tyCaret
              txCaret = 0
           End If
        End If
        txCaret = edmGetWordLimits(hWnd, 0, @ed)
     End If
     If @ed.confineCaret Then
        @ed.xCaretOld = Min&(txCaret, edmTxtGetLineLen(@ed, tyCaret))
     End If
     SendMessage hWnd, %EDM_CARETHANDLER, txCaret, tyCaret
     Exit Function

  Case %VK_UP
     If @ed.ScrollMode Then
        SendMessage hWnd, %WM_VSCROLL, MakLng(%SB_LINEUP, 0), 0
        Exit Function
     End If

     If (GetKeyState(%VK_SHIFT) And &H8000) = 0 And @ed.hTimer = 0 Then
        If @ed.SelLen Then SendMessage hWnd, %EDM_RESETSEL, -1, 0
        If @ed.SelLen = 0 Then @ed.LineUpdate = 3
     End If

     If @ed.yCaret < @ed.TopLine Then
        @ed.yCaret = @ed.TopLine + 1
        @ed.LineUpdate = 0
     End If

     If @ed.yCaret > @ed.TopLine + @ed.wnHeight Then
        @ed.yCaret = @ed.TopLine + @ed.wnHeight
        @ed.LineUpdate = 0
     End If

     If @ed.yCaret = 0 Then
        @ed.LineUpdate = 0
        Exit Function
     End If

     txCaret = @ed.xCaret : tyCaret = @ed.yCaret - 1

     If @ed.confineCaret Then
        If txCaret < @ed.xCaretOld Then
           txCaret = Min&(@ed.xCaretOld, edmTxtGetLineLen(@ed, tyCaret))
        End If
     End If

     SendMessage hWnd, %EDM_CARETHANDLER, txCaret, tyCaret
     Exit Function

  Case %VK_SCROLL
     @ed.ScrollMode = SendMessage(hWnd, %EDM_GETSCROLLOCK, 0, 0)
     SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0
     SendMessage hWnd, %EDM_REFRESH, 0, -1

  Case %VK_NUMLOCK
     SendMessage GetParent(hWnd), %EDM_SETINFO, 0, 1

  Case %VK_DOWN
     If @ed.ScrollMode Then
        SendMessage hWnd, %WM_VSCROLL, MakLng(%SB_LINEDOWN, 0), 0
        Exit Function
     End If

     If (GetKeyState(%VK_SHIFT) And &H8000) = 0 And @ed.hTimer = 0 Then
        If @ed.SelLen Then SendMessage hWnd, %EDM_RESETSEL, -1, 0
        If @ed.SelLen = 0 Then @ed.LineUpdate = 2
     End If

     If @ed.yCaret < @ed.TopLine Then
        @ed.yCaret = @ed.TopLine - 1
        @ed.LineUpdate = 0
     End If

     If @ed.yCaret > @ed.TopLine + @ed.wnHeight - 1 Then
        @ed.yCaret = @ed.TopLine + @ed.wnHeight - 2
        @ed.LineUpdate = 0
     End If

     If @ed.yCaret = @ed.UBtxBuf Then
        @ed.LineUpdate = 0
        Exit Function
     End If
     txCaret = @ed.xCaret : tyCaret = Min&(@ed.yCaret + 1, @ed.UBtxBuf)

     If @ed.confineCaret Then
        If txCaret < @ed.xCaretOld Then
           txCaret = Min&(@ed.xCaretOld, edmTxtGetLineLen(@ed, tyCaret))
        End If
     End If

     SendMessage hWnd, %EDM_CARETHANDLER, txCaret, tyCaret
     Exit Function

  Case Else '32 TO 255, 9
     Exit Function
  End Select

  If @ed.yCaret < 0 Then @ed.yCaret = 0
  If @ed.xCaret < 0 Then @ed.xCaret = 0
  If @ed.yCaret > @ed.UBtxBuf Then @ed.yCaret = @ed.UBtxBuf
  If @ed.confineCaret Then
     If @ed.xCaret > edmTxtGetLineLen(@ed, @ed.yCaret) Then @ed.xCaret = edmTxtGetLineLen(@ed, @ed.yCaret)
  End If
  If @ed.xCaret < @ed.iHscrollPos Then @ed.iHscrollPos = Max&(0, @ed.xCaret - 10)
  If @ed.xCaret > @ed.iHscrollPos + @ed.wnWidth Then @ed.iHscrollPos = @ed.xCaret - @ed.wnWidth + 10

  @ed.xPos = @ed.lMargin + @ed.charWidth * (- @ed.iHscrollPos)
  edmSetScrollData hwnd, @ed.wnWidth / 3, @ed.iHscrollMax, @ed.iHscrollPos, %SB_HORZ
  edmSetScrollData hwnd, @ed.wnHeight, Max&(0, @ed.UBtxBuf), @ed.TopLine, %SB_VERT
  InvalidateRect hwnd, ByVal %NULL, 0 : UpdateWindow hwnd
  SetCaretPos  @ed.xPos + @ed.xCaret * @ed.charWidth, (@ed.yCaret - @ed.TopLine) * @ed.charHeight
  SendMessage GetParent(hWnd), %EDM_SETINFO, 0, 1
  @ed.LineUpdate = 0

End Function

'
' %WM_KEYUP message handler
'
Function ctmKeyUp(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
  Local ed As EDM32DATA Ptr, nm As NMHDR
  ed = GetWindowLong(hWnd, 0)

  nm.hwndFrom = hWnd
  nm.idFrom   = @ed.id
  nm.code     = (%NM_FIRST-4)  '%NM_RETURN
  SendMessage @ed.hParent, %WM_NOTIFY, @ed.id , ByVal VarPtr(nm)
End Function

'
' WM_KILLFOCUS message handler
'
Function ctmKillFocus(ByVal hWnd As Dword, ByVal wParam As Long) As Long
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)

  Local nm As NMHDR
  nm.hwndFrom = hWnd
  nm.idFrom   = @ed.id
  nm.code     = (%NM_FIRST-8) '%NM_KILLFOCUS
  SendMessage @ed.hParent, %WM_NOTIFY, @ed.id , ByVal VarPtr(nm)

  If @ed.ReadOnly Then Exit Function
  HideCaret hWnd
  DestroyCaret

End Function

'
' %WM_SIZE message handler
'
Function ctmSize(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
  If hWnd = 0 Or wParam = %SIZE_MINIMIZED Then Exit Function
  Local sBarWidth As Long, ed As EDM32DATA Ptr, wRect As RECT
  ed = GetWindowLong(hWnd, 0)
  @ed.LineUpdate = 0
  If @ed.charWidth = 0 Or @ed.charHeight = 0 Then Exit Function

  GetClientRect hWnd, wRect
  sBarWidth = GetSystemMetrics(%SM_CXVSCROLL)
  @ed.wnWidth = Max&(1, (wRect.nRight - sBarWidth) / @ed.charWidth - 1) 'Get window size in characters
  @ed.wnHeight = Max&(1, (HiWrd(lParam) - sBarWidth) \ @ed.charHeight + 1)

  @ed.iHscrollMax = Max&(0, 1024 - @ed.wnWidth) '1024, max number of characters in a line..
  '@ed.iHscrollMax = MAX&(0, edmTxtGetMaxLineLen(hWnd) - @ed.wnWidth / 3)
  @ed.iHscrollPos = Min&(@ed.iHscrollPos, @ed.iHscrollMax)
  edmSetScrollData hwnd, @ed.wnWidth / 3, @ed.iHscrollMax, @ed.iHscrollPos, %SB_HORZ
  edmSetScrollData hwnd, @ed.wnHeight, Max&(0, @ed.UBtxBuf), @ed.TopLine, %SB_VERT

  InvalidateRect hwnd, ByVal %NULL, 0 : UpdateWindow hwnd
End Function

'
' %WM_TIMER message handler
'
Function ctmTimer(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
  If wParam = 22 Then      'scroll up when selecting and mouse pointer is above control
     SendMessage hWnd, %WM_KEYDOWN, %VK_UP, 0
  ElseIf wParam = 23 Then  'scroll down when selecting and mouse pointer is below control
     SendMessage hWnd, %WM_KEYDOWN, %VK_DOWN, 0
  End If
End Function

'
' WM_VSCROLL message handler
'
Function ctmVscroll(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
  Local ed As EDM32DATA Ptr, si As SCROLLINFO
  ed = GetWindowLong(hWnd, 0)

  Select Case LoWrd(wParam)
     Case %SB_LINEDOWN
        @ed.TopLine = Min&(@ed.TopLine + 1, @ed.UBtxBuf - @ed.wnHeight + 1)
        If @ed.LineUpdate <> 2 Then
           If @ed.SelLen = 0 And @ed.OldTopLine < @ed.UBtxBuf - @ed.wnHeight Then @ed.LineUpdate = 2
        End If

     Case %SB_LINEUP
        @ed.TopLine = Max&(0, @ed.TopLine - 1)
        If @ed.LineUpdate <> 3 Then
           If @ed.SelLen = 0 And @ed.OldTopLine > 0 Then @ed.LineUpdate = 3
        End If

     Case %SB_PAGEDOWN   : @ed.TopLine = Min&(@ed.TopLine + @ed.wnHeight, @ed.UBtxBuf - @ed.wnHeight + 1)
     Case %SB_PAGEUP     : @ed.TopLine = Max&(0, @ed.TopLine - @ed.wnHeight)
     Case %SB_THUMBTRACK
        si.cbSize = SizeOf(si)
        si.fMask = %SIF_TRACKPOS
        GetScrollInfo hWnd, %SB_VERT, si   '<- enables tracking above 16-bit limit, 65536 lines..
        @ed.TopLine = si.nTrackPos

     Case %SB_BOTTOM     : @ed.TopLine = Max&(0, @ed.UBtxBuf - @ed.wnHeight + 1)
     Case %SB_TOP        : @ed.TopLine = 0
     Case Else
        If @ed.hScrollLabel Then
           If DestroyWindow(@ed.hScrollLabel) <> 0 Then @ed.hScrollLabel = 0
        End If
        Exit Function
  End Select
  @ed.TopLine = Max&(0, @ed.TopLine)
  edmSetScrollData hwnd, @ed.wnHeight, Max&(0, @ed.UBtxBuf), @ed.TopLine, %SB_VERT
  InvalidateRect hwnd, ByVal %NULL, 0
  UpdateWindow hWnd

  If LoWrd(wParam) = %SB_THUMBTRACK Then
     If @ed.showScrollInfo Then           'if to show info label on thumb scroll
        If @ed.hScrollLabel = 0 Then @ed.hScrollLabel = MakeScrollInfoLabel(hWnd)
        If @ed.hScrollLabel <> 0 Then
           Local r As RECT, txt As String
           GetClientRect hWnd, r
           txt = @ed.infoLeadText & Str$(@ed.TopLine + 1) + " -" + _  'visible top line
                                    Str$(@ed.TopLine + @ed.wnHeight)  'visible bottom line
           SetWindowText @ed.hScrollLabel, ByVal StrPtr(txt)
           SetWindowPos @ed.hScrollLabel, %NULL, _
                        r.nRight - 105, _
                        r.nBottom / 2 - 9, 0, 0, _
                        %SWP_NOSIZE Or %SWP_NOZORDER Or %SWP_SHOWWINDOW 'SetWindowPos redraws automatically
           UpdateWindow @ed.hScrollLabel                                'UpdateWindow restores NC borders
        End If
     End If
  End If

  SetCaretPos @ed.xPos + @ed.xCaret * @ed.charWidth, (@ed.yCaret - @ed.TopLine) * @ed.charHeight

End Function

'
' WM_HSCROLL message handler
'
Function ctmHscroll(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
  Local hScrlInc As Long, ed As EDM32DATA Ptr
  ed = GetWindowLong(hWnd, 0)

  Select Case LoWrd(wParam)
     Case %SB_LINELEFT      : hScrlInc = -1
     Case %SB_LINERIGHT     : hScrlInc =  1
     Case %SB_PAGELEFT      : hScrlInc = -@ed.wnWidth / 3
     Case %SB_PAGERIGHT     : hScrlInc =  @ed.wnWidth / 3
     Case %SB_THUMBTRACK    : hScrlInc = HiWrd(wParam) - @ed.iHscrollPos
     Case Else              : hScrlInc = 0
  End Select

  hScrlInc = Max&(-@ed.iHscrollPos, Min&(hScrlInc, @ed.iHscrollMax - @ed.iHscrollPos))

  If hScrlInc <> 0 Then
     @ed.iHscrollPos = @ed.iHscrollPos + hScrlInc
     @ed.xPos = @ed.lMargin + @ed.charWidth * (- @ed.iHscrollPos)
     edmSetScrollData hwnd, @ed.wnWidth / 3, @ed.iHscrollMax, @ed.iHscrollPos, %SB_HORZ
     InvalidateRect hwnd, ByVal %NULL, 0
'     UpdateWindow hWnd
     SetCaretPos @ed.xPos + @ed.xCaret * @ed.charWidth, (@ed.yCaret - @ed.TopLine) * @ed.charHeight
  End If

End Function

'
' WM_PAINT message handler
'
Function ctmPaint(ByVal hWnd As Dword, ByVal wParam As Long) As Long
  Local y        As Long, StartP As Long, EndP As Long, hFontOld As Dword, _
        hBrush   As Dword, hBrushOld As Dword, NotInSel As Long, hPen As Dword, _
        ed       As EDM32DATA Ptr, _
        txt      As String, _
        ps       As PAINTSTRUCT, _
        wRect    As RECT, rct As RECT
  ReDim pl(2)    As POINTAPI

  ed = GetWindowLong(hWnd, 0)
  If @ed.OldHscrollPos <> @ed.iHscrollPos Then @ed.LineUpdate = 0

  GetClientRect hWnd, wRect
  GdiFlush
  BeginPaint hWnd, ps                              'initiate window painting

  If @ed.LineUpdate = 0 Or @ed.memDC = 0 Then
     If @ed.hbit  Then DeleteObject SelectObject(@ed.memDC, @ed.hbit)
     If @ed.memDC Then DeleteDC @ed.memDC
     @ed.memDC = CreateCompatibleDC(ps.hdc)
     @ed.hbit  = CreateCompatibleBitmap(ps.hdc, wRect.nRight, wRect.nBottom)
     If @ed.hbit  Then @ed.hbit = SelectObject(@ed.memDC, @ed.hbit)
  End If
  hBrush = CreateSolidBrush(@ed.ColorPage) 'GetSysColorBrush(%COLOR_WINDOW)
  If hBrush   Then hBrushOld = SelectObject(@ed.memDC, hBrush)    'select new brush into DC, old is returned
  If @ed.hFont Then hFontOld = SelectObject(@ed.memDC, @ed.hFont) 'select new font into memDC, old is returned
  SetBkColor @ed.memDC, @ed.ColorTxtBk

  If @ed.LineUpdate > 0 And @ed.SelLen = 0 Then    'if single line scroll and no selection
     If @ed.LineUpdate = 1 Then                    'character input/delete..
        StartP = @ed.yCaret - @ed.TopLine
        @ed.LineUpdate = 0
     Else                                          'scrolling
        If @ed.TopLine = @ed.OldTopLine + 1 Then   'scrolling down
           StartP = @ed.wnHeight - 1
        Else                                       'else scrolling up
           StartP = 0
        End If
     End If
     EndP          = StartP
     wRect.nTop    = StartP * @ed.charHeight
     wRect.nBottom = (EndP + 1) * @ed.charHeight
  Else
     StartP = 0                '- 1                'ELSE redraw all visible lines
     EndP   = Min&(@ed.wnHeight - 1, Max&(0, @ed.UBtxBuf - @ed.TopLine))
  End If

  If @ed.LineUpdate > 0 Then     'if single line scroll and no selection
     Local rc2 As Rect
     GetClientRect hWnd, rc2
     If @ed.LineUpdate = 2 Then  '<- scroll down - text moves up one line
        rc2.nTop = rc2.nTop + @ed.charHeight
        rc2.nBottom = @ed.wnHeight * @ed.charHeight
        BitBlt @ed.memDC, 0, 0, rc2.nRight, rc2.nBottom, @ed.memDC, 0, @ed.charHeight, %SRCCOPY
     Else
        rc2.nBottom = (@ed.wnHeight - 1) * @ed.charHeight
        BitBlt @ed.memDC, 0, @ed.charHeight, rc2.nRight, rc2.nBottom, @ed.memDC, 0, 0, %SRCCOPY
     End If
  End If

  FillRect @ed.memDC, wRect, hBrush

  'DRAW LEFT, VERTICAL BAR
  rct = wRect : rct.nLeft = @ed.xPos - @ed.lMargin : rct.nRight = rct.nLeft + @ed.lMargin - 2
  If @ed.LineUpdate = 2 And rc2.nBottom > rct.nBottom Then rct.nBottom = rc2.nBottom
  @ed.LineUpdate = 0
  FillRect @ed.memDC, rct, @ed.hBrushScrlBar

  hPen = CreatePen(%PS_SOLID, 1, GetSysColor(%COLOR_SCROLLBAR)) 'RGB(128,128,128))
  If hPen Then
     hPen = SelectObject(@ed.memDC, hPen)
     '******************************
     'MoveTo @ed.memDC, @ed.xPos - 2, rct.nTop'0
     LineTo @ed.memDC, @ed.xPos - 2, rct.nBottom 'wRect.nBottom
     DeleteObject SelectObject(@ed.memDC, hPen)
  End If

  For y = StartP To EndP
     NotInSel = 1 'exclude fully selected lines for better speed
     If @ed.SelLen Then
        If @ed.SelStartY < @ed.SelEndY - 1 Then
           If y + @ed.TopLine > @ed.SelStartY And y + @ed.TopLine < @ed.SelEndY Then NotInSel = 0
        ElseIf @ed.SelEndY < @ed.SelStartY  + 1 Then 'backwards selection
           If y + @ed.TopLine > @ed.SelEndY And y + @ed.TopLine < @ed.SelStartY Then NotInSel = 0
        End If
     End If

     If NotInSel = 1 Then
        txt = edmTxtGetLine(hWnd, y + @ed.TopLine)
        If @ed.ShowNonPrintChars Then
           txt = txt & Chr$(182)
           Replace " " With Chr$(183) In txt
        End If
        SetTextColor @ed.memDC, @ed.ColorTxt
        TextOut @ed.memDC, @ed.xPos, y * @ed.charHeight, ByVal StrPtr(txt), Len(txt)

        If @ed.ShowSyntaxColor Then
           Select Case @ed.FileType
              Case 2 : ctmPaintBasic hWnd, txt, y * @ed.charHeight
              Case 3 : ctmPaintHtml  hWnd, txt, y * @ed.charHeight
              Case 4 : ctmPaintRc    hWnd, txt, y * @ed.charHeight
           End Select
        End If
     End If

     If edmBookMarked(hWnd, y + @ed.TopLine) Then 'Bookmark check & painting
        If @ed.hBrushBMark Then SelectObject @ed.memDC, @ed.hBrushBMark
        hPen = CreatePen(%PS_SOLID, 1, RGB(128,0,0))
        If hPen Then hPen = SelectObject(@ed.memDC, hPen)
        pl(0).x = @ed.xPos - @ed.lMargin + 2 : pl(0).y = y * @ed.charHeight + 2
        pl(1).x = @ed.xPos - 4 : pl(1).y = y * @ed.charHeight + @ed.charHeight / 2
        pl(2).x = pl(0).x      : pl(2).y = y * @ed.charHeight + @ed.charHeight - 2
        Polygon @ed.memDC, ByVal VarPtr(pl(0)), 3
        DeleteObject SelectObject(@ed.memDC, hPen)
        If hBrush Then SelectObject @ed.memDC, hBrush
     End If
  Next

  If @ed.SelLen Then ctmPaintSelection hWnd

  'DRAW HORIZONTAL LINES
  If @ed.ShowGridLines Then
     hPen = CreatePen(%PS_SOLID, 1, @ed.ColorGridLine)
     If hPen Then
        hPen = SelectObject(@ed.memDC, hPen)
        For y = 1 To @ed.wnHeight
          '*******************************************
          'MoveTo @ed.memDC, 0, y * @ed.charHeight
           LineTo @ed.memDC, wRect.nRight, y * @ed.charHeight
        Next
        DeleteObject SelectObject(@ed.memDC, hPen)
     End If
  End If

  BitBlt ps.hdc, 0, 0, ps.rcPaint.nRight, ps.rcPaint.nBottom, @ed.memDC, 0, 0, %SRCCOPY
  @ed.OldTopLine = @ed.TopLine
  @ed.OldHscrollPos = @ed.iHscrollPos

  If hFontOld  Then SelectObject @ed.memDC, hFontOld    'select old font back into memDC
  If hBrushOld Then SelectObject @ed.memDC, hBrushOld   'select original brush back
  If hBrush    Then DeleteObject hBrush                 'delete temporary background brush
  EndPaint hWnd, ps                                     'end window painting
End Function

'
' Syntax color parser for PB/DLL code
'
Sub ctmPaintBasic(ByVal hWnd As Dword, ByVal Buf As String, ByVal yLine As Long)
  Local ii As Long, wPos As Long, cPos As Long, Ac As Long, Result As Long, wFlag As Long, _
        q As Dword Ptr, q2 As Dword Ptr, Letter As Byte Ptr, ed As EDM32DATA Ptr, _
        txt As String, xWord As String, tx2 As String, org As String

  ed = GetWindowLong(hWnd, 0)                     'get private data
  org = Buf                                       'store original line
  edmUCASE Buf                                    'make proper UCASE for search

  Local linePos As Long
  linePos = @ed.txt + edmTxtGetLineStart(@ed, (yLine / @ed.charHeight) + @ed.TopLine) - 1
  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ' Loop through the line, using a pointer for better speed
  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  Letter = StrPtr(Buf) : wFlag = 0
  For ii = 1 To Len(Buf) + 1
     Select Case @Letter 'The characters we need to inlude in a word here
        Case  65 To 90, 35 To 38, 48 To 57, 63, 95
           If wFlag = 0 Then
              wFlag = 1 : wPos = ii               'word starts here
           End If

        Case 39 ' uncomment (rem) character -> '
           txt = Mid$(org, ii)
           SetTextColor @ed.memDC, @ed.ColorSyntaxRem
           TextOut @ed.memDC, @ed.xPos + (ii - 1) * @ed.charWidth, yLine, ByVal StrPtr(txt), Len(txt)
           Exit Sub                                 'done, next line please

        Case 34 ' text within string quotes -> "text"  (fixed to handle quotes within quotes)
           wPos = InStr(ii + 1, Buf, $Dq)    'Find match
           If wPos Then
              txt = Mid$(org, ii, wPos - ii + 1)
              SetTextColor @ed.memDC, @ed.ColorSyntaxString
              TextOut @ed.memDC, @ed.xPos + (ii - 1) * @ed.charWidth, yLine, ByVal StrPtr(txt), Len(txt)
              wPos = (wPos - ii)
              ii = ii + wPos
              Letter = Letter + wPos
              wFlag = 0
           End If

        Case 33 ' asm character -> !
           If ii = 1 Or (ii > 1 And Peek(Letter - 1) = 32) Then
              txt = Mid$(org, ii)
              cPos = InStr(txt, Any ";'")
              SetTextColor @ed.memDC, @ed.ColorSyntaxAsm

              If cPos Then 'asm uncomment character -> ; or '
                 tx2 = Mid$(txt, cPos)
                 txt = Left$(txt, cPos - 1)
                 TextOut @ed.memDC, @ed.xPos + (ii - 1) * @ed.charWidth, yLine, ByVal StrPtr(txt), Len(txt)
                 SetTextColor @ed.memDC, @ed.ColorSyntaxRem
                 TextOut @ed.memDC, @ed.xPos + (ii - 1 + Len(txt)) * @ed.charWidth, yLine, _
                               ByVal StrPtr(tx2), Len(tx2)
              Else
                 TextOut @ed.memDC, @ed.xPos + (ii - 1) * @ed.charWidth, yLine, ByVal StrPtr(txt), Len(txt)
              End If
              Exit Sub
           End If

        Case Else  'word is ready
           If wFlag Then
              xWord = Mid$(Buf, wPos, ii - wPos)   'Get word

              Ac = Max&(0, Asc(xWord) - 64)
              If Ac < 30 Then
                 q2 = @ed.pEndCount + Ac * 4
                 q = @ed.pStartCount + Ac * 4
                 Result = edmArrayScan(xWord, @ed.pData, @ed.pDataEndPos, @q, @q2)
              End If

              If Result Then
                 Result = 0
                 If @ed.kwCase = 1 Then 'uppercase - is default, so do nothing
                 ElseIf @ed.kwCase = 0 Then 'no change, so use original text
                    xWord = Mid$(org, wPos, ii - wPos)
                 ElseIf @ed.kwCase = 2 Then 'mixed case
                    edmMCASE xWord
                 ElseIf @ed.kwCase = 3 Then 'lower case
                    edmLCASE xWord
                 End If
                 SetTextColor @ed.memDC, @ed.ColorSyntaxCode
                 TextOut @ed.memDC, @ed.xPos + (wPos - 1) * @ed.charWidth, yLine, ByVal StrPtr(xWord), Len(xWord)

              Else
                 If xWord = "#PBFORMS" Then  'extra for the #PBFORMS statement
                    txt = "#PBForms" + Mid$(org, ii)
                    SetTextColor @ed.memDC, @ed.ColorSyntaxPBForms
                    TextOut @ed.memDC, @ed.xPos + (ii - 9) * @ed.charWidth, yLine, ByVal StrPtr(txt), Len(txt)
                    Exit Sub

                 ElseIf xWord = "REM" Then  'extra for the uncomment word, REM
                    txt = "REM" + Mid$(org, ii)
                    SetTextColor @ed.memDC, @ed.ColorSyntaxRem
                    TextOut @ed.memDC, @ed.xPos + (ii - 4) * @ed.charWidth, yLine, ByVal StrPtr(txt), Len(txt)
                    Exit Sub

                 ElseIf xWord = "ASM" Then  'extra for ASM
                    txt = "ASM" + Mid$(org, ii)
                    cPos = InStr(txt, ";")
                    SetTextColor @ed.memDC, @ed.ColorSyntaxAsm

                    If cPos Then 'asm uncomment character -> ;
                       tx2 = Mid$(txt, cPos)
                       txt = Left$(txt, cPos - 1)
                       TextOut @ed.memDC, @ed.xPos + (ii - 4) * @ed.charWidth, yLine, ByVal StrPtr(txt), Len(txt)
                       SetTextColor @ed.memDC, @ed.ColorSyntaxRem
                       TextOut @ed.memDC, @ed.xPos + (ii - 4 + Len(txt)) * @ed.charWidth, yLine, ByVal StrPtr(tx2), Len(tx2)
                    Else
                       TextOut @ed.memDC, @ed.xPos + (ii - 4) * @ed.charWidth, yLine, ByVal StrPtr(txt), Len(txt)
                    End If
                    Exit Sub
                 End If
              End If
              wFlag = 0
           End If
     End Select
     Incr Letter
  Next ii
End Sub

'
' Syntax color parser for HTML code
'
Sub ctmPaintHtml (ByVal hWnd As Dword, ByVal Buf As String, ByVal yLine As Long)
  Local ii As Long, wPos As Long, Ac As Long, _
        Result As Long, wFlag As Long, IsHtml As Long, _
        q As Long Ptr, q2 As Long Ptr, Letter As Byte Ptr, ed As EDM32DATA Ptr, _
        txt As String, xWord As String, org As String

  ed = GetWindowLong(hWnd, 0)               'get private data
  org = Buf                                 'store original line
  edmUCASE Buf                              'make proper UCASE for search
  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ' Loop through the line, using a pointer for better speed
  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  Letter = StrPtr(Buf) : wFlag = 0
  For ii = 1 To Len(Buf) + 1
     Select Case @Letter 'The characters we need to inlude in a word
        Case 60 : IsHtml = -1 '<

        Case  65 To 90, 35 To 38, 48 To 57, 63, 95, 45
           If IsHtml And wFlag = 0 Then
              wFlag = -1 : wPos = ii
           End If

        Case 34 ' text within string quotes -> "text"  (fixed to handle quotes within quotes)
           If IsHtml Then
              wPos = InStr(ii + 1, Buf, $Dq) 'Find match
              If wPos Then
                 txt = Mid$(org, ii, wPos - ii + 1)
                 SetTextColor @ed.memDC, @ed.ColorSyntaxString
                 TextOut @ed.memDC, @ed.xPos + (ii - 1) * @ed.charWidth, yLine, ByVal StrPtr(txt), Len(txt)
                 wPos = (wPos - ii)
                 ii = ii + wPos
                 Letter = Letter + wPos
                 wFlag = 0
              End If
           End If

        Case Else  'word is ready
           If wFlag Then
              xWord = Mid$(Buf, wPos, ii - wPos)   'Get word
              Ac = Max&(0, Asc(xWord) - 64)
              If Ac < 30 Then
                 q2 = @ed.hEndCount + Ac * 4
                 q = @ed.hStartCount + Ac * 4
                 Result = edmArrayScan(xWord, @ed.hData, @ed.hDataEndPos, @q, @q2)
              End If

              If Result Then
                 Result = 0
                 If @ed.kwCase = 1 Then 'uppercase - is default, so do nothing
                 ElseIf @ed.kwCase = 0 Then 'no change
                    xWord = Mid$(org, wPos, ii - wPos)
                 ElseIf @ed.kwCase = 2 Then 'mixed case
                    edmMCASE xWord
                 ElseIf @ed.kwCase = 3 Then 'lower case
                    edmLCASE xWord
                 End If

                 SetTextColor @ed.memDC, @ed.ColorSyntaxCode
                 TextOut @ed.memDC, @ed.xPos + (wPos - 1) * @ed.charWidth, yLine, ByVal StrPtr(xWord), Len(xWord)
              End If
              wFlag = 0
           End If
           If @Letter = 62 Then IsHtml = 0
     End Select
     Incr Letter
  Next ii

End Sub

'
' Syntax color parser for RC code
'
Sub ctmPaintRc(ByVal hWnd As Dword, ByVal Buf As String, ByVal yLine As Long)
  Local ii As Long, wPos As Long, Ac As Long, _
        Result As Long, wFlag As Long, _
        q As Dword Ptr, q2 As Dword Ptr, Letter As Byte Ptr, ed As EDM32DATA Ptr, _
        txt As String, xWord As String, org As String

  ed = GetWindowLong(hWnd, 0)               'get private data
  org = Buf                                 'store original line
  edmUCASE Buf                              'make proper UCASE for search
  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ' Loop through the line, using a pointer for better speed
  '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  Letter = StrPtr(Buf) : wFlag = 0
  For ii = 1 To Len(Buf) + 1
     Select Case @Letter 'The characters we need to inlude in a word
        Case  65 To 90, 35 To 38, 48 To 57, 63, 95
           If wFlag = 0 Then
              wFlag = -1 : wPos = ii
           End If

        Case 47 ' uncomment character -> //
           txt = Mid$(org, ii)  ' to make sure its //
           If Asc(txt, 2) = 47 Then 'MID$(txt, 2, 1) = "/"
              SetTextColor @ed.memDC, @ed.ColorSyntaxRem
              TextOut @ed.memDC, @ed.xPos + (ii - 1) * @ed.charWidth, yLine, ByVal StrPtr(txt), Len(txt)
              Exit Sub 'next line, please..
           End If

        Case 34 ' text within string quotes -> "text"  (fixed to handle quotes within quotes)
           wPos = InStr(ii + 1, Buf, $Dq) 'Find match
           If wPos Then
              txt = Mid$(org, ii, wPos - ii + 1)
              SetTextColor @ed.memDC, @ed.ColorSyntaxString
              TextOut @ed.memDC, @ed.xPos + (ii - 1) * @ed.charWidth, yLine, ByVal StrPtr(txt), Len(txt)
              wPos = (wPos - ii)
              ii = ii + wPos
              Letter = Letter + wPos
              wFlag = 0
           End If

        Case Else  'word is ready
           If wFlag Then
              xWord = Mid$(Buf, wPos, ii - wPos)   'Get word

              Ac = Max&(0, Asc(xWord) - 64)
              If Ac < 30 Then
                 q2 = @ed.rEndCount + Ac * 4
                 q = @ed.rStartCount + Ac * 4
                 Result = edmArrayScan(xWord, @ed.rData, @ed.rDataEndPos, @q, @q2)
              End If

              If Result Then
                 Result = 0
                 txt = Mid$(org, wPos, ii - wPos)
                 SetTextColor @ed.memDC, @ed.ColorSyntaxCode
                 TextOut @ed.memDC, @ed.xPos + (wPos - 1) * @ed.charWidth, yLine, ByVal StrPtr(txt), Len(txt)
              End If
              wFlag = 0
           End If

     End Select
     Incr Letter
  Next ii
End Sub

'
' Draw selection in editor
'
Sub ctmPaintSelection(ByVal hWnd As Dword)
  Local X1 As Long, X2 As Long, Y1 As Long, Y2 As Long, ii As Long, txt As String, rc As RECT
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)

  SetBkColor @ed.memDC, GetSysColor(%COLOR_HIGHLIGHT)
  SetTextColor @ed.memDC, GetSysColor(%COLOR_HIGHLIGHTTEXT)

  If @ed.SelStartY = @ed.SelEndY Then 'one line
     If @ed.SelStartX < @ed.SelEndX Then
        X1 = @ed.SelStartX : X2 = @ed.SelEndX
     Else
        X1 = @ed.SelEndX : X2 = @ed.SelStartX
     End If
     Y1 = @ed.SelStartY : Y2 = @ed.SelEndY
     txt = Mid$(edmTxtGetLine(hWnd, Y1), X1 + 1, X2 - X1)
     If @ed.ShowNonPrintChars Then
        txt = txt & Chr$(182)
        Replace " " With Chr$(183) In txt
     End If
     TextOut @ed.memDC, @ed.xPos + X1 * @ed.charWidth, (Y1 - @ed.TopLine) * @ed.charHeight, _
                   ByVal StrPtr(txt), Len(txt)

  Else
     If @ed.SelStartY < @ed.SelEndY Then 'several lines
        X1 = @ed.SelStartX : X2 = @ed.SelEndX
        Y1 = @ed.SelStartY
        Y2 = Min&(@ed.SelEndY, (@ed.TopLine + @ed.wnHeight))
     Else
        X1 = @ed.SelEndX : X2 = @ed.SelStartX
        Y1 = @ed.SelEndY
        Y2 = Min&(@ed.SelStartY, (@ed.TopLine + @ed.wnHeight))
     End If

     If Y1 >= @ed.TopLine And Y1 < (@ed.TopLine + @ed.wnHeight) Then
        txt = Mid$(edmTxtGetLine(hWnd, Y1), X1 + 1)
        If @ed.ShowNonPrintChars = 0 Then
           txt = txt & " "
        Else
           txt = txt & Chr$(182)
           Replace " " With Chr$(183) In txt
        End If
        TextOut @ed.memDC, @ed.xPos + X1 * @ed.charWidth, (Y1 - @ed.TopLine) * @ed.charHeight, ByVal StrPtr(txt), Len(txt)
     End If

     For ii = Y1 + 1 To Y2 - 1
        If ii >= @ed.TopLine And ii < (@ed.TopLine + @ed.wnHeight) Then
           txt = edmTxtGetLine(hWnd, ii)
           If @ed.ShowNonPrintChars = 0 Then
              txt = txt & " "
           Else
              txt = txt & Chr$(182)
              Replace " " With Chr$(183) In txt
           End If
           TextOut @ed.memDC, @ed.xPos, (ii - @ed.TopLine) * @ed.charHeight, ByVal StrPtr(txt), Len(txt)
        End If
     Next

     If Y2 >= @ed.TopLine And Y2 < (@ed.TopLine + @ed.wnHeight) Then
        txt = Left$(edmTxtGetLine(hWnd, Y2), X2)
        If @ed.ShowNonPrintChars Then
           txt = txt & Chr$(182)
           Replace " " With Chr$(183) In txt
        End If
        TextOut @ed.memDC, @ed.xPos, (Y2 - @ed.TopLine) * @ed.charHeight, ByVal StrPtr(txt), Len(txt)
     End If
  End If
End Sub

'
' MOUSE ACTIONS - %WM_LBUTTONDOWN message handler
'
Function ctmMouseLButtonDown(ByVal hWnd  As Dword,_
                            ByVal wParam As Long, ByVal lParam As Long) As Long
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)

  SetFocus hWnd : SetCapture hWnd
  @ed.xCaret = (LoWrd(lParam) - @ed.xPos) / @ed.charWidth     'horizontal position of cursor
  @ed.yCaret = HiWrd(lParam) \ @ed.charHeight + @ed.TopLine 'vertical position of cursor
  If @ed.yCaret > @ed.TopLine + @ed.wnHeight - 1 Then @ed.yCaret = @ed.TopLine + @ed.wnHeight - 1
  If @ed.yCaret > @ed.UBtxBuf Then @ed.yCaret = @ed.UBtxBuf
  If @ed.xCaret < 0 Then @ed.xCaret = 0
  If @ed.confineCaret Then
     If @ed.xCaret > edmTxtGetLineLen(@ed, @ed.yCaret) Then @ed.xCaret = edmTxtGetLineLen(@ed, @ed.yCaret)
  End If

  If (wParam And %MK_LBUTTON) Then
     If LoWrd(lParam) <  @ed.lMargin - 2 - (@ed.iHscrollPos * @ed.charWidth) Then 'clicked in margin
        If (wParam And %MK_SHIFT) Then
           If @ed.yCaret < @ed.SelStartY Then
              @ed.SelStartY = @ed.yCaret : @ed.SelStartX = 0
           Else
              @ed.SelEndX = edmTxtGetLineLen(@ed, @ed.yCaret) : @ed.xCaret = @ed.SelStartX
              @ed.SelEndY = @ed.yCaret
           End If
           @ed.SelLen = edmTxtGetSelLength(@ed)

        Else
           @ed.SelStartX = 0
           @ed.xCaret    = @ed.SelStartX
           @ed.SelStartY = @ed.yCaret
           If @ed.yCaret < @ed.UBtxBuf Then
              @ed.SelEndX   = 0
              Incr @ed.yCaret
           Else
              @ed.SelEndX = edmTxtGetLineLen(@ed, @ed.yCaret)
           End If
           @ed.SelEndY   = @ed.yCaret
           @ed.SelLen    = edmTxtGetSelLength(@ed)
        End If
        @ed.SelClickedY = @ed.SelStartY
        @ed.dClicked = 2
     Else
        If (wParam And %MK_SHIFT) Then
           @ed.SelEndX = @ed.xCaret : @ed.SelEndY = @ed.yCaret
           @ed.SelLen = edmTxtGetSelLength(@ed)
        Else
           SendMessage hWnd, %EDM_RESETSEL, 0, 0
        End If
        @ed.dClicked = 1
     End If
  End If

  SendMessage hWnd, %EDM_REFRESH, 0, 0

  Local nm As NMHDR
  nm.hwndFrom = hWnd
  nm.idFrom   = @ed.id
  nm.code     = (%NM_FIRST-2) '%NM_CLICK
  SendMessage @ed.hParent, %WM_NOTIFY, @ed.id , ByVal VarPtr(nm)

End Function

'
' %WM_LBUTTONUP message handler
'
Function ctmMouseLButtonUp(ByVal hWnd   As Dword,_
                           ByVal wParam As Long, ByVal lParam As Long) As Long
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)

  If @ed.confineCaret Then @ed.xCaretOld = @ed.xCaret
  @ed.SelLen = edmTxtGetSelLength(@ed)
  @ed.dClicked = 0
  ReleaseCapture
  If @ed.hTimer Then
     KillTimer hWnd, @ed.hTimer
     @ed.hTimer = 0
  End If

End Function

'
' %WM_LBUTTONDBLCLK message handler
'
Function ctmMouseLButtonDblClk(ByVal hWnd As Dword, _
                               ByVal wParam As Long, ByVal lParam As Long) As Long
  Local ed As EDM32DATA Ptr, nm As NMHDR
  ed = GetWindowLong(hWnd, 0)

  SetCapture hWnd
  @ed.xCaret = (LoWrd(lParam) - @ed.xPos) / @ed.charWidth     'horizontal position of cursor
  @ed.yCaret = HiWrd(lParam) \ @ed.charHeight + @ed.TopLine 'vertical position of cursor
  If @ed.yCaret < 0 Then @ed.yCaret = 0
  If @ed.yCaret > @ed.UBtxBuf Then @ed.yCaret = @ed.UBtxBuf
  If @ed.xCaret < 0 Then @ed.xCaret = 0
  If @ed.confineCaret Then
     If @ed.xCaret > edmTxtGetLineLen(@ed, @ed.yCaret) Then @ed.xCaret = edmTxtGetLineLen(@ed, @ed.yCaret)
  End If

  If (wParam And %MK_LBUTTON) Then
     If LoWrd(lParam) < @ed.lMargin - 2 - (@ed.iHscrollPos * @ed.charWidth) Then
        @ed.dClicked = 4
        edmSelectSub hWnd, 0, @ed
     Else
        @ed.dClicked = 3
        edmSelectWord hWnd, 0, @ed
     End If
  End If

  SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0
  @ed.dSelStartX  = @ed.SelStartX
  @ed.dSelEndX    = @ed.SelEndX
  @ed.dSelStartY  = @ed.SelStartY
  @ed.SelClickedY = @ed.SelStartY
  SendMessage hWnd, %EDM_REFRESH, 0, 0

  nm.hwndFrom = hWnd
  nm.idFrom   = @ed.id
  nm.code     = (%NM_FIRST-3) '%NM_DBLCLK
  SendMessage @ed.hParent, %WM_NOTIFY, @ed.id , ByVal VarPtr(nm)

End Function

'
' %WM_MOUSEMOVE message handler
'
Function ctmMouseMove(ByVal hWnd   As Dword,_
                      ByVal wParam As Long, ByVal lParam As Long) As Long

  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)
  If @ed.dClicked = 4 Then Exit Function '<- if double-clicked in left margin
  Local txCaret As Long, tyCaret As Long, pt As POINTAPI, wRect As RECT

  If @ed.dClicked <> 2 Then
     If LoWrd(lParam) < @ed.lMargin - 2 - (@ed.iHscrollPos * @ed.charWidth) Then
        If GetCursor() <> @ed.CursArrow Then
           SetCursor @ed.CursArrow
        End If
     ElseIf LoWrd(lParam) >= @ed.lMargin - 2 - (@ed.iHscrollPos * @ed.charWidth) Then
        If GetCursor() <> @ed.CursIBeam Then
           SetCursor @ed.CursIBeam
        End If
     End If
  End If

  If (wParam And %MK_LBUTTON) Then
     If @ed.dClicked = 0 Then Exit Function

     GetCursorPos pt : GetWindowRect hWnd, wRect

     If pt.y < wRect.nTop + 3 Then 'create new timer for auto scrolling if outside area
        If @ed.hTimer = 0 Then @ed.hTimer = SetTimer(hWnd, 22, 10, 0)
        'let it slip throughto provide "accelerated" scroll on mouse move above/below..
     ElseIf pt.y > wRect.nBottom - 2 - GetSystemMetrics(%SM_CYHSCROLL) Then
        If @ed.hTimer = 0 Then @ed.hTimer = SetTimer(hWnd, 23, 10, 0)
        'let it slip throughto provide "accelerated" scroll on mouse move above/below..
     Else
        If @ed.hTimer Then KillTimer hWnd, @ed.hTimer
        @ed.hTimer = 0
     End If

     If pt.y < wRect.nTop + 3 Then
        tyCaret = @ed.TopLine - ((wRect.nTop - pt.y) \ @ed.charHeight + 1)
     Else
        tyCaret = @ed.TopLine + HiWrd(lParam) \ @ed.charHeight
     End If

     If pt.x < wRect.nLeft  + 3 Then
        txCaret = 0
     Else
        txCaret = (LoWrd(lParam) - @ed.xPos) / @ed.charWidth
     End If

     If tyCaret < 0 Then tyCaret = 0
     If tyCaret > @ed.TopLine + @ed.wnHeight Then tyCaret = @ed.TopLine + @ed.wnHeight
     If tyCaret > @ed.UBtxBuf Then tyCaret = @ed.UBtxBuf
     If txCaret < 0 Then txCaret = 0
     If @ed.confineCaret Then
        If txCaret > edmTxtGetLineLen(@ed, tyCaret) Then txCaret = edmTxtGetLineLen(@ed, tyCaret)
     End If

     If LoWrd(lParam) <  @ed.lMargin - 2 - (@ed.iHscrollPos * @ed.charWidth) Then 'in margin
     Else
        If txCaret = @ed.xCaret And tyCaret = @ed.yCaret Then Exit Function
     End If

     If tyCaret > @ed.TopLine + @ed.wnHeight - 1 Then
        @ed.TopLine = Min&(tyCaret - @ed.wnHeight + 1, @ed.UBtxBuf - @ed.wnHeight + 1)
     ElseIf tyCaret < @ed.TopLine Then
        @ed.TopLine = Max&(0, tyCaret)
     End If

     @ed.xCaret = txCaret : @ed.yCaret = tyCaret
     If @ed.dClicked = 1 Then
        @ed.SelEndX = @ed.xCaret : @ed.SelEndY = @ed.yCaret
        @ed.SelLen = edmTxtGetSelLength(@ed)
        If @ed.confineCaret = 0 Then
            If @ed.SelStartX > edmTxtGetLineLen(@ed, @ed.SelStartY) Then
               @ed.SelStartX = edmTxtGetLineLen(@ed, @ed.SelStartY)
            End If
            If @ed.SelEndX > edmTxtGetLineLen(@ed, @ed.SelEndY) Then
               @ed.SelEndX = edmTxtGetLineLen(@ed, @ed.SelEndY)
            End If
        End If

     ElseIf @ed.dClicked = 2 Then 'clicked in marging
        @ed.SelEndY = @ed.yCaret

        If @ed.SelEndY < @ed.SelClickedY Then 'upwards..
           @ed.SelEndX   = 0
           @ed.SelStartX = 0
           @ed.xCaret    = 0
           @ed.SelEndY = @ed.yCaret
           @ed.SelStartY = Min&(@ed.SelClickedY + 1, @ed.UBtxBuf)

        Else
           @ed.SelStartX = 0
           @ed.xCaret    = 0
           @ed.SelStartY = @ed.SelClickedY
           If @ed.yCaret < @ed.UBtxBuf Then
              @ed.SelEndX   = 0
              Incr @ed.yCaret
              Incr @ed.SelEndY
           Else
              @ed.SelEndX = edmTxtGetLineLen(@ed, @ed.yCaret)
           End If
        End If

        @ed.SelLen = edmTxtGetSelLength(@ed)


     ElseIf @ed.dClicked = 3 Then                'doubleclicked on word
        If @ed.yCaret = @ed.dSelStartY Then      'on same line
           @ed.SelStartY = @ed.yCaret : @ed.SelEndY = @ed.yCaret
           If @ed.xCaret < @ed.dSelStartX Then                                    'move before word
              @ed.SelStartX = @ed.xCaret
              @ed.SelEndX = @ed.dSelEndX
           ElseIf @ed.xCaret >= @ed.dSelStartX And @ed.xCaret < @ed.dSelEndX Then 'move within word
              @ed.xCaret = @ed.dSelEndX
              @ed.SelStartX = @ed.dSelStartX
              @ed.SelEndX = @ed.dSelEndX
           ElseIf @ed.xCaret > @ed.dSelEndX Then                                  'move after word
              @ed.SelStartX = @ed.dSelStartX
              @ed.SelEndX = @ed.xCaret
           End If

        ElseIf @ed.yCaret < @ed.dSelStartY Then
           @ed.SelStartY = @ed.yCaret : @ed.SelEndY = @ed.dSelStartY
           @ed.SelStartX = @ed.xCaret : @ed.SelEndX = @ed.dSelEndX

        ElseIf @ed.yCaret > @ed.dSelStartY Then
           @ed.SelStartY = @ed.dSelStartY : @ed.SelEndY = @ed.yCaret
           @ed.SelStartX = @ed.dSelStartX : @ed.SelEndX = @ed.xCaret
        End If

        @ed.SelLen = edmTxtGetSelLength(@ed)
     End If

     SendMessage hWnd, %EDM_REFRESH, 0, 0
  End If
End Function

'
' %WM_MOUSEWHEEL message handler
'
Function ctmMouseWheel(ByVal hWnd   As Dword,_
                       ByVal wParam As Long, ByVal lParam As Long) As Long

  Local zDelta As Integer
  zDelta = HiWrd(wParam)

  If (LoWrd(wParam) And %MK_MBUTTON) = %MK_MBUTTON Then 'horizontal scroll
     If zDelta > 0 Then                                 'scroll to the left  (Arrow left)
        SendMessage hWnd, %WM_HSCROLL, MakLng(%SB_LINELEFT, 0), 0
     Else                                               'scroll to the right (Arrow right
        SendMessage hWnd, %WM_HSCROLL, MakLng(%SB_LINERIGHT, 0), 0
     End If
  Else                                                  'vertical scroll
     If zDelta > 0 Then                                 'scroll upwards   (Arrow up)
        SendMessage hWnd, %WM_VSCROLL, MakLng(%SB_LINEUP, 0), 0
     Else                                               'scroll downwards (Arrow down)
        SendMessage hWnd, %WM_VSCROLL, MakLng(%SB_LINEDOWN, 0), 0
     End If
  End If

End Function

'
' %WM_RBUTTONDOWN message handler
'
Function ctmMouseRButtonDown(ByVal hWnd   As Dword,_
                             ByVal wParam As Long, ByVal lParam As Long) As Long
  Local ed As EDM32DATA Ptr, nm As NMHDR

  ed          = GetWindowLong(hWnd, 0)
  nm.hwndFrom = hWnd
  nm.idFrom   = @ed.id
  nm.code     = (%NM_FIRST-5) '%NM_RCLICK
  SendMessage @ed.hParent, %WM_NOTIFY, @ed.id , ByVal VarPtr(nm)

End Function

'
' %WM_RBUTTONDBLCLK message handler
'
Function ctmMouseRButtonDblClk(ByVal hWnd   As Dword,_
                               ByVal wParam As Long, ByVal lParam As Long) As Long
  Local ed As EDM32DATA Ptr, nm As NMHDR

  ed          = GetWindowLong(hWnd, 0)
  nm.hwndFrom = hWnd
  nm.idFrom   = @ed.id
  nm.code     = (%NM_FIRST-6) '%NM_RDBLCLK
  SendMessage @ed.hParent, %WM_NOTIFY, @ed.id , ByVal VarPtr(nm)

End Function

'
' Select a sub or function
'
Sub edmSelectSub(ByVal hWnd As Dword, KeepSelect As Long, ed As EDM32DATA)
  Local ch As String, Bc As Long, sFlag As Long

  For Bc = ed.yCaret To 0 Step - 1
     ch = edmTxtGetLine(hWnd, Bc)

     If Len(ch) Then
        ch = LTrim$(edmUCASER(ch))
        If Bc < ed.yCaret And Left$(ch, 12) = "END FUNCTION" Or _
           Bc < ed.yCaret And Left$(ch, 7) = "END SUB" Then

           If ed.yCaret < ed.UBtxBuf Then
              ed.SelEndX   = 0
              Incr ed.yCaret
           Else
              ed.SelEndX = edmTxtGetLineLen(ed, ed.yCaret)
           End If
           ed.SelLen = edmTxtGetSelLength(ed)
           Exit Sub
        End If

        If Left$(ch, 4) = "SUB " Then
           sFlag = 1 : ed.SelStartY = Bc : ed.SelStartX = 0 : Exit For
        ElseIf Left$(ch, 9) = "FUNCTION " And InStr(ch, "=") = 0 Then
           sFlag = 2 : ed.SelStartY = Bc : ed.SelStartX = 0 : Exit For
        ElseIf Left$(ch, 18) = "CALLBACK FUNCTION " Or Left$(ch, 16) = "STATIC FUNCTION " Then
           sFlag = 2 : ed.SelStartY = Bc : ed.SelStartX = 0 : Exit For
        ElseIf Bc = 0 Then
           If ed.yCaret < ed.UBtxBuf Then
              ed.SelEndX = 0
              Incr ed.yCaret
           Else
              ed.SelEndX = edmTxtGetLineLen(ed, ed.yCaret)
           End If
           ed.SelLen = edmTxtGetSelLength(ed)
           Exit Sub

        End If
     End If
  Next

  For Bc = ed.yCaret To ed.UBtxBuf
     ch = edmTxtGetLine(hWnd, Bc)

     If Len(ch) Then
        If sFlag = 1 Then
           If Left$(LTrim$(edmUCASER(ch)), 7) = "END SUB" Then
              ed.SelEndY = Bc
              If Bc < ed.UBtxBuf Then
                 ed.SelEndX   = 0
                 Incr ed.SelEndY
              Else
                 ed.SelEndX = Len(ch)
              End If
              ed.xCaret = ed.SelStartX : ed.yCaret = ed.SelStartY : Exit For
           End If
        ElseIf sFlag = 2 Then
           If Left$(LTrim$(edmUCASER(ch)), 12) = "END FUNCTION" Then
              ed.SelEndY = Bc
              If Bc < ed.UBtxBuf Then
                 ed.SelEndX   = 0
                 Incr ed.SelEndY
              Else
                 ed.SelEndX = Len(ch)
              End If
              ed.xCaret = ed.SelStartX : ed.yCaret = ed.SelStartY : Exit For
           End If
        End If
     End If
  Next

  If sFlag > 0 Then
     SendMessage hWnd, %EDM_SETTOPLINE, Max&(0, ed.SelStartY - 3), 0
  End If
  ed.SelLen = edmTxtGetSelLength(ed)
End Sub

'
' Select a word
'
Sub edmSelectWord(ByVal hWnd As Dword, KeepSelect As Long, ed As EDM32DATA)
  Local ch As String, Ac As Long, Pos As Long, Letter As Byte Ptr
  ch = edmTxtGetLine(hWnd, ed.yCaret)

  Letter = StrPtr(ch) + ed.xCaret
  For Ac = ed.xCaret To 0 Step - 1 'pick out beginning of word
     If Ac = ed.xCaret Then        'if clicking right side of word, make sure it still may be selected
        If Ac > 0 And IsFalse(IsCharAlphaNumeric(@Letter)) Then
           Decr Ac : Decr Letter                           'by start checking from previous characters
        End If
     End If
     Select Case @Letter
        Case 32, 34, 40 To 44, 46, 47, 58 To 62, 91 To 94, 123 To 191, 215, 247, 0 To 31
           Exit For
     End Select
     Decr Letter
  Next

  Select Case Peek(StrPtr(ch) + Ac + 1) 'should not begin with - or '
     Case 39, 45 : Incr Ac
  End Select

  Pos = InStr(-1, Mid$(ch, Ac + 2, ed.xCaret - Ac), "--") 'single - is ok, but not multiple
  If Pos = 0 Then
     ed.SelStartX = Ac + 1
  Else
     ed.SelStartX = Ac + Pos + 2
  End If

  Letter = StrPtr(ch) + ed.xCaret
  For Ac = ed.xCaret To Len(ch)     'pick out end of word
     Select Case @Letter
        Case 32, 34, 40 To 44, 46, 47, 58 To 62, 91 To 94, 123 To 191, 215, 247, 0 To 31
           Exit For
     End Select
     Incr Letter
  Next

  Pos = InStr(Mid$(ch, ed.xCaret + 1, Ac - ed.xCaret), "--")
  If Pos = 0 Then
     ed.SelEndX = Ac
  Else
     ed.SelEndX = ed.xCaret + Pos - 1
  End If

  ed.xCaret = ed.SelEndX
  ed.SelLen = ed.SelLen + (ed.SelEndX - ed.SelStartX)

End Sub

'********************************************************************
' Set Scroll bars
'********************************************************************
Sub edmSetScrollData(ByVal hWnd As Dword, ByVal vhPage As Long, ByVal vhMax As Long, _
                     ByVal vhPos As Long, ByVal vhBar As Long)
  Local si As SCROLLINFO

  si.cbSize = SizeOf(si)
  si.fMask  = %SIF_ALL Or %SIF_DISABLENOSCROLL
  si.nMin   = 0
  si.nMax   = Max&(0, vhMax)
  si.nPage  = vhPage
  si.nPos   = vhPos
  SetScrollInfo hwnd, vhBar, si, 1

End Sub

'
' Combined Replace all / Count all function
'
Function edmTxtReplaceAll(ByVal hWnd As Dword, ByVal s As String, ByVal r As String, _
                          ByVal mc As Long, ByVal wo As Long, CountOnly As Long) As Long

  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)
  If Len(s) = 0 Or @ed.txtLen < Len(s) Then Exit Function
  s = edmTxtFindCRLF(s)

  Local x As Long, x2 As Long, fnd As Long
  Local Letter As Byte Ptr, Letter2 As Byte Ptr, ckLet As Byte Ptr, OldLetter As Byte Ptr
  Local oldx As Long, Flag As Long, sln As Long, tmpCountOnly As Long, dif As Long, rln As Long, oln As Long
  Local fLet As Byte, txt As String, txt2 As String

  txt = edmTxtGetSelText(hWnd, 0)                          'get all text
  If mc = 0 Then                                           'if not MATCH CASE
     edmUCASE txt : edmUCASE s                             'compare both search and text as UCASE strings
  End If

  oln = Len(txt)                                           'get whole text's length
  sln = Len(s)                                             'get search text's length
  tmpCountOnly = 1                                         'set tmp flag so we can get count first
  GoTo BeginSearch                                         'ignore replace at this point

BeginReplace:
  r = edmTxtFindCRLF(r)
  tmpCountOnly = 0                                         'now we have count, so reset this one
  rln = Len(r)                                             'get replace text's length
  dif = fnd * (rln - sln)                                  'calculate occurrences and caclculate needed space
  txt2 = Space$(oln + dif)                                 'allocate space enough for replacements
  Letter2 = StrPtr(txt2)                                   'set pointer to resulting text

BeginSearch:
  Letter  = StrPtr(txt)                                    'set pointer to text
  fLet    = Asc(s)                                         'get first letter in search text

  For  x = 1 To oln
     If @Letter = fLet Then                                'if first letter matches
        OldLetter = Letter : oldx = x : flag = 1           'store positions and set found flag

        If wo Then                                         'if whole words only
           If x > 1 Then                                   'if position is > 1
              ckLet = Letter - 1                           'check letter before
              flag  = (IsCharAlphaNumeric(@ckLet) = 0)     'flag becomes zero if not ok
              If flag = 0 Then                             'if not ok
                 If InStr("_-", Chr$(@ckLet)) Then flag = 1 'these are allowed
              End If
           End If
           If Flag Then                                    'if still ok
              If x < oln - sln Then                        'if position is < len(txt) - len(search)
                 ckLet = Letter + sln                      'check letter after
                 flag  = (IsCharAlphaNumeric(@ckLet) = 0)  'flag becomes zero if not ok
                 If flag = 0 Then                          'if not ok
                    If InStr("_-", Chr$(@ckLet)) Then flag = 1 'these are allowed
                 End If
              End If
           End If
        End If

        If Flag Then                                       'if still ok
           For x2 = 2 To sln                               'check against rest of search string
              Incr Letter : Incr x                         'increase pointers
              If @Letter <> Asc(s, x2) Then                'if anything different
                 Letter = OldLetter : x = oldx             'reset position
                 flag = 0 : Exit For                       'reset flag and move on
              End If
           Next
        End If
     End If

     If tmpCountOnly Then                  'if initial count
        If flag Then                       'match, so increase counter and reset flag
           Incr fnd : flag = 0
        End If
     Else                                  'else, if replace action
        If flag = 0 Then                   'if no match
           ckLet = @ed.txt + x - 1         'set pointer to original string
           @Letter2 = @ckLet               'just copy bytes from original string
        Else                               'else perfect match, so
           Poke$ Letter2, r                'poke replace string into place
           Letter2 = Letter2 + rln - 1     'increase pointers in resulting string
           flag = 0                        'and reset flag
        End If
        Incr Letter2
     End If

     Incr Letter
  Next

  If CountOnly = 0 Then
     If @ed.ReadOnly = 1 Then Exit Function
     If fnd And tmpCountOnly Then GoTo BeginReplace
  End If

  If fnd And CountOnly = 0 And Len(txt2) Then
     x = @ed.xCaret : x2 = @ed.yCaret : oldx = @ed.TopLine

     SendMessage hWnd, %EDM_DELETE, 1, 0
     edmTxtPutString @ed, txt2, 0, 0, -1, -1, 23

     If @ed.undoPos Then
        @ed.@undo[@ed.undoPos - 1].uActionCount = 1
     End If

     @ed.xCaret = x : @ed.yCaret = x2 : @ed.TopLine = oldx
     SendMessage hWnd, %EDM_CHECKBOUNDS, 0, 0
     SendMessage hWnd, %EDM_REFRESH, 0, -1
  End If

  Function = fnd
End Function

'
' Find next function
'
Function edmTxtFindCRLF(ByVal s As String) As String
  Local Pos As Long

  Pos = InStr(edmUCASER(s), "^P")                          'Enable search for line feeds (^p)
  If Pos And Len(s) = 2 Then
     s = $CrLf
  ElseIf Pos And Len(s) > 2 Then
     Do
        If Pos = 1 Then
           s = $CrLf & Mid$(s, Pos + 2)
        ElseIf Pos > 1 And Mid$(s, Pos - 1, 1) <> "^" Then
           s = Left$(s, Pos - 1) & $CrLf & Mid$(s, Pos + 2)
        ElseIf Pos > 1 And Mid$(s, Pos - 1, 1) = "^" Then
           s = Left$(s, Pos - 1) & Mid$(s, Pos + 1)
        End If
        Pos = InStr(Pos + 2, edmUCASER(s), "^P")
     Loop Until Pos = 0
  End If

  Function = s
End Function

'
' Find next function
'
Function edmTxtFindNext(ByVal hWnd As Dword, ByVal s As String, ByVal dr As Long, _
                        ByVal mc As Long, ByVal wo As Long, ByVal eRedraw As Long) As Long

  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)

  If Len(s) = 0 Or @ed.txtLen < Len(s) Then Exit Function
  Local Pos As Long, fnd As Long, x As Long, y As Long, sln As Long, txt As String

  s = edmTxtFindCRLF(s)
  Pos = edmTxtGetPos(hWnd, @ed.xCaret, @ed.yCaret) + 1        'get current position
  txt = edmTxtGetSelText(hWnd, 0)                             'get text
  sln = Len(s)

  If mc = 0 Then                                              'MATCH CASE
     edmUCASE txt : edmUCASE s                                'ignore case
  End If

StartHere:
     If dr = 0 Then                                           'DIRECTION
        If Pos <= Len(txt) - sln + 1 Then
           fnd = InStr(Pos, txt, s)                           'search forward
        Else
           fnd = 0
        End If
     Else
        fnd = InStr(-1, Left$(txt, Pos - 1 - @ed.SelLen), s)  'search backwards
     End If

  If fnd Then
     If wo Then                                            'WHOLE WORDS ONLY
        If fnd > 1 Then
           x = Asc(txt, fnd - 1)
           If IsCharAlphaNumeric(x) Then 'OR INSTR("_-", CHR$(x)) = 0
              If dr Then 'search upwards
                 Pos = fnd
              Else
                 Pos = fnd  + sln
              End If
              GoTo StartHere
           End If
        End If
        If fnd + Len(s) < Len(txt) Then
           x = Asc(txt, fnd + sln)
           If IsCharAlphaNumeric(x) Then 'OR INSTR("_-", CHR$(x)) = 0
              If dr Then 'search upwards
                 Pos = fnd
              Else
                 Pos = fnd + sln
              End If
              GoTo StartHere
           End If
        End If
     End If

     y = Tally(Left$(txt, fnd), $CrLf)
     If y Then
        x = fnd - InStr(-1, Left$(txt, fnd), $CrLf) - 2
     Else
        x = fnd - 1
     End If
     @ed.SelStartX = x : @ed.SelEndX = x + Len(s)

     x = InStr(-1, s, $CrLf)
     If x Then @ed.SelEndX = sln - x - 1

     @ed.SelStartY = y : @ed.SelEndY = y
     y = Tally(s, $CrLf)
     If y Then @ed.SelEndY = @ed.SelEndY + y

     @ed.SelLen = sln
     @ed.xCaret = @ed.SelEndX : @ed.yCaret = @ed.SelEndY
     If eRedraw Then
        @ed.TopLine = Min&(Max&(0, @ed.SelStartY - 1), Max&(0, @ed.UBtxBuf - @ed.wnHeight + 1))
        SendMessage @ed.hWnd, %EDM_REFRESH, 0, 0
     End If
  End If

  Function = fnd
End Function

'
' Check boundaries
'
Sub edmCheckXYbounds(ed As EDM32DATA)
  If ed.yCaret < 0 Then ed.yCaret = 0
  If ed.xCaret < 0 Then ed.xCaret = 0
  If ed.yCaret > ed.UBtxBuf Then ed.yCaret = ed.UBtxBuf
  If ed.confineCaret Then
     If ed.xCaret > edmTxtGetLineLen(ed, ed.yCaret) Then ed.xCaret = edmTxtGetLineLen(ed, ed.yCaret)
  End If

  If ed.TopLine > ed.UBtxBuf - ed.wnHeight + 1 Then
     ed.TopLine = Max&(0, ed.UBtxBuf - ed.wnHeight + 1)
  End If

  If ed.yCaret < ed.TopLine Then
     SendMessage ed.hWnd, %EDM_SETTOPLINE, Max&(0, ed.yCaret), 0
  End If

End Sub

'
' Get text from ClipBoard
'
Function edmClipboardGetText () As String
  Local Bytes As Long, cbPtr As Byte Ptr, zPtr As Asciiz Ptr

  If OpenClipboard(0) Then               ' try to claim clipboard for ourselves
     cbPtr = GetClipboardData(%CF_TEXT)  ' see if there is any text data available
     If cbPtr Then
        zPtr = GlobalLock(cbPtr)         ' lock the memory address/handle and get pointer to text
        If zPtr Then Function = @zPtr    ' if success, zPtr points to the text
        GlobalUnlock cbPtr               ' decrement the lock count
     End If
     CloseClipboard                      ' release clipboard
  End If

End Function

'
' Put text into ClipBoard
'
Function edmClipboardPutText (ByVal txt As String) As Long
  Local cbPtr As Long, hData As Long, Bytes As Long

  txt = txt + Chr$(0)                    ' make it an ASCIIz string
  Bytes = Len(txt)                       ' get # of bytes to load
  If Bytes < 2 Then Exit Function        ' if no bytes to load

  hData = GlobalAlloc(&H2002, Bytes)     ' get memory address of clipboard
  cbPtr = GlobalLock(hData)              ' lock and get pointer to data
  Poke$ cbPtr, txt                       ' paste the text into memory
  GlobalUnlock hData                     ' unlock memory

  If IsFalse OpenClipboard(0) Then       ' if clipboard isn't available
    GlobalFree hData                     ' free up memory
    Exit Function                        ' Return false
  End If

  EmptyClipboard                         ' empty whatever's in there now
  SetClipboardData %CF_TEXT, ByVal hData ' point the clipboard at it as text
  CloseClipboard                         ' release clipboard

  Function = %TRUE                       ' return true
End Function

'
' %EDM_SORTLINES and %EDM_COMMENT procedure for comment in/out, in/outdent and sorting lines
'
Function edmFullLineProc(ByVal hWnd As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
  Local ii As Long, jj As Long, tmpCaretX As Long, tmpCaretY As Long, Pos As Long, smallStep As Long, _
        tmpSelStartX As Long, tmpSelEndX As Long, tmpSelStartY As Long, tmpSelEndY As Long, _
        ed As EDM32DATA Ptr, txt As String, buf As String, buf2 As String, rmchar As String
  ed = GetWindowLong(hWnd, 0)
  If @ed.ReadOnly = 1 Or @ed.txtLen = 0 Then Exit Function

  tmpCaretX = @ed.xCaret
  tmpCaretY = @ed.yCaret
  If @ed.SelStartY = @ed.SelEndY Then
     tmpSelStartX = Min&(@ed.SelStartX, @ed.SelEndX)
     tmpSelEndX   = Max&(@ed.SelStartX, @ed.SelEndX)
     tmpSelStartY = @ed.SelStartY
     tmpSelEndY   = tmpSelStartY
  ElseIf @ed.SelStartY < @ed.SelEndY Then
     tmpSelStartX = @ed.SelStartX
     tmpSelEndX   = @ed.SelEndX
     tmpSelStartY = @ed.SelStartY
     tmpSelEndY   = @ed.SelEndY
  Else
     tmpSelStartX = @ed.SelEndX
     tmpSelEndX   = @ed.SelStartX
     tmpSelStartY = @ed.SelEndY
     tmpSelEndY   = @ed.SelStartY
  End If

  @ed.SelStartX = 0
  @ed.SelEndX   = edmTxtGetLineLen(@ed, tmpSelEndY)
  @ed.SelStartY = tmpSelStartY
  @ed.SelEndY   = tmpSelEndY
  @ed.SelLen    = edmTxtGetSelLength(@ed)
  @ed.LineUpdate = 0

  If @ed.SelLen Then
     Select Case lParam
        Case 0 'REM OUT SELECTED LINE(s)
           rmchar = "'"
           'IF @ed.FileType = 4 THEN rmchar = "//"
           txt = Space$(@ed.SelLen + Max&(1, (tmpSelEndY - tmpSelStartY)) * 3) 'lines * (added char + CRLF)
           jj = 1
           For ii = tmpSelStartY To tmpSelEndY
              Buf = edmTxtGetLine(hWnd, ii)
              If Len(Buf) Then
                 Mid$(txt, jj) = rmchar & Buf & $CrLf
                 jj = jj + Len(Buf) + 3
              Else
                 Mid$(txt, jj) = Buf & $CrLf
                 jj = jj + Len(Buf) + 2
              End If
           Next
           txt = Left$(txt, jj - 3) 'remove last line feed, plus one
           tmpCaretX = tmpCaretX + 1
           tmpSelStartX = tmpSelStartX + 1
           tmpSelEndX = tmpSelEndX + 1
           Pos = 16

        Case 1 'REMOVE REM OUT FROM SELECTED LINE(s)
           rmchar = "'"
           'IF @ed.FileType = 4 THEN rmchar = "//"
           txt = Space$(@ed.SelLen + Max&(1, (tmpSelEndY - tmpSelStartY)) * 2) 'minimum add (lines * CRLF)
           jj = 1
           For ii = tmpSelStartY To tmpSelEndY
              Buf = edmTxtGetLine(hWnd, ii)
              If Len(Buf) And Left$(LTrim$(Buf), 1) = rmchar Then
                 Pos = InStr(Buf, rmchar)
                 If Pos > 1 Then
                    Buf2 = Left$(Buf, Pos - 1) & Mid$(Buf, Pos + 1)
                 Else
                    Buf2 = Mid$(Buf, 2)
                 End If

                 Mid$(txt, jj) = Buf2 & $CrLf
                 jj = jj + Len(Buf2) + 2
              Else
                 Mid$(txt, jj) = Buf & $CrLf
                 jj = jj + Len(Buf) + 2
              End If
           Next
           txt = Left$(txt, jj - 3) 'remove last line feed, plus one
           tmpCaretX = Max&(0, tmpCaretX - 1)
           tmpSelStartX = Max&(0, tmpSelStartX - 1)
           tmpSelEndX = Max&(0, tmpSelEndX - 1)
           Pos = 17

        Case 2 'INDENT SELECTED LINE(s)
           txt = Space$(@ed.SelLen + Max&(1, (tmpSelEndY - tmpSelStartY + 1)) * (@ed.tabSize + 2)) 'lines * (@ed.tabSize + LEN($CRLF))
           jj = 1
           For ii = tmpSelStartY To tmpSelEndY
              Buf = edmTxtGetLine(hWnd, ii)
              If Len(Buf) Then
                 Mid$(txt, jj) = Space$(@ed.tabSize) & Buf & $CrLf
                 jj = jj + @ed.tabSize + Len(Buf) + 2
              Else
                 Mid$(txt, jj) = Buf & $CrLf
                 jj = jj + Len(Buf) + 2
              End If
           Next
           txt = Left$(txt, jj - 3) 'remove last line feed, plus one
           tmpCaretX = tmpCaretX + @ed.tabSize
           tmpSelStartX = tmpSelStartX + @ed.tabSize
           tmpSelEndX = tmpSelEndX + @ed.tabSize
           Pos = 14

        Case 3  'OUTDENT SELECTED LINE(s)
           For ii = tmpSelStartY To tmpSelEndY
              Buf = edmTxtGetLine(hWnd, ii)
              If Len(Buf) Then  'must be at least tabSize spaces on each line to remove
                 If Left$(Buf, @ed.tabSize) <> Space$(@ed.tabSize) Then
                    If Asc(Buf) <> 32 Then Exit Function
                    smallStep = 1
                 End If
              End If
           Next

           txt = Space$(@ed.SelLen) 'minimum space
           jj = 1
           For ii = tmpSelStartY To tmpSelEndY
              Buf = edmTxtGetLine(hWnd, ii)
              If Len(Buf) Then
                 If smallStep = 0 Then 'move tabSize steps to the left
                    Buf = Mid$(Buf, @ed.tabSize + 1)
                 Else
                    Buf = Mid$(Buf, 2) 'move one step to the left
                 End If
                 Mid$(txt, jj) = Buf & $CrLf
                 jj = jj + Len(Buf) + 2
              Else
                 Mid$(txt, jj) = $CrLf
                 jj = jj + 2
              End If
           Next
           txt = Left$(txt, jj - 3) 'remove last line feed, plus one
           tmpCaretX = Max&(0, tmpCaretX - 1)
           tmpSelStartX = Max&(0, tmpSelStartX - 1)
           tmpSelEndX = Max&(0, tmpSelEndX - 1)
           Pos = 15

        Case 4  'SORT SELECTED LINE(s)
           If tmpSelEndY - tmpSelStartY Then
              ReDim arr(tmpSelEndY - tmpSelStartY) As String
              For ii = tmpSelStartY To tmpSelEndY
                 arr(ii - tmpSelStartY) = edmTxtGetLine(hWnd, ii)
              Next

              If wParam = 0 Then
                 Pos = 18
                 Array Sort Arr()
              ElseIf wParam = 1 Then
                 Pos = 19
                 Array Sort Arr(), Collate UCase
              ElseIf wParam = 2 Then
                 Pos = 20
                 Array Sort Arr(), Descend
              ElseIf wParam = 3 Then
                 Pos = 21
                 Array Sort Arr(), Collate UCase, Descend
              End If

              jj = 1
              txt = Space$(@ed.SelLen) 'minimum space
              For ii = tmpSelStartY To tmpSelEndY
                 Buf = arr(ii - tmpSelStartY) & $CrLf
                 Mid$(txt, jj) = Buf
                 jj = jj + Len(Buf)
              Next
              txt = Left$(txt, jj - 3) 'remove last line feed, plus one
           Else
              GoTo ResetProc
           End If

     End Select

     SendMessage hWnd, %EDM_DELETE, 0, 0
     edmTxtPutString @ed, txt, @ed.xCaret, @ed.yCaret, -1, -1, Pos
     If @ed.undoPos Then @ed.@undo[@ed.undoPos - 1].uActionCount = 1
     Function = 1 'return 1 on success, in case we ever need to check..
  End If

ResetProc:
  @ed.xCaret = tmpCaretX
  @ed.yCaret = tmpCaretY
  @ed.SelStartX = tmpSelStartX
  @ed.SelEndX   = tmpSelEndX
  @ed.SelStartY = tmpSelStartY
  @ed.SelEndY   = tmpSelEndY
  @ed.SelLen    = edmTxtGetSelLength(@ed)
  SendMessage hWnd, %EDM_REFRESH, 0, -1
End Function

'
' Get name of loaded file
'
Function edmGetFileName(ByVal hWnd As Dword) As String
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)
  Function = @ed.fName
End Function

'
' Set name of loaded file
'
Sub edmSetFileName(ByVal hWnd As Dword, ByVal fName As String)
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)
  @ed.fName = fName
End Sub

'
' edmLCASE routine, alters given string directly - very fast..
'
Sub edmLCASE(a$)
  If Len(a$) = 0 Then Exit Sub
  #Register None

  ! mov eax, a$            ; eax = pointer to string handle
  ! mov eax, [eax]         ; eax = pointer to string data
  ! mov ecx, [eax-4]       ; move length of string into ecx

  bgnLoLoop:
     ! mov dl, [eax]       ; move current char into dl
     ! cmp dl, 65          ; compare against value 65 (A)
     ! jb NoUpChar         ; if dl < 65  then get next character
     ! cmp dl, 91          ; compare against value 91
     ! jb MakeLower        ; if dl < 91 it is in 65-90 range, make lowercase and get next
     ! cmp dl, 192         ; compare against value 192 () - extended ANSI
     ! jb NoUpChar         ; if dl < 192 it is in 91-191 range, do nothing to it
     ! cmp dl, 215         ; compare against value 192 ()
     ! jb MakeLower        ; if dl < 215 it is in 192-214 range, make lowercase and get next
     ! je NoUpChar         ; if dl = 215 (), do nothing
     ! cmp dl, 223         ; compare against value 223 ( - has no lower case equivilant)
     ! jb MakeLower        ; if dl < 223 it is in 216-222 range, make lowercase and get next
     ! jmp NoUpChar        ; else - next character

  MakeLower:
     ! add dl, 32          ; make lowercase by adding 32 to dl's value
     ! mov [eax], dl       ; write changed char back into a$

  NoUpChar:
     ! inc eax             ; get next character
     ! dec ecx             ; decrease ecx (length) counter
     ! jnz bgnLoLoop       ; exit if 0, else iterate

End Sub

'
' Get type of character set - ansi, symbol.. a must for some fonts.
'
Function edmEnumCharSet (elf As ENUMLOGFONT, ntm As NEWTEXTMETRIC, _
                         ByVal FontType As Long, CharSet As Long) As Long
  CharSet = elf.elfLogFont.lfCharSet
End Function

'
' Create a desirable font and return its handle. Original code by Dave Navarro
'
Function edmMakeFont(ByVal FontName As String, ByVal PointSize As Long, ByVal fBold As Long, _
                     ByVal fItalic As Long, ByVal fUnderline As Long) As Dword
  Local hDC As Dword, CharSet As Long, CyPixels As Long

  hDC = GetDC(%HWND_DESKTOP)
    CyPixels  = GetDeviceCaps(hDC, %LOGPIXELSY)
    EnumFontFamilies hDC, ByVal StrPtr(FontName), CodePtr(edmEnumCharSet), ByVal VarPtr(CharSet)
  ReleaseDC %HWND_DESKTOP, hDC
  PointSize = 0 - (PointSize * CyPixels) \ 72

  Function = CreateFont(PointSize, 0, _  'height, width(default=0)
             0, 0, _                     'escapement(angle), orientation
             fBold, _                    'weight (%FW_DONTCARE = 0, %FW_NORMAL = 400, %FW_BOLD = 700)
             fItalic, _                  'Italic
             fUnderline, _               'Underline
             %FALSE, _                   'StrikeThru
             CharSet, %OUT_TT_PRECIS, _
             %CLIP_DEFAULT_PRECIS, %DEFAULT_QUALITY, _
             %FF_DONTCARE , ByCopy FontName)
End Function

'
' Set editor's font
'
Function edmSetFontEditor(ByVal hWnd As Dword, ByVal FontName As String, ByVal PointSize As Long, _
                          ByVal fBold As Long, ByVal fItalic As Long, ByVal fUnderline As Long) As Dword

  If hWnd And Len(FontName) Then
     Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)
     If @ed.hFont Then DeleteObject @ed.hFont
     @ed.hFont = edmMakeFont(FontName, PointSize, fBold, fItalic, fUnderline)
     SendMessage hWnd, %WM_SETFONT, @ed.hFont, 0   'set new font in the editor
  End If

  Function = @ed.hFont  'in case calling program needs it for something else
End Function

'
' edmMCASE routine, alters given string directly - very fast..
'
Sub edmMCASE(a$)
  If Len(a$) = 0 Then Exit Sub
  #Register None
  Local wordFlag As Byte

  ! mov eax, a$            ; eax = pointer to string handle
  ! mov eax, [eax]         ; eax = pointer to string data
  ! mov ecx, [eax-4]       ; length of string

  bgnMLoop:
     ! mov dl, [eax]       ; move current char into edx
     ! cmp dl, 39
     ! je mWordCheck        ; if dl = 39 => ASC(39)
     ! cmp dl, 45
     ! je mWordCheck       ; if dl = 45 => ASC(45)
     ! cmp dl, 65          ; CASE 65 TO 90 => ASC(65 to 90)
     ! jb mNextChar        ;    if dl < 65
     ! cmp dl, 91
     ! jb mUWordCheck      ;    if dl < 91
     ! cmp dl, 97          ; CASE 96 TO 122
     ! jb mNextChar        ;    if dl < 97
     ! cmp dl, 123
     ! jb mLWordCheck      ;    if dl < 123
     ! cmp dl, 146
     ! je mWordCheck       ; if dl = 146
     ! cmp dl, 180
     ! je mWordCheck       ; if dl = 180
     ! cmp dl, 192         ; CASE 192 TO 214, 216 TO 222  - extended UCASE ANSI characters
     ! jb mNextChar        ;    if dl < 192
     ! cmp dl, 215
     ! jb mUWordCheck      ;    if dl < 215
     ! je mNextChar        ;    if dl = 215
     ! cmp dl, 223
     ! jb mUWordCheck      ;    if dl < 223
     ! cmp dl, 224         ; CASE 224 TO 246, 248 TO 254  - extended LCASE ANSI characters
     ! jb mNextChar        ;    if dl < 224
     ! cmp dl, 247
     ! jb mLWordCheck      ;    if dl < 247
     ! je mNextChar        ;    if dl = 247
     ! cmp dl, 255
     ! jbe mLWordCheck     ;    if dl <= 255

  mNextChar:
     ! cmp wordFlag, 1     ; check if word flag is 1
     ! je mSetFlag0        ; if it is, jump ahead to reset it
     ! jmp GetmNextChar    ; else get next char

  mWordCheck:
     ! cmp wordFlag, 0     ; check if word flag is 0
     ! je mSetFlag1        ; if it is, jump ahead to set it
     ! jmp GetmNextChar    ; else get next char

  mUWordCheck:
     ! add dl, 32          ; make lowercase
     ! mov [eax], dl       ; write changed char back into a$
     ! cmp wordFlag, 0     ; check if word flag is 0
     ! je mSetFlag1        ; if it is, jump ahead to set it
     ! jmp GetmNextChar    ; else get next char

  mLWordCheck:
     ! cmp wordFlag, 0     ; check if word flag is 0
     ! je mSetFlag1        ; if it is, jump ahead to set it
     ! jmp GetmNextChar    ; else get next char

  mSetFlag0:
     ! dec wordFlag        ; reset word flag
     ! jmp GetmNextChar    ; get next char

  mSetFlag1:
     ! cmp dl, 39          ; if dl = 39 then
     ! je mNextChar        ; jump to next char
     ! cmp dl, 45          ; if dl = 45 then
     ! je mNextChar        ; jump to next char
     ! cmp dl, 146         ; if dl = 146 then
     ! je mNextChar        ; jump to next char
     ! cmp dl, 180         ; if dl = 180 then
     ! je mNextChar        ; jump to next char
     ! sub dl, 32          ; else make uppercase
     ! mov [eax], dl       ; write changed char back into a$
     ! inc wordFlag        ; set word flag
     ! jmp GetmNextChar    ; get next char

  GetmNextChar:
     ! inc eax             ; get next character
     ! dec ecx             ; decrease the ecx (length) counter
     ! jnz bgnMLoop        ; iterate if not zero

End Sub

'
' edmUCASE routine, alters given string directly - very fast..
'
Sub edmUCASE(a$)
  If Len(a$) = 0 Then Exit Sub
  #Register None

  ! mov eax, a$            ; eax = pointer to string handle
  ! mov eax, [eax]         ; eax = pointer to string data
  ! mov ecx, [eax-4]       ; move length of string into ecx

  BgnUpLoop:
     ! mov dl, [eax]       ; move current char into dl
     ! cmp dl, 97          ; compare against value 97 (a)
     ! jb NoDnChar         ; if dl < 97 it is in 0-97 range, do nothing - get next
     ! cmp dl, 123         ; compare against value 123
     ! jb MakeUpper        ; if dl < 123 it is in 97-122 range, make Upper case and get next
     ! cmp dl, 224         ; compare against value 224 () - extended ANSI characters
     ! jb NoDnChar         ; if dl < 224 it is in 123-223 range, do nothing - get next
     ! cmp dl, 247         ; compare against value 247 ()
     ! jb MakeUpper        ; if dl < 247 it is in 224-246 range, make Upper case and get next
     ! je NoDnChar         ; else dl = 247 (), do nothing - get next
     ! cmp dl, 255         ; compare against value 255 ()
     ! jb MakeUpper        ; if dl < 255 it is in 248-254 range, make Upper case and get next
     ! jmp NoDnChar        ; else dl = 255, do nothing - get next

  MakeUpper:
     ! sub dl, 32          ; make uppercase subtracting 32 from dl's value
     ! mov [eax], dl       ; write changed char back into a$

  NoDnChar:
     ! inc eax             ; get next character
     ! dec ecx             ; decrease ecx (length) counter
     ! jnz bgnUpLoop       ; iterate if not zero

End Sub

'
' edmUCASER function, returns UCASE string without altering original
'
Function edmUCASER(ByVal st As String) As String
  #Register None
  Local p As String Ptr
  p = StrPtr(st)

  ! mov eax, p              ; move pointer to string into eax
  ! mov ecx, [eax-4]        ; move length of string into ecx (counter)
  ! cmp ecx, 0              ; compare length with 0
  ! je exitedmUCASER           ; if it equals 0 (len = 0), then exit

  beginedmUCASER:
     ! mov dl, [eax]        ; move current char into dl
     ! cmp dl, 97           ; compare against value 97 (a)
     ! jb nextedmUCASER     ; if dl < 97  then get next character
     ! cmp dl, 123          ; compare against value 123
     ! jb makeedmUCASER     ; if dl < 123 it is in 97-122 range, make Uppercase and get next
     ! cmp dl, 224          ; compare against value 224 () - extended ANSI
     ! jb nextedmUCASER     ; if dl < 224 it is in 123-224 range, do nothing to it
     ! cmp dl, 247          ; compare against value 247
     ! jb makeedmUCASER     ; if dl < 247 it is in 224-247 range, make Uppercase and get next
     ! je nextedmUCASER     ; if dl = 247, do nothing
     ! cmp dl, 255          ; compare against value 255
     ! jb makeedmUCASER     ; if dl < 255 it is in 248-255 range, make Uppercase and get next
     ! jmp nextedmUCASER    ; else, on to next character

  makeedmUCASER:
     ! sub dl, 32           ; make lowercase by subtracting 32 from dl (which holds asc value)
     ! mov [eax], dl        ; move changed char back into eax and fall through to nextedmUCASER

  nextedmUCASER:
     ! inc eax              ; get next character
     ! dec ecx              ; decrease ecx (length) counter
     ! jnz beginedmUCASER   ; iterate if not zero (end of string)
     Function = st          ' ecx was 0, return result on exit

  exitedmUCASER:
End Function

'
' edmTCASE routine - toggle UPPER to lower and lower to UPPER..
'
Sub edmTCASE(a$)
  If Len(a$) = 0 Then Exit Sub
  #Register None

  ! mov eax, a$            ; eax = pointer to string handle
  ! mov eax, [eax]         ; eax = pointer to string data
  ! mov ecx, [eax-4]       ; length of string

  BgnMixLoop:
     ! mov edx, [eax]      ; move current char into edx
     ! cmp dl, 65          ; CASE 65 TO 90
     ! jb NoMixChar        ;    if dl < 65  then get next character
     ! cmp dl, 90
     ! jbe MixToLower      ;    if dl <= 90 then make lowercase and get next
     ! cmp dl, 97          ; CASE 97 TO 122
     ! jb NoMixChar        ;    if dl < 97  then next character
     ! cmp dl, 123
     ! jb MixToUpper       ;    if dl < 123 then make upper case and get next
     ! cmp dl, 192         ; CASE 192 TO 214, 216 TO 222  - extended ANSI characters
     ! jb NoMixChar        ;    if dl < 192
     ! cmp dl, 215
     ! jb MixToLower       ;    if dl < 215
     ! je NoMixChar        ;    if dl = 215
     ! cmp dl, 223
     ! jb MixToLower       ;    if dl <= 223
     ! cmp dl, 224         ; CASE 224 TO 246, 248 TO 254   - extended ANSI characters
     ! jb NoMixChar        ;    if dl < 224
     ! cmp dl, 247
     ! jb MixToUpper       ;    if dl < 247
     ! je NoMixChar        ;    if dl = 247
     ! cmp dl, 255
     ! jb MixToUpper       ;    if dl < 255
     ! jmp NoMixChar       ; CASE ELSE - get next character

  MixToUpper:
     ! sub dl, 32          ; make uppercase
     ! mov [eax], dl       ; write changed char back into a$
     ! jmp NoMixChar

  MixToLower:
     ! add dl, 32          ; make lowercase
     ! mov [eax], dl       ; write changed char back into a$

  NoMixChar:
     ! inc eax             ; get next character
     ! dec ecx              ; decrease the ecx (length) counter
     ! jnz bgnMixLoop

End Sub

'
' Asm wordcount - much faster and more correct than MS Word's crappy one..
'(MS Word counts separated dots as words, among many other strange things)
'
Function edmWordCount(a$) As Long
  If Len(a$) = 0 Then Exit Function
  #Register None
  Local wCount As Long, Pos As Long, wordFlag As Byte

  ! mov eax, a$            ; eax = pointer to string handle
  ! mov eax, [eax]         ; eax = pointer to string data
  ! mov ecx, [eax-4]       ; length of string

  BgnCtLoop:
     ! mov dl, [eax]       ; move current char into dl
     ! cmp dl, 39
     ! je WordCheck        ; if dl = 39 => ASC(39)
     ! cmp dl, 45
     ! je WordCheck        ; if dl = 45 => ASC(45)
     ! cmp dl, 65          ; CASE 65 TO 90 => ASC(65 to 90)
     ! jb NextChar         ;    if dl < 65
     ! cmp dl, 91
     ! jb WordCheck        ;    if dl < 91
     ! cmp dl, 95          ; CASE 95 TO 122
     ! jb NextChar         ;    if dl < 97
     ! cmp dl, 123
     ! jb WordCheck        ;    if dl < 123
     ! cmp dl, 146
     ! je WordCheck        ; if dl = 146
     ! cmp dl, 180
     ! je WordCheck        ; if dl = 180

     ! cmp dl, 192         ; CASE 192 TO 214, 216 TO 222  - extended ANSI characters
     ! jb NextChar         ;    if dl < 192
     ! cmp dl, 215
     ! jb WordCheck        ;    if dl < 215
     ! je NextChar         ;    if dl = 215
     ! cmp dl, 223
     ! jb WordCheck        ;    if dl < 223

     ! cmp dl, 224         ; CASE 224 TO 246, 248 TO 255  - extended ANSI characters
     ! jb NextChar         ;    if dl < 224
     ! cmp dl, 247
     ! jb WordCheck        ;    if dl < 247
     ! je NextChar         ;    if dl = 247
     ! cmp dl, 255
     ! jbe WordCheck       ;    if dl <= 255

  NextChar:
     ! cmp wordFlag, 1
     ! je SetFlag0
     ! jmp GetNextChar

  WordCheck:
     ! cmp wordFlag, 0
     ! je SetFlag1
     ! jmp GetNextChar

  SetFlag0:
     ! dec wordFlag
       ! cmp dh, 45             ; check last character
       ! je GetNextChar         ; if last character (dh) = 45 (-), jump out
         ! inc pos              ; add one to last pos
         ! cmp pos, eax         ; compare current pos and last - if equal, single character word
         ! je chkSingle         ; if equal (len = 1), then validate - usually only vocables are valid..
     ! inc wCount
     ! jmp GetNextChar

  chkSingle: 'aeiouyAEIOUY = OK
             ' -> the rest is done below
     ! cmp dh, 65        ; A
     ! jb GetNextChar    ; if < A

     ! cmp dh, 97        ; a
     ! je SetFlag2
     ! cmp dh, 101       ; e
     ! je SetFlag2
     ! cmp dh, 105       ; i
     ! je SetFlag2
     ! cmp dh, 111       ; o
     ! je SetFlag2
     ! cmp dh, 117       ; u
     ! je SetFlag2
     ! cmp dh, 121       ; y
     ! je SetFlag2
     ! cmp dh, 224       ; 
     ! je SetFlag2
     ! cmp dh, 225       ; 
     ! je SetFlag2
     ! cmp dh, 228       ; 
     ! je SetFlag2
     ! cmp dh, 229       ; 
     ! je SetFlag2
     ! cmp dh, 230       ; 
     ! je SetFlag2
     ! cmp dh, 232       ; 
     ! je SetFlag2
     ! cmp dh, 233       ; 
     ! je SetFlag2
     ! cmp dh, 242       ; 
     ! je SetFlag2
     ! cmp dh, 243       ; 
     ! je SetFlag2
     ! cmp dh, 246       ; 
     ! je SetFlag2
     ! cmp dh, 248       ; 
     ! je SetFlag2
     ! cmp dh, 253       ; 
     ! je SetFlag2

     ! cmp dh, 65        ; A
     ! je SetFlag2
     ! cmp dh, 69        ; E
     ! je SetFlag2
     ! cmp dh, 73        ; I
     ! je SetFlag2
     ! cmp dh, 79        ; O
     ! je SetFlag2
     ! cmp dh, 85        ; U
     ! je SetFlag2
     ! cmp dh, 89        ; Y
     ! je SetFlag2
     ! cmp dh, 192       ; 
     ! je SetFlag2
     ! cmp dh, 193       ; 
     ! je SetFlag2
     ! cmp dh, 196       ; 
     ! je SetFlag2
     ! cmp dh, 197       ; 
     ! je SetFlag2
     ! cmp dh, 198       ; 
     ! je SetFlag2
     ! cmp dh, 200       ; 
     ! je SetFlag2
     ! cmp dh, 201       ; 
     ! je SetFlag2
     ! cmp dh, 210       ; 
     ! je SetFlag2
     ! cmp dh, 211       ; 
     ! je SetFlag2
     ! cmp dh, 214       ; 
     ! je SetFlag2
     ! cmp dh, 216       ; 
     ! je SetFlag2
     ! cmp dh, 221       ; 
     ! je SetFlag2

     ! jmp GetNextChar

  SetFlag2:
     ! inc wCount
     ! jmp GetNextChar

  SetFlag1:
     ! cmp dl, 39          ; not these as first character
     ! je NextChar
     ! cmp dl, 45
     ! je NextChar
     ! cmp dl, 95
     ! je NextChar
     ! cmp dl, 96
     ! je NextChar
     ! cmp dl, 180
     ! je NextChar
     ! inc wordFlag        ; else first character is ok, so make flag 1
     ! mov dh, [eax]       ; move current char into dh
     ! mov pos, eax        ; move current pos into pos

  GetNextChar:
     ! inc eax             ; get next character
     ! dec ecx             ; decrease the ln (length) counter
     ! jnz bgnCtLoop       ; jump to begin if not zero

  'Ending procedure
     ! inc ecx             ; incr counter for an eventual final check for an ending word..
     ! cmp wordFlag, 1     ; if word flag is set, that is
     ! je SetFlag0         ; then repeat this one to get and count last word
     ! mov ecx, wCount     ; now, move wordcount into ecx, and
     ! mov Function, ecx   ; move ecx into Function's return value (same as FUNCTION = wCount)

End Function

'
' A fast wordwrap function
'
Function edmWrapText(ByVal txt As String, ByVal nw As Long, ByVal strip As Long) As String
  Local ii As Long, Pos As Long, Letter As Byte Ptr, Letter2 As Byte Ptr

  Letter = StrPtr(txt)
  For ii = 1 To Len(txt)
     If strip Then 'if binary file, readonly - replace "impossible" characters
        Select Case @Letter
           Case 0 To 8, 11 To 12, 14 To 31
              @Letter = 32
        End Select
     End If

     If @Letter = 13 Then
        Pos = 0                   'proper linefeed, reset line-length counter
     ElseIf @Letter = 32 Then
        Letter2 = Letter          'space, store the address
     End If

     Incr Letter : Incr Pos

     If Pos > nw + 1 Then         'line-length exceeds max, so..
        If Letter2 > 0 Then       'if we have an address..
           @Letter2 = 13          'make most recent space a line-feed instead and
           Pos = Letter - Letter2 'reset line-length counter for next line
        End If
     End If

  Next ii

  Function = txt
End Function

'
' Get subs and functions to a combobox (or listbox)
'
Sub edmGetSubs(ByVal hWnd As Dword, ByVal hCombo As Dword, ByVal hList As Dword)
  Local ii As Long, jj As Long, hCtrl As Dword, tAsc As Long, txt As String, oldtxt As String
  Local C_GETCURSEL As Long,   C_GETTEXTLEN As Long, C_GETTEXT As Long, _
        C_RESETCONTENT As Long, C_ADDSTRING As Long, C_SETCURSEL As Long, _
        C_SETITEMDATA As Long, C_FINDSTRINGEXACT As Long

  If hCombo Then
     hCtrl = hCombo
     C_GETCURSEL       = %CB_GETCURSEL
     C_GETTEXTLEN      = %CB_GETLBTEXTLEN
     C_GETTEXT         = %CB_GETLBTEXT
     C_RESETCONTENT    = %CB_RESETCONTENT
     C_ADDSTRING       = %CB_ADDSTRING
     C_SETCURSEL       = %CB_SETCURSEL
     C_SETITEMDATA     = %CB_SETITEMDATA
     C_FINDSTRINGEXACT = %CB_FINDSTRINGEXACT
  ElseIf hList Then
     hCtrl = hList
     C_GETCURSEL       = %LB_GETCURSEL
     C_GETTEXTLEN      = %LB_GETTEXTLEN
     C_GETTEXT         = %LB_GETTEXT
     C_RESETCONTENT    = %LB_RESETCONTENT
     C_ADDSTRING       = %LB_ADDSTRING
     C_SETCURSEL       = %LB_SETCURSEL
     C_SETITEMDATA     = %LB_SETITEMDATA
     C_FINDSTRINGEXACT = %LB_FINDSTRINGEXACT
  End If

  If hCtrl Then
     ii = SendMessage(hCtrl, C_GETCURSEL, 0, 0)                 'Get old selection, if any
     If ii <> %CB_ERR Then
        jj = SendMessage(hCtrl, C_GETTEXTLEN, ii, 0)
        oldtxt = Space$(jj)
        jj = SendMessage(hCtrl, C_GETTEXT, ii, StrPtr(oldtxt))
     End If

     SendMessage hCtrl, C_RESETCONTENT, 0, 0
     txt = "(Declarations)"
     SendMessage hCtrl, C_ADDSTRING, 0, ByVal StrPtr(txt)
     SendMessage hCtrl, C_SETITEMDATA, 0, 1
     If hCombo Then SendMessage hCtrl, C_SETCURSEL, 0, 0
     If hWnd = 0 Then Exit Sub                            'only wanted to initialize, so..
  End If

  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0) 'now, get subs and functions
  If @ed.UBtxBuf = 0 Then Exit Sub                        'if no or only one line, exit
  Local ac As Long, kk As Long, sFlag As Long, Buf As String, xWord As String

  If hCtrl = 0 Then edmBookMarkSet hWnd, 0, -1      'set one at top..
  For ii = 0 To @ed.UBtxBuf
     txt = edmTxtGetLine(hWnd, ii)
     If Len(txt) < 4 Then Iterate    'line is too short..

     If sFlag = 0 Then
        Buf = UCase$(LTrim$(txt))
        tAsc = Asc(Buf)
        If tAsc = 70 And Left$(Buf, 9) = "FUNCTION " Then
           txt = LTrim$(Mid$(LTrim$(txt), 10))
           sFlag = 2 : xWord = Parse$(txt, Any " (", 1) : kk = ii + 1
           If xWord = "_" Then sFlag = 3
        ElseIf tAsc = 83 And Left$(Buf, 4) = "SUB " Then
           txt = LTrim$(Mid$(LTrim$(txt), 5))
           sFlag = 1 : xWord = Parse$(txt, Any " (", 1) : kk = ii + 1
           If xWord = "_" Then sFlag = 3
        ElseIf tAsc = 67 And Left$(Buf, 9) = "CALLBACK " Then
           txt = LTrim$(Mid$(LTrim$(txt), 10))
           If Asc(txt) = 70 And UCase$(Left$(txt, 9)) = "FUNCTION " Then
              txt = LTrim$(Mid$(txt, 10))
              sFlag = 2 : xWord = Parse$(txt, Any " (", 1) : kk = ii + 1
              If xWord = "_" Then sFlag = 3
           End If
        ElseIf tAsc = 83 And Left$(Buf, 7) = "STATIC " Then
           txt = LTrim$(Mid$(txt, 8))
           If Asc(txt) = 70 And UCase$(Left$(txt, 9)) = "FUNCTION " Then
              txt = LTrim$(Mid$(txt, 10))
              sFlag = 2 : xWord = Parse$(txt, Any " (", 1) : kk = ii + 1
              If xWord = "_" Then sFlag = 3
           End If
        End If

     ElseIf sFlag = 3 Then 'line was wrapped right after first word
        sFlag = 2 : xWord = Parse$(LTrim$(txt), Any " (", 1)

     Else
        txt = UCase$(LTrim$(txt))
        Select Case sFlag
        Case 1
           If Asc(txt) = 69 And Left$(txt, 7) = "END SUB" Then
              If hCtrl Then
                 jj = SendMessage(hCtrl, C_ADDSTRING, 0, ByVal StrPtr(xWord))
                 SendMessage hCtrl, C_SETITEMDATA, jj, kk
              Else
                 edmBookMarkSet hWnd, kk - 1, -1
              End If
              sFlag = 0
           End If
        Case 2
           If Asc(txt) = 69 And Left$(txt, 12) = "END FUNCTION" Then
              If hCtrl Then
                 jj = SendMessage(hCtrl, C_ADDSTRING, 0, ByVal StrPtr(xWord))
                 SendMessage hCtrl, C_SETITEMDATA, jj, kk
              Else
                 edmBookMarkSet hWnd, kk - 1, -1
              End If
              sFlag = 0
           End If
        End Select
     End If
  Next

  If hCtrl Then
     If Len(oldtxt) Then
        ii = SendMessage(hCtrl, C_FINDSTRINGEXACT, -1, ByVal StrPtr(oldtxt))
        If ii <> %CB_ERR Then  'CB/LB_ERR has same value, -1
            SendMessage hCtrl, C_SETCURSEL, ii, 0
        End If
     End If
  Else
     SendMessage hWnd, %EDM_REFRESH, 0, -1
  End If

End Sub

'
' Get start or end position of a word
'
Function edmGetWordLimits(ByVal hWnd As Dword, ByVal StartOrEnd As Long, ed As EDM32DATA) As Long
  Local Ac As Long, Bc As Long, cPos As Long, xp As Long, lLen As Long, Letter As Byte Ptr

  xp   = ed.txt
  cPos = edmTxtGetPos(hWnd, ed.xCaret, ed.yCaret)

  If StartOrEnd Then      'Ctrl+VK_LEFT (or Ctrl+BackSpace) pressed - Start pos
     If ed.xCaret > 0 Then
        Letter = xp + cPos - 1
        For Ac = ed.xCaret - 1 To 1 Step - 1
           If @Letter <> 32 Then Exit For
           Decr Letter
        Next

        If Ac Then
           Decr Letter : Bc = Ac
           For Ac = Bc To 1 Step - 1
              If @Letter = 32 Then Exit For
              Decr Letter
           Next
        End If

     Else
        If ed.yCaret > 0 Then
           Decr ed.yCaret
           Ac = edmTxtGetLineLen(ed, ed.yCaret)
        End If
     End If

  Else                    'Ctrl + VK_RIGHT pressed - End pos
    lLen = edmTxtGetLineLen(ed, ed.yCaret)
    If ed.xCaret < lLen Then
       Letter = xp + cPos + 1 : Ac = ed.xCaret + 1

       Do While @Letter <> 32
          Incr Ac : Incr Letter
       Loop Until Ac >= lLen

       If Ac < lLen Then
          Do While @Letter = 32
             Incr Ac : Incr Letter
          Loop Until Ac >= lLen
       End If

       If ed.yCaret = ed.UBtxBuf Then Ac = lLen
    End If
  End If

  Function = Ac
End Function

'
' Load PB keywords into memory block, and index them
'
Function LoadPBdata(ed As EDM32DATA) As Long
  Local ii As Long, jj As Long, kk As Long, rc As Long, q As Dword Ptr, q2 As Dword Ptr

  rc = DataCount
  ReDim dArray(rc - 1) As String   'zero based, so -1

  For ii = 1 To rc                  'read the data into the array
     dArray(ii-1) = UCase$(Read$(ii))
     kk = kk + Len(dArray(ii-1))
  Next

  Array Sort dArray()              'Sort the array

  ed.pData       = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, kk)
  ed.pDataEndPos = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, rc * 4 + 4)
  ed.pStartCount = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, 124)
  ed.pEndCount   = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, 124)

  If ed.pData Then                      ' place local array into control specific memory block
     kk = 0 : q = ed.pDataEndPos
     For ii = 0 To rc - 1
        @q = kk : Incr q
        Poke$ ed.pData + kk, dArray(ii) ' Poke text into memory block
        kk = kk + Len(dArray(ii))       ' End-position in memory block
     Next
     @q = kk                            ' Store end-positions
  End If

  jj = 64 : q = ed.pStartCount : q2 = ed.pEndCount 'Index on first character, %&#, etc = 0, A = 1, etc..
  For ii = 0 To UBound(dArray)
     kk = Asc(dArray(ii))
     If kk > jj Then    'A - Z
        q2 = ed.pEndCount + (jj - 64) * 4 : @q2 = Max&(0, ii - 1)
        jj = kk 'ASC(dArray(ii))
        q  = ed.pStartCount + (jj - 64) * 4 : @q = ii
     End If
  Next
  q2 = ed.pEndCount + (jj - 64) * 4 : @q2 = ii - 1

  Function = rc  'Return the count, in case we ever should need it..

'PBWin 7.0 syntax color data - think at least most of it..  :-)
Data #BLOAT, #COMPILE, #DEBUG, #DIM, #ELSE, #ELSEIF, #ENDIF, #IF, #INCLUDE, #OPTION, #REGISTER, #RESOURCE
Data #SEGMENT, #STACK, #TOOLS, $BEL, $BS, $COMPILE, $CR, $CRLF, $DEBUG, $DIM, $DQ, $ELSE, $ELSEIF, $ENDIF
Data $EOF, $ESC, $FF, $IF, $INCLUDE, $LF, $NUL, $OPTION, $REGISTER, $RESOURCE, $SEGMENT, $SPC, $STACK
Data $TAB, $VT, %DEF, %FALSE, %NULL, %PB_EXE, %TRUE
Data ABS, ACCEL, ACCEPT, ACCESS, ACODE$, ADD, ADDR, ALIAS, ALL, AND, ANY, APPEND, ARRAY, ARRAYATTR
Data AS, ASC, ASCEND, ASCIZ, ASCIIZ, AT, ATN, ATTACH, ATTRIB, BAR, BASE, BAUD, BDECL, BEEP
Data BIN$, BINARY, BIT, BITS%, BITS&, BITS?, BITS??, BITS???, BREAK, BUTTON, BYCMD, BYCOPY, BYREF
Data BYTE, BYVAL, CALC, CALL, CALLBACK, CALLSTK, CALLSTK$, CALLSTKCOUNT, CASE, CATCH, CBCTL, CBCTLMSG
Data CBHNDL, CBLPARAM, CBMSG, CBWPARAM, CBYT, CCUR, CCUX, CD, CDBL, CDECL, CDWD, CEIL, CEXT, CHDIR
Data CHDRIVE, CHECK, CHECK3STATE, CHECKBOX, CHOOSE, CHOOSE&, CHOOSE%, CHOOSE$, CHR$, CINT, CLIENT, CLNG
Data CLOSE, CLS, CLSID$, CODEPTR, COLLATE, COLOR, COMBOBOX, COMM, COMMAND$, CON, CONNECT, CONST, CONTROL
Data COS, CQUD, CREATE, CSET, CSET$, CSNG, CTSFLOW, CUR, CURDIR$, CURRENCY, CURRENCYX, CUX, CVBYT, CVCUR
Data CVCUX, CVD, CVDWD, CVE, CVI, CVL, CVQ, CVS, CVWRD, CWRD, DATA, DATACOUNT, DATE$, DECLARE, DECR, DEFAULT
Data DEFBYT, DEFCUR, DEFCUX, DEFDBL, DEFDWD, DEFEXT, DEFINT, DEFLNG, DEFQUD, DEFSNG, DEFSTR, DEFWRD, DELETE
Data DESCEND, DIALOG, DIM, DIR$, DISABLE, DISKFREE, DISKSIZE, DISPATCH, DLL, DLLMAIN, DO, DOEVENTS, DOUBLE
Data DOWN, DRAW, DSRFLOW, DSRSENS, DTRFLOW, DTRLINE, DWORD, ELSE, ELSEIF, EMPTY, ENABLE, END, ENVIRON$
Data EOF, EQV, ERASE, ERR, ERRAPI, ERRCLEAR, ERROR, ERROR$, EXE, EXIT, EXP, EXP10, EXP2, EXPLICIT, EXPORT
Data EXT, EXTENDED, EXTRACT$, FILEATTR, FILECOPY, FILENAME$, FILESCAN, FILL, FINALLY, FIX, FLOW, FLUSH, FOCUS
Data FONT, FOR, FORMAT$, FORMFEED, FRAC, FRAME, FREEFILE, FROM, FUNCTION, FUNCNAME$, GET, GET#, GET$
Data GETATTR, GLOBAL, GOSUB, GOTO, GUID$, GUIDTXT$, HANDLE, HEX$, HIBYT, HIINT, HIWRD, HOST, ICASE, ICON
Data IDN, IF, IFACE, IIF, IIF&, IIF%, IIF$, IMAGE, IMAGEX, IMGBUTTON, IMGBUTTONX, IMP, IN, INCR, INP, INOUT
Data INPUT, INPUT#, INPUTBOX$, INSERT, INSTR, INT, INTERFACE, INTEGER, INV, ISFALSE, ISNOTHING
Data ISOBJECT, ISTRUE, ITERATE, JOIN$, KILL, LABEL, LBOUND, LCASE$, LEFT, LEFT$, LEN, LET, LIB, LIBMAIN
Data LINE, LISTBOX, LOBYT, LOC, LOCAL, LOCK, LOF, LOG, LOG10, LOG2, LOINT, LONG, LOOP, LOWRD, LPRINT
Data LSET, LSET$, LTRIM$, MACRO, MACROTEMP, MAIN, MAKDWD, MAKINT, MAKLNG, MAKPTR, MAKWRD, MAT, MAX, MAX$
Data MAX%, MAX&, MCASE$, MEMBER, MENU, MID$, MIN, MIN$, MIN%, MIN&, MKBYT$, MKCUR$, MKCUX$, MKD$
Data MKDIR, MKDWD$, MKE$, MKI$, MKL$, MKQ$, MKS$, MKWRD$, MOD, MODAL, MODELESS, MOUSEPTR, MSGBOX
Data NAME, NEW, NEXT, NONE, NOT, NOTHING, NOTIFY, NULL, OBJACTIVE, OBJECT, OBJPTR, OBJRESULT, OCT$, OF
Data OFF, ON, OPEN, OPT, OPTION, OPTIONAL, OR, OUT, OUTPUT, PAGE, PARITY, PARITYCHAR, PARITYREPL, PARITYTYPE
Data PARSE, PARSE$, PARSECOUNT, PBD, PBMAIN, PEEK, PEEK$, PIXELS, POINTER, POKE, POKE$, POPUP, PORT, POST
Data PRESERVE, PRINT, PRINT#, PRIVATE, PROFILE, PROGID$, PTR, PUT, PUT$, QUAD, QWORD, RANDOM, RANDOMIZE, READ
Data READ$, RECEIVE, RECORDS, RECV, REDIM, REDRAW, REGEXPR, REGISTER, REGREPL, REMAIN$, REMOVE$, REPEAT$
Data REPLACE, RESET, RESUME, RET16, RET32, RET87, RETAIN$, RETP16, RETP32, RETPRM, RETURN, RGB, RIGHT
Data RIGHT$, RING, RLSD, RMDIR, RND, ROTATE, ROUND, RSET, RSET$, RTRIM$, RTSFLOW, RXBUFFER, RXQUE, SCAN
Data SCROLLBAR, SDECL, SEEK, SELECT, SEND, SERVER, SET, SETATTR, SETEOF, SGN, SHARED, SHELL
Data SHIFT, SHOW, SIGNED, SIN, SINGLE, SIZE, SIZEOF, SLEEP, SORT, SPACE$, SPC, SQR, STATE, STATIC, STATUS
Data STDCALL, STEP, STOP, STR$, STRDELETE$, STRING, STRING$, STRINSERT$, STRPTR, STRREVERSE$, SUB, SUSPEND
Data SWAP, SWITCH, SWITCH&, SWITCH%, SWITCH$, TAB, TAB$, TAGARRAY, TALLY, TAN, TCP, TEXT, TEXTBOX, THEN
Data THREAD, THREADCOUNT, THREADID, TIME$, TIMEOUT, TIMER, TO, TOGGLE, TRACE, TRIM$, TRN, TRY, TXBUFFER
Data TXQUE, TYPE, UBOUND, UCASE, UCASE$, UCODE$, UDP, UNION, UNITS, UNLOCK, UNTIL, UP, USER, USING, USING$
Data VAL, VARIANT, VARIANT#, VARIANT$, VARIANTVT, VARPTR, VERIFY, VERSION3, VERSION4, VERSION5
Data WEND, WHILE, WIDTH, WIDTH#, WINMAIN, WITH, WORD, WRITE, WRITE#, XOR, XINPFLOW, XOUTFLOW, ZER

End Function

'
' Load HTML keywords into memory block
'
Function LoadHTMLdata(ed As EDM32DATA) As Long
  Local ii As Long, jj As Long, kk As Long, rc As Long, q As Long Ptr, q2 As Long Ptr

  rc = DataCount
  ReDim dArray(rc - 1) As String   'zero based, so -1

  For ii = 1 To rc                  'read the data into the array
     dArray(ii-1) = Read$(ii)
     kk = kk + Len(dArray(ii-1))
  Next
  Array Sort dArray() 'Sort and index the array on first character, %&#, etc = 0, A = 1, B = 2, etc..

  ed.hData       = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, kk)
  ed.hDataEndPos = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, rc * 4 + 4)
  ed.hStartCount = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, 124)
  ed.hEndCount   = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, 124)

  If ed.hData Then
     kk = 0 : q = ed.hDataEndPos
     For ii = 0 To rc - 1
        @q = kk : Incr q
        Poke$ ed.hData + kk, dArray(ii)  ' Poke text into memory block
        kk = kk + Len(dArray(ii))        ' Set End pos into
     Next
     @q = kk
  End If

  jj = 64 : q = ed.hStartCount : q2 = ed.hEndCount
  For ii = 0 To UBound(dArray)
     kk = Asc(dArray(ii))
     If kk > jj Then    'A - Z
        q2 = ed.hEndCount + (jj - 64) * 4 : @q2 = Max&(0, ii - 1)
        jj = kk 'ASC(dArray(ii))
        q = ed.hStartCount + (jj - 64) * 4 : @q = ii
     End If
  Next
  q2 = ed.hEndCount + (jj - 64) * 4 : @q2 = ii - 1

  Function = rc  'Return the count, in case we ever should need it..

'HTML syntax color data
Data A, ABBR, ACCESSKEY, ACRONYM, ACTION, ADDRESS, ALIGN, ALINK, ALT, APPLET, ARCHIVE, AREA, AU, AXIS
Data B, BACKGROUND, BANNER, BASE, BASEFONT, BGCOLOR, BGSOUND, BIG, BLINK, BLOCKQUOTE
Data BODY, BORDER, BORDERCOLORLIGHT, BOTTOM, BQ, BR, BUTTON
Data CAPTION, CDATA, CELLPADDING, CELLSPACING, CENTER, CHAR, CHAROFF, CHARSET, CHECKED, CITE, CLASS, CLASSID
Data CLEAR, CODE, CODEBASE, CODETYPE, COL, COLGROUP, COLOR, COLS, COLSPAN, COMMENT, CONTENT, COORDS
Data DAT, DATA, DATETIME, DECLARE, DD, DEFER, DFN, DEL, DIR, DISABLED, DISABLES, DIV, DL, DOCTYPE, DT
Data EM, EMBED, FACE, FIG, FIELDSET, FN, FONT, FORM, FRAME, FRAMEBORDER, FRAMESET
Data H1, H2, H3, H4, H5, H6, HEAD, HEADERS, HEIGHT, HIGH, HR, HREF, HREFLANG, HSPACE, HTML, HTTP-EQUIV
Data I, ID, IFRAME, IMG, INPUT, INS, ISMAP, JUSTIFY, KBD
Data LABEL, LANG, LANGUAGE, LEGEND, LEFT, LEFTMARGIN, LI, LINK, LISTING, LONGDESC, LOOP, LOW
Data MAP, MARGINHEIGHT, MARGINLEFT, MARGINRIGHT, MARGINWIDTH, MARQUEE, MAXLENGTH, MEDIA, MENU, META
Data METHOD, MIDDLE, MULTICOL
Data NAME, NEXTID, NOBR, NOFRAMES, NORESIZE, NOSAVE, NOSCRIPT, NOSHADE, NOTE, NOWRAP
Data OBJECT, OL, ONBLUR, ONCHANGE, ONFOCUS, ONLOAD, ONMOUSEOUT, ONMOUSEOVER, ONMOUSEUP, ONSELECT
Data ONUNLOAD, OPTGROUP, OPTION, OVERLAY
Data P, PARAM, PERSON, PLAINTEXT, PLUGINSPAGE, PRE, PROFILE, PUBLIC
Data Q, QUALITY, RANGE, READONLY, RECT, REF, REL, RESET, REV, RIGHT, RIGHTMARGIN, ROW, ROWS, ROWSPAN
Data S, SAMP, SCHEME, SCOPE, SCRIPT, SCRIPTA, SCROLLING, SELECT, SELECTED, SHAPE, SIZE, SMALL, SOUND
Data SPACER, SPAN, SRC, STANDBY, STRIKE, STRONG, STYLE, SUB, SUBMIT, SUMMARY, SUP
Data TAB, TABINDEX, TABLE, TARGET, TBODY, TD, TEXT, TEXTAREA, TFOOT, TH, THEAD, TITLE, TOP, TOPMARGIN
Data TR, TT, TYPE, U, UL, URL, USEMAP, VALIGN, VALUE, VALUETYPE, VAR, VLINK, VSPACE, WBR, WIDTH, WRAP, XMP

End Function

'
' Load RC keywords into memory block
'
Function LoadRCdata(ed As EDM32DATA) As Long
  Local ii As Long, jj As Long, kk As Long, rc As Long, q As Dword Ptr, q2 As Dword Ptr

  rc = DataCount
  ReDim dArray(rc - 1) As String   'zero based, so -1

  For ii = 1 To rc                  'read the data into the array
     dArray(ii-1) = UCase$(Read$(ii))
     kk = kk + Len(dArray(ii-1))
  Next

  Array Sort dArray() 'Sort and index the array on first character, %&#, etc = 0, A = 1, B = 2, etc..

  ed.rData       = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, kk)
  ed.rDataEndPos = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, rc * 4 + 4)
  ed.rStartCount = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, 124)
  ed.rEndCount   = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, 124)

  If ed.rData Then
     kk = 0 : q = ed.rDataEndPos
     For ii = 0 To rc - 1
        @q = kk : Incr q
        Poke$ ed.rData + kk, dArray(ii)  ' Poke text into memory block
        kk = kk + Len(dArray(ii))        ' Set End pos into
     Next
     @q = kk
  End If

  jj = 64 : q = ed.rStartCount : q2 = ed.rEndCount
  For ii = 0 To UBound(dArray)
     kk = Asc(dArray(ii))
     If kk > jj Then    'A - Z
        q2 = ed.rEndCount + (jj - 64) * 4 : @q2 = Max&(0, ii - 1)
        jj = kk
        q = ed.rStartCount + (jj - 64) * 4 : @q = ii
     End If
  Next
  q2 = ed.rEndCount + (jj - 64) * 4 : @q2 = ii - 1

  Function = rc  'Return the count, in case we ever should need it..

'RC syntax color data, extracted from RCDLL.DLL, the RC help file and other sources..  :-)
'Note, first line, definition words, must be lower case in RC file. Is upper case
'here for search resons, but will *not* be automatically changed in text, since RC.EXE
'is case sensitive.

Data #INCLUDE, #DEFINE, #ELIF, #ELSE, #ENDIF, #IF, #IFDEF, #IFNDEF, #INCLUDE, #PRAGMA, #UNDEF
Data ACCELERATOR, ACCELERATORS, ALT, ANICURSOR, ANIICON, ASCII, AUTO3STATE
Data AUTOCHECKBOX, AUTORADIOBUTTON, AVI, BEDIT, BEGIN, BITMAP, BLOCK, BUTTON
Data CAPTION, CHARACTERISTICS, CHECKBOX, CHECKED, CLASS, COMBOBOX, CONTROL, CTEXT, CURSOR
Data DEFPUSHBUTTON, DIALOG, DIALOGEX, DISCARDABLE, DLGINCLUDE, DLGINIT, EDIT, EDITTEXT, END, EXSTYLE
Data FILEFLAGS, FILEFLAGSMASK, FILEOS, FILESUBTYPE, FILETYPE, FILEVERSION, FIXED, FONT, FONTDIR
Data GRAYED, GROUP_CURSOR, GROUP_ICON, GROUPBOX, HELP, ICON, IEDIT, IMPURE, INACTIVE
Data LANGUAGE, LISTBOX, LOADONCALL, LTEXT, MAINMENU, MENU, MENUBARBREAK, MENUBREAK, MENUEX
Data MENUITEM, MESSAGETABLE, MOVEABLE, NOINVERT, NONSHARED, NOT, OBJECTS, OWNERDRAW
Data PLUGPLAY, POPUP, PRELOAD, PRODUCTVERSION, PURE, PUSHBOX, PUSHBUTTON, RADIOBUTTON, RCDATA, RTEXT
Data SCROLLBAR, SEPARATOR, SHARED, SHIFT, STATE3, STATIC, STRING, STRINGTABLE, STYLE, SOUND
Data USERBUTTON, VALUE, VERSION, VERSIONINFO, VIRTKEY, VS_VERSION_INFO, WAVE, VXD


End Function

'********************************************************************
' Create Bitmap for left side's selection/bookmark bar
'********************************************************************
Function MakeScrlBitMap(ByVal hDlg As Dword) As Dword
  Local Loop1 As Long, Loop2 As Long, hdc As Dword, hdcMem As Dword, _
        hBitmap As Dword, hBitmapOld As Dword, RC As RECT

  hdc = GetDC(hDlg)
  hdcMem = CreateCompatibleDC(hdc)
  hBitmap = CreateCompatibleBitmap(hdc, 8, 8)
  hBitmapOld = SelectObject(hdcMem, hBitmap)

  RC.nleft = 0 : RC.ntop = 0 : RC.nright = 8 : RC.nbottom = 8
  FillRect hdcMem, RC, GetSysColorBrush(%COLOR_SCROLLBAR)

  For Loop1 = 0 To 7 Step 1
     If Loop1 Mod 2 = 0 Then
        For Loop2 = 0 To 7 Step 2
           SetPixel hdcMem, Loop2, Loop1, GetSysColor(%COLOR_BTNHIGHLIGHT)
        Next
     Else
        For Loop2 = 1 To 7 Step 2
           SetPixel hdcMem, Loop2, Loop1, GetSysColor(%COLOR_BTNHIGHLIGHT)
        Next
     End If
  Next
  Function = CreatePatternBrush (hBitmap)

  SelectObject hdcMem, hBitmapOld
  DeleteObject hBitmap
  DeleteDC hdcMem
  ReleaseDC hDlg, hdc

End Function

'
' Create Tooltips control
'
Function MakeScrollInfoLabel(ByVal hParent As Dword) As Dword
  Local ed As EDM32DATA Ptr, lf As LOGFONT, hF As Dword, hW As Dword, hDC As Dword, lh As Long
  ed = GetWindowLong(hParent, 0)

  hF = GetStockObject(%DEFAULT_GUI_FONT)  'get font data for label height
  If hF Then GetObject hF, SizeOf(lf), ByVal VarPtr(lf)

  hDc = GetDc(%HWND_DESKTOP)
    If lf.lfHeight < 12 Then lf.lfHeight = 12                      'minimum height
    lh = Abs((lf.lfHeight * GetDeviceCaps(hDC, %LOGPIXELSY)) \ 72) 'get height vs big/small system font
    If lh < 16 Then lh = 16                                        'ensure height never is less
  ReleaseDC %HWND_DESKTOP, hDC

  hW = CreateWindowEx (0, "STATIC",_                             'window class name
                       ByVal 0,_                                 'window text
                       %WS_CHILD Or %WS_BORDER Or %SS_CENTER, _  'window style
                       0, 0, 100, lh, _                          'initial x-y-w-h
                       hParent,_                                 'parent window handle
                       0,_                                       'control id
                       @ed.hInst, _                              'program instance handle
                       ByVal %NULL)                              'creation parameter

  If hF And hW Then SendMessage hW, %WM_SETFONT, hF, 0

  Function = hW                 '<- return what we got, hopefully control's handle
End Function

'
' Scan specified keyword memoryblock for match
'
Function edmArrayScan(ByVal txt As String, ByVal mb As Long, ByVal Idx As Long, _
                      ByVal dStart As Long, ByVal dEnd As Long) As Long
  Local r As Long, p As Long, q As Dword Ptr, q2 As Dword Ptr, wL As Long, wp As Byte Ptr

  wp = StrPtr(txt)                 ' mb points to memory block with words
  wL = Len(txt)                    ' Idx points to earch word's position in mb

  q = Idx + dStart * 4 - 4
  For p = dStart To dEnd           ' Word number - Array element
     q  = q + 4                    ' previous word's end pos = this word's start pos
     If Peek(mb + @q) = @wp Then   ' if first character macthes
        q2 = q + 4                 ' this word's end pos
        If @q2 - @q = wL Then      ' if length matches
           For r = 1 To wL - 1     ' check the rest of the characters against each other
              Incr wp              ' Note: have tried PEEK$ here - but this loop is much faster
              If Peek(mb + @q + r) <> @wp Then Exit For  'no match, continue
           Next

           If r = wL Then          ' then all matches, exit
              Function = %TRUE : Exit Function 'enough and fastest to return %TRUE
           Else                                'if to return pos instead, use:  dStart + p + 1
              wp = StrPtr(txt)     'reset pointer and iterate
           End If
        End If
     End If
  Next

End Function

'
' Add/insert bookmark space to bookmark's memoryblock
' Note: There are some "issues" with ReAlloc, so it's safer to allocate
' a new memoryblock, copy old and free old. A bit overhead for a short
' while there, but sometimes safety must go before cost..
'
Sub edmBookMarkInsert(ed As EDM32DATA, ByVal Pos As Long, ByVal Count As Long)
  Local mOld As Dword, mNew As Dword

  mNew = HeapAlloc(ed.heapbMarks, %HEAP_ZERO_MEMORY, (ed.UBtxBuf + Count - 1) * 4 + 128) 'Allocate new memblock
  mOld = ed.bMarks                              'Old block's pointer

  MoveMemory ByVal mNew, _                    'New memblock, pos 0
             ByVal mOld, _                    'Old memblock, pos 0
             Pos * 4                          'Move all Left of given pos there

  MoveMemory ByVal (mNew + (Pos * 4) + (Count * 4)), _   'New memblock, given pos + 4 (inserted)
             ByVal (mOld + (Pos * 4)), _                 'Old memblock, given pos
            (ed.UBtxBuf * 4) - (Pos * 4) + 4             'Move all right of pos to new location

  HeapFree ed.heapbMarks, 0&, ByVal ed.bMarks 'Free old block
  ed.bMarks = mNew                            'Assign new block's handle to structure

End Sub

'
' Delete bookmark space at given position in bookmark's memoryblock
'
Sub edmBookMarkDelete(ed As EDM32DATA, ByVal Pos As Long, ByVal Count As Long)
  Local mOld As Dword, mNew As Dword

  mNew = HeapAlloc(ed.heapbMarks, %HEAP_ZERO_MEMORY, (ed.UBtxBuf - (Count - 1)) * 4 + 128) 'Allocate new memory block
  mOld = ed.bMarks                                'Pointer to old block

  If Pos Then
     MoveMemory ByVal mNew, _         'New memblock, pos 0
                ByVal mOld, _         'Old memblock, pos 0
                Pos * 4               'Move all up to given pos there
  End If

  If Pos < ed.UBtxBuf Then
     MoveMemory ByVal (mNew + (Pos * 4)), _               'New memblock, given pos
                ByVal (mOld + ((Pos + Count) * 4)), _     'Old memblock, given pos + 4 (avoid pos)
               (ed.UBtxBuf * 4) - ((Pos + Count - 1) * 4) 'Move all right of pos to new location
  End If

  HeapFree ed.heapbMarks, 0&, ByVal ed.bMarks   'Free old block
  ed.bMarks = mNew                              'Assign new block's handle to structure

End Sub

'
' Set/Reset bookmark, %TRUE or %FALSE
'
Sub edmBookMarkSet(ByVal hWnd As Dword, ByVal y As Long, ByVal bmVal As Long)
  Local p As Dword Ptr, ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)
  p  = @ed.bMarks + (y * 4)
  @p = bmVal
End Sub

'
' Get bookmark status for given line, return %TRUE or %FALSE
'
Function edmBookMarked(ByVal hWnd As Dword, ByVal y As Long) As Long
  Local p As Dword Ptr, ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)
  p = @ed.bMarks + (y * 4)
  Function = @p
End Function

'
' Reset all bookmarks = free and allocate new bookmark memoryblock from scratch
'
Sub edmBookMarkResetAll(ByVal hWnd As Dword)
  Local ii As Long, p As Dword Ptr
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)

  p  = @ed.bMarks
  For ii = 0 To @ed.UBtxBuf
     If @p Then @p = 0
     Incr p
  Next
End Sub

'
' Reset all bookmarks = free and allocate new bookmark memoryblock from scratch
'
Sub edmBookMarkInitialize(ByVal hWnd As Dword)
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)

  If @ed.bMarks Then HeapFree @ed.heapbMarks, 0&, ByVal @ed.bMarks
  @ed.bMarks = HeapAlloc(@ed.heapbMarks, %HEAP_ZERO_MEMORY, 128)
End Sub

'
' Get bookmark count
'
Function edmBookMarkGetCount(ByVal hWnd As Dword) As Long
  Local ii As Long, jj As Long, p As Dword Ptr
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)

  p  = @ed.bMarks
  For ii = 0 To @ed.UBtxBuf
     If @p Then Incr jj
     Incr p
  Next
  Function = jj
End Function

'
' Get next bookmark, if any
'
Function edmBookMarkGetNext(ByVal hWnd As Dword, ByVal y As Long) As Long
  Local ii As Long, p As Dword Ptr, ed As EDM32DATA Ptr
  ed = GetWindowLong(hWnd, 0)

  Incr y : p = @ed.bMarks + (y * 4)
  For ii = y To @ed.UBtxBuf
     If @p Then 'return an increased value, so results for line 0 can be detected
        Function = ii + 1 : Exit For
     End If
     Incr p
  Next

End Function

'
' Get previous bookmark, if any
'
Function edmBookMarkGetPrev(ByVal hWnd As Dword, ByVal y As Long) As Long
  Local ii As Long, p As Dword Ptr, ed As EDM32DATA Ptr
  ed = GetWindowLong(hWnd, 0)

  Decr y : p = @ed.bMarks + (y * 4)
  For ii = y To 0 Step - 1
     If @p Then 'return an increased value, so results for line 0 can be detected
        Function = ii + 1 : Exit For
     End If
     Decr p
  Next

End Function

'
' Get bookmarks to list (I use this in a goto dialog)
'
Function edmBookMarksToList(ByVal hWnd As Dword, ByVal lsthWnd As Long) As Long
  Local ii As Long, jj As Long, p As Dword Ptr, txt As String
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)
  SendMessage lsthWnd, %LB_RESETCONTENT, 0, 0

  p  = @ed.bMarks
  For ii = 0 To @ed.UBtxBuf
     If @p Then
        txt = edmTxtGetLine(hWnd, ii)
        If Len(txt) > 32 Then txt = Left$(txt, 32) + ".."
        txt = "Line" & Str$(ii + 1) & $Tab & txt
        SendMessage lsthWnd, %LB_ADDSTRING, 0, ByVal StrPtr(txt)
        SendMessage lsthWnd, %LB_SETITEMDATA, jj, ii + 1
        Incr jj
     End If
     Incr p
  Next
  Function = jj
End Function

'
' String memory block routines
' Get longest line's length
'
Function edmTxtGetMaxLineLen(ByVal hWnd As Dword) As Long
  Local ii As Long, jj As Long, p1 As Dword Ptr, p2 As Dword Ptr
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)

  If @ed.UBtxBuf > 0 Then                'if more than one line
     For ii = 1 To @ed.UBtxBuf            'loop through all lines
        p1 = @ed.LineEnd + ((ii - 1) * 4) 'Previous line's end
        p2 = @ed.LineEnd + (ii * 4)       'Current line's end
        jj = Max&(jj, @p2 - @p1)           'Current - Previous
     Next
  Else
     jj = @ed.txtLen + 2                  'else line's length equals text length
  End If
  Function = Max&(0, jj - 2)              'return max line length

End Function

'
' Get given line's length
'
Function edmTxtGetLineLen(ed As EDM32DATA, ByVal y As Long) As Long
  Local p1 As Dword Ptr, p2 As Dword Ptr

  If y = 0 Then                      'line zero's end equals its line length
     p2 = ed.LineEnd
     Function = @p2
  Else
     p1 = ed.LineEnd + ((y - 1) * 4) 'Previous line's end
     p2 = ed.LineEnd + (y * 4)       'Current line's end
     Function = @p2 - @p1 - 2        'Current - Previous - 2 (for $CRLF)
  End If

End Function

'
' String memory block routines
' Get given line's start position in text
'
Function edmTxtGetLineStart(ed As EDM32DATA, ByVal y As Long) As Long
  Local p As Dword Ptr

  If y = 0 Then                     'line zero starts at zero..
     Function = 0
  Else
     p = ed.LineEnd + ((y - 1) * 4) 'Previous line's end
     Function = @p + 2              ' + 2 (for $CRLF)
  End If

End Function

'
' Convert linear position into Y and X
'
Sub edmTxtGetXYpos(ed As EDM32DATA, ByVal Pos As Long, x As Long, y As Long)
  If ed.txtLen = 0 Then Exit Sub
  Local lp As Long, p As Dword Ptr

  p = ed.LineEnd

  For lp = 0 To ed.UBtxBuf - 1
     If @p >= Pos Then
        p = Max&(ed.LineEnd, p - 4)
        Exit For
     End If
     p = p + 4
  Next
  y = lp
  If y = 0 Then
     x = Pos
  Else
     x = Max&(0, Pos - @p - 2)
  End If

End Sub

'
' Convert Y and X into linear position
'
Function edmTxtGetPos(ByVal hWnd As Dword, ByVal x As Long, ByVal y As Long) As Long
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)

  If @ed.txtLen Then
    Local ls As Long, p As Dword Ptr

    If y Then                          'Line 0 always starts at zero
       p = @ed.LineEnd + ((y - 1) * 4) 'Previous line's end = current line's start
       ls = @p + 2                     'zero for first line, + 2 for all others (CRLF)
    End If

     Function = ls + x                 'add given x pos for current line
  End If

End Function

'
' Get Selected length
'
Function edmTxtGetSelLength(ed As EDM32DATA) As Long
  Local p1 As Long, p2 As Long

  p1 = edmTxtGetPos(ed.hWnd, ed.SelStartX, ed.SelStartY) 'selStart
  p2 = edmTxtGetPos(ed.hWnd, ed.SelEndX, ed.SelEndY)     'selEnd
  If p2 < p1 Then Swap p1, p2                    'make sure start is before end
  Function = p2 - p1                             'return SelLen

End Function

'
' Get selected text
'
Function edmTxtGetSelText(ByVal hWnd As Dword, ByVal selFlag As Long) As String
  Local p1 As Long, p2 As Long
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)

  If selFlag Then                                          'if selFlag <> 0, return selected text
     p1 = edmTxtGetPos(hWnd, @ed.SelStartX, @ed.SelStartY) 'selStart
     p2 = edmTxtGetPos(hWnd, @ed.SelEndX, @ed.SelEndY)     'selEnd
     If p2 < p1 Then Swap p1, p2                           'make sure start is before end

  Else                                                     'if selFlag = 0, return all text
     p1 = 0
     p2 = @ed.txtLen
  End If
  Function = Peek$(@ed.txt + p1, p2 - p1)                  'return text

End Function

'
' Get line of text
'
Function edmTxtGetLine(ByVal hWnd As Dword, ByVal y As Long) As String
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)
  If @ed.txtLen = 0 Then Exit Function
  Local lp As Long, ls As Long, p As Dword Ptr, q As Dword Ptr

  lp = @ed.LineEnd           'pointer to line end's memory block

  If y Then                  'Line 0 always starts at zero
     q = lp + ((y - 1) * 4)  'Previous line's end = current line's start
     ls = @q + 2             'zero for first line, + 2 for all others (CRLF)
  End If
  p = lp + (y * 4)           'current line's end
  lp = @p - ls               'current line's length
  If lp > 0 Then Function = Peek$(@ed.txt + ls, lp) 'return line's text

End Function

'
' Get summary of file to compile
'
Function edmTxtGetFileSummary(ByVal hWnd As Dword) As String
  Local ii As Long, isExe As Long, isRes As Long, Pos As Long, pos2 As Long, _
        txt As String, incFiles As String
  Local ed As EDM32DATA Ptr : ed = GetWindowLong(hWnd, 0)

  For ii = 0 To @ed.UBtxBuf
     txt = LTrim$(UCase$(edmTxtGetLine(hWnd, ii)))
     If isExe = 0 Then
        If (Asc(txt) = 35 Or Asc(txt) = 36) And Mid$(txt, 2, 8) = "COMPILE " Then   'only one compile statement per file, at beginning
           If InStr(Mid$(txt, 8), "EXE") Then isExe = 1 'set flag because no need to look for this anymore
           Pos = InStr(txt, $Dq)
           If Pos Then
              Incr Pos
              pos2 = InStr(Pos, txt, $Dq)
              If pos2 Then
                 incFiles = incFiles + "COMPILE:  " + Mid$(txt, Pos, pos2 - Pos) + $CrLf
              End If
           Else
              If isExe Then
                 incFiles = incFiles + "COMPILE EXE" + $CrLf
              Else
                 incFiles = incFiles + "COMPILE DLL" + $CrLf
              End If
           End If
        End If
     End If

     If isRes = 0 Then
        If (Asc(txt) = 35 Or Asc(txt) = 36) And Mid$(txt, 2, 9) = "RESOURCE " Then  'only one resource file per file, anywhere in code
           Pos = InStr(txt, $Dq)
           If Pos Then
              Incr Pos
              pos2 = InStr(Pos, txt, $Dq)
              incFiles = incFiles + "RESOURCE: " + Mid$(txt, Pos, pos2 - Pos)+ $CrLf
              isRes = 1  'set flag because no need to look for this anymore
           End If
        End If
     End If

     If (Asc(txt) = 35 Or Asc(txt) = 36) And Mid$(txt, 2, 8) = "INCLUDE " Then  'can be several INCLUDE statements in file, anywhere in code..
        Pos = InStr(txt, $Dq)
        If Pos Then
           Incr Pos
           pos2 = InStr(Pos, txt, $Dq)
           incFiles = incFiles + "INCLUDE:  " + Mid$(txt, Pos, pos2 - Pos) + $CrLf
        End If
     End If
  Next

  Function = incFiles 'isExe
End Function

'
' Put text string into memory block - Central place for expanding text
'
Sub edmTxtPutString(ed As EDM32DATA, ByVal s As String, ByVal x As Long, ByVal y As Long, _
                    ByVal mFlag As Long, ByVal uFlag As Long, ByVal iType As Long)

  Local ii As Long, L As Long, xp As Long, lp As Long
  Local x1 As Long, x2 As Long, y1 As Long, y2 As Long, sLen As Long

  sLen = Len(s)
  If sLen = 0 Then Exit Sub

  If ed.txtLen + sLen > HeapSize(ed.heapTxt, 0&, ByVal ed.txt) - 512 Then
     edmTxtResizeBlocks ed, ed.txtLen + sLen
  End If

  xp = ed.txt
  lp = ed.LineEnd
  L  = edmTxtGetPos(ed.hWnd, ed.xCaret, ed.yCaret) 'current position
  If sLen > 1 Then ii = Tally(s, $CrLf)     'Count line feeds

  If L < ed.txtLen Then                    'If text is to be inserted
     MoveMemory ByVal (xp + L + sLen), _   'Destination text block
                ByVal (xp + L), _          'Source
                ed.txtLen - L              'Size of block to move
  End If

  Poke$ xp + L, s                            'put text
  ed.txtLen = ed.txtLen + sLen               'increase text length variable

  If ii Then                                        'if any new lines were added
     If ed.xCaret = 0 Then
        edmBookMarkInsert ed, ed.yCaret, ii         'insert to bookmarks memory block
     Else
        edmBookMarkInsert ed, ed.yCaret + 1, ii     'insert to bookmarks memory block
     End If
     edmTxtLineEndResize ed, ii                     'and resize LineEnd memory block
     ed.UBtxBuf = ed.UBtxBuf + ii                   'add number of line feeds
     edmTxtUpDateLineEnds ed, ed.txt, ed.LineEnd    'Update LineEnds
     ed.LineUpdate = 0                              'make sure entire window is redrawn
  Else
     edmTxtAddToLineEnds ed, sLen                   'add the increased position to following line end values
     If uFlag Then ed.LineUpdate = 1
  End If

  If ed.Changed = 0 Then ed.Changed = -1           'set changed flag

  x1 = ed.xCaret : y1 = ed.yCaret
  If mFlag And sLen > 1 Then                 'if not simple character input
     ed.yCaret = ed.yCaret + ii               'add line feeds to yCaret
     L = InStr(-1, s, $CrLf)          'find last line feed
     If L Then
        ed.xCaret = Max&(0, sLen - L - 1)    'set xCaret accordingly
     Else
        ed.xCaret = ed.xCaret + sLen
     End If
     x2 = ed.xCaret : y2 = ed.yCaret
     ed.SelStartX = ed.xCaret : ed.SelEndX = ed.xCaret
     ed.SelStartY = ed.yCaret : ed.SelEndY = ed.yCaret : ed.SelLen = 0
  Else
     x2 = x1 + 1 : y2 = y1
     If uFlag = 0 Then Incr ed.xCaret
  End If

  If ed.confineCaret Then
     ed.xCaretOld = Max&(0, ed.xCaret)
  End If

  If uFlag Then edmUndoAdd ed, 1, iType, s, x1, y1, x2, y2
End Sub

'
' Delete text in memory block - Central place for shrinking text
'
Sub edmTxtDeleteSel(ed As EDM32DATA, ByVal selFlag As Long, ByVal uFlag As Long, ByVal iType As Long)
  If ed.txtLen = 0 Or ed.ReadOnly = 1 Then Exit Sub
  Local L As Long, xp As Long, lp As Long
  Local p1 As Long, p2 As Long, t As Byte Ptr, s As String
  Local x1 As Long, x2 As Long, y1 As Long, y2 As Long', crFlag AS LONG

  xp = ed.txt
  lp = ed.LineEnd

  If selFlag = 1 Then                         'No selection, delete next character (Del)
     p1 = edmTxtGetPos(ed.hWnd, ed.xCaret, ed.yCaret)
     x1 = ed.xCaret : y1 = ed.yCaret
     t = xp + p1
     If @t = 13 Then                          'if Carriage return character (CR, of CRLF)
        p2 = Min&(ed.txtLen, p1 + 2)          'delete line feed, 2 characters
        x2 = 0 : y2 = ed.yCaret + 1
     Else                                     'else regular, 1 character
        p2 = Min&(ed.txtLen, p1 + 1)
        x2 = ed.xCaret + 1 : y2 = ed.yCaret
     End If

  ElseIf selFlag = 2 Then                     'No selection, delete previous character (Back)
     p1 = Max&(0, edmTxtGetPos(ed.hWnd, ed.xCaret, ed.yCaret) - 1)
     x2 = ed.xCaret : y2 = ed.yCaret
     t = xp + p1
     If @t = 10 Then                          'if line feed character (LF, of CRLF)
        p1 = Max&(0, p1 - 1)                  'delete line feed, 2 characters
        p2 = Min&(ed.txtLen, p1 + 2)
        y1 = ed.yCaret - 1 : x1 = edmTxtGetLineLen(ed, y1)
     Else                                     'else regular, 1 character
        p2 = p1 + 1
        x1 = ed.xCaret - 1 : y1 = ed.yCaret
     End If

  Else
     p1 = edmTxtGetPos(ed.hWnd, ed.SelStartX, ed.SelStartY) 'selStart
     p2 = edmTxtGetPos(ed.hWnd, ed.SelEndX, ed.SelEndY)     'selEnd
     If p2 < p1 Then Swap p1, p2                    'make sure start is before end
  End If
  L = p2 - p1                                       'SelLen

  If uFlag And L Then
     s = Peek$(xp + p1, L)    'grab selection for undo buffer
  End If

  If ed.txtLen - p2 > 0 Then
     MoveMemory ByVal (xp + p1), _          'Destination
                ByVal (xp + p2), _          'Source
                ed.txtLen - p2              'Size of block to move
  End If

  ed.txtLen = ed.txtLen - L
  Poke$ xp + ed.txtLen, String$(L, 0)

  If ed.txtLen < HeapSize(ed.heapTxt, 0&, ByVal ed.txt) + 512 Then
     edmTxtResizeBlocks ed, ed.txtLen
  End If

  edmTxtUpDateLineEnds ed, ed.txt, ed.LineEnd    'update LineEnds
  edmTxtLineEndResize ed, 0                      'resize LineEnd memory block
  If ed.Changed = 0 Then ed.Changed = -1         'set changed flag

  If ed.SelStartY < ed.SelEndY Then
     ed.xCaret = ed.SelStartX
     ed.yCaret = ed.SelStartY
     If uFlag And SelFlag = 0 Then
        x1 = ed.SelStartX : y1 = ed.SelStartY
        x2 = ed.SelEndX   : y2 = ed.SelEndY
     End If
     ed.LineUpdate = 0
  ElseIf ed.SelStartY > ed.SelEndY Then
     ed.xCaret = ed.SelEndX
     ed.yCaret = ed.SelEndY
     If uFlag And SelFlag = 0 Then
        x1 = ed.SelEndX   : y1 = ed.SelEndY
        x2 = ed.SelStartX : y2 = ed.SelStartY
     End If
     ed.LineUpdate = 0
  ElseIf ed.SelStartY = ed.SelEndY Then
     ed.xCaret = Min&(ed.SelStartX, ed.SelEndX)
     ed.yCaret = ed.SelStartY
     If uFlag Then
        If y1 = y2 Then ed.LineUpdate = 1
        If SelFlag = 0 Then
           x1 = Min&(ed.SelStartX, ed.SelEndX) : y1 = ed.SelStartY
           x2 = Max&(ed.SelStartX, ed.SelEndX) : y2 = ed.SelEndY
        End If
     End If
  End If

  If ed.confineCaret Then
     ed.xCaretOld = Max&(0, ed.xCaret)
  End If

  If uFlag Then edmUndoAdd ed, 2, iType, s, x1, y1, x2, y2
End Sub

'
' Set new text to editor, like WM_SETTEXT..
'
Sub edmSetText(ByVal hWnd As Dword, ByVal txt As String)
  Local ed As EDM32DATA Ptr
  ed = GetWindowLong(hWnd, 0)

  If IsFalse(edmTxtInit(@ed)) Then
     SendMessage @ed.hParent, %WM_COMMAND, MakLng(@ed.id, %EN_ERRSPACE), hwnd
     'MessageBox hWnd, "Could not allocate memory for Editor control" & CHR$(0), _
     '                 "Error message" & CHR$(0), %MB_OK
     Exit Sub
  End If

  If Len(txt) Then
     If InStr(txt, Chr$(0)) Then 'binary - remove zeros
        If InStr(txt, Chr$(0, 0)) Then Replace Chr$(0, 0) With "  " In txt 'replace doubles with space
        If InStr(txt, Chr$(0)) Then txt = Remove$(txt, Chr$(0))            'remove the rest
     End If
     If InStr(txt, $Tab) Then Replace $Tab With Space$(@ed.tabSize) In txt ' replace tabs..

     Replace $CrLf With $Cr In txt   'convert to easier handled format
     Replace $Lf With $Cr In txt     'in case some odd CHR$(10) still is there..

     If @ed.WordWrap Then
        txt = edmWrapText(txt, @ed.wnWidth, 1)
     End If

     Replace $Cr With $CrLf In txt           'restore line feeds again

     edmTxtPutString @ed, txt, 0, 0, 0, 0, 0
  End If

  @ed.xCaret = 0 : @ed.yCaret = 0 : @ed.Changed = 0 : @ed.LineUpdate = 0
  edmUndoInit @ed
  SendMessage hWnd, %EDM_RESETSEL, 0, 0
  SendMessage hWnd, %EDM_REFRESH, 0, 0
End Sub

'
' Add or subtract one to LineEnd array
'
Sub edmTxtAddToLineEnds(ed As EDM32DATA, ByVal addThis As Long)
  Local p As Dword Ptr, q As Dword Ptr

  p = ed.LineEnd + ed.yCaret * 4    'start from current line
  q = ed.LineEnd + ed.UBtxBuf * 4   'last line

  !mov eax, addThis    ; add this value
  !mov edx, p          ; buffer address
  !mov ecx, q          ; offset for exit condition

  atl:
    !add [edx], eax    ; put fill chars at address in edx
    !add edx, 4        ; increment edx by 4 (DWORD)
    !cmp edx, ecx      ; compare to counter
    !jbe atl           ; jump out if below or equal (jbe)

End Sub

'
' Update entire LineEnd array. Time critical code - therefore some ASM
'
Sub edmTxtUpDateLineEnds(ed As EDM32DATA, ByVal xp As Long, ByVal lp As Long)
  #Register None
  Register ln As Long, lc As Long
  Local pn As Long, pl As Dword Ptr

  pn = ed.txtLen           ' put text's total length into ln
  pl = lp                  ' pointer to LineEnd's memory block

  ! mov ecx, pn            ; Length
  ! cmp ecx, 0             ; check length
  ! je ExitUdlLoop         ; jump out if zero
  ! mov eax, xp            ; pointer to ed.txt's memory block

  BgnUdlLoop:
     ! cmp ecx, ln         ; exit when ln = txtLen
     ! jne ContinueUdlLoop ; not equal ln - continue
        ! jmp ExitUdlLoop  ; else jump out of here

  ContinueUdlLoop:
     ! mov dl, [eax]       ; move character into dl
     ! cmp dl, 13          ; if dl = 13, line feed character
     ! je PutLEnd          ; jump to PutEnd if it equals 13
     ! inc eax             ; else get next character
     ! inc ln              ; increase ln (length) counter
     ! jmp BgnUdlLoop      ; jump to beginning again

  PutLEnd:
     ! push eax
     @pl = ln              ' put line's end position into array element
     ! pop eax
     ! add pl, 4           ; move to next array element
     ! inc eax             ; move to next character
     ! inc lc              ; increase line counter
     ! inc ln              ; increase ln (length) counter
     ! jmp BgnUdlLoop      ; jump to beginning again

  ExitUdlLoop:
     @pl = ed.txtLen       ' last one must equal text length
     ed.UBtxBuf = lc       ' total number of lines
End Sub

'
' Initialize - clear all
'
Function edmTxtInit(ed As EDM32DATA) As Dword

  If ed.txt         Then HeapFree ed.heapTxt, 0&, ByVal ed.txt
  If ed.heapTxt     Then ed.txt     = HeapAlloc(ed.heapTxt, %HEAP_ZERO_MEMORY, 4096)
  If ed.LineEnd     Then HeapFree ed.heapLineEnd, 0&, ByVal ed.LineEnd
  If ed.heapLineEnd Then ed.LineEnd = HeapAlloc(ed.heapLineEnd, %HEAP_ZERO_MEMORY, 128)

  edmBookMarkInitialize ed.hWnd
  ed.txtLen    = 0 : ed.UBtxBuf = 0
  ed.yCaret    = 0 : ed.xCaret  = 0 : ed.TopLine = 0
  ed.SelStartX = 0 : ed.SelEndX = 0
  ed.SelStartY = 0 : ed.SelEndY = 0 : ed.SelLen  = 0
  Function = ed.txt
End Function

'
' Resize memory
'
Sub edmTxtResizeBlocks(ed As EDM32DATA, ByVal nSize As Long)
  Local mOld As Dword, mNew As Dword

  mNew = HeapAlloc(ed.heapTxt, %HEAP_ZERO_MEMORY, nSize + 4096) 'Allocate new memory block
  mOld = ed.txt                                          'Pointer to old block

  If ed.txtLen < HeapSize(ed.heapTxt, 0&, ByVal ed.txt) Then
     If ed.txtLen Then
        MoveMemory ByVal mNew, ByVal mOld, ed.txtLen     'Move old text to new memory
     End If
  End If

  HeapFree ed.heapTxt, 0&, ByVal ed.txt                  'Free old block
  ed.txt = mNew                                          'Assign new block's handle to structure

End Sub

'
' Resize LineEnd memory block
'
Sub edmTxtLineEndResize(ed As EDM32DATA, ByVal ii As Long)
  Local mOld As Dword, mNew As Dword

  mNew = HeapAlloc(ed.heapLineEnd, %HEAP_ZERO_MEMORY, (ed.UBtxBuf + ii) * 4 + 128) 'Allocate new memblock
  mOld = ed.LineEnd                             'Pointer to old block

  MoveMemory ByVal mNew, _                      'New memblock, pos 0
             ByVal mOld, _                      'Old memblock, pos 0
             ed.UBtxBuf * 4 + 4                 'Move old to new

  HeapFree ed.heapLineEnd, 0&, ByVal ed.LineEnd 'Free old block
  ed.LineEnd = mNew                             'Pointer to new block

End Sub

'
' UNDO/REDO functions
' Initialize undo - clear all
'
Function edmUndoInit(ed As EDM32DATA) As Dword

  If ed.Undotxt     Then HeapFree ed.heapUndotxt, 0&, ByVal ed.Undotxt
  If ed.heapUndotxt Then ed.Undotxt = HeapAlloc(ed.heapUndotxt, %HEAP_ZERO_MEMORY, 1024)
  If ed.Undo        Then HeapFree ed.heapUndo, 0&, ByVal ed.Undo
  If ed.heapUndo    Then ed.Undo = HeapAlloc(ed.heapUndo, %HEAP_ZERO_MEMORY, 128)
  ed.undoPos = 0 : ed.undoCount = 0

  Function = ed.Undotxt
End Function

'
' UNDO/REDO ACTIONS
' Add undo action
'
Sub edmUndoAdd(ed As EDM32DATA, ByVal act As Long, ByVal iType As Long, ByVal txt As String, _
               ByVal x1 As Long, ByVal y1 As Long, ByVal x2 As Long, ByVal y2 As Long)

  Local mOldtxt As Dword, mNewtxt As Dword, mOld As Dword, mNew As Dword, oldEnd As Long

  If ed.undoPos < ed.undoCount Then ed.undoCount = ed.undoPos 'reset undoCount if undo been used

  If Len(txt) Then  'Undo txt
     If ed.undoCount Then oldEnd = ed.@undo[ed.undoCount - 1].uTxtBlockEnd     'last end = size of all text
     mNewtxt = HeapAlloc(ed.heapUndotxt, %HEAP_ZERO_MEMORY, oldEnd + Len(txt) + 1024)
     mOldtxt = ed.Undotxt                                    'Old block's pointer

     MoveMemory ByVal mNewtxt, ByVal mOldtxt, oldEnd         'Move old to new
     Poke$ mNewtxt + oldEnd, txt                             'add text to new
     ed.@undo[ed.undoCount].uTxtBlockEnd = oldEnd + Len(txt) ' end of this text block

     HeapFree ed.heapUndotxt, 0&, ByVal ed.Undotxt           'Free old text block
     ed.Undotxt = mNewtxt                                    'Assign new block's handle to structure
  End If

  'Undo data, resize block by creating new and moving old to new
  mNew = HeapAlloc(ed.heapUndo, %HEAP_ZERO_MEMORY, (ed.undoCount + 2) * Len(UNDODATA))
  mOld = ed.Undo                                 'Old block's pointer
  MoveMemory ByVal mNew, ByVal mOld, (ed.undoCount + 1) * Len(UNDODATA) 'Move old to new
  HeapFree ed.heapUndo, 0&, ByVal ed.Undo        'Free old block
  ed.Undo = mNew                                 'Assign new block's handle to structure

  ed.@undo[ed.undoCount].uAction       = act     ' type of undo/redo action
  ed.@undo[ed.undoCount].uActionCount  = 0       ' number of undo actions in a row
  ed.@undo[ed.undoCount].uActionStartX = x1      ' x startpos of undo/redo action
  ed.@undo[ed.undoCount].uActionStartY = y1      ' y startpos of undo/redo action
  ed.@undo[ed.undoCount].uActionEndX   = x2      ' x endpos of undo/redo action
  ed.@undo[ed.undoCount].uActionEndY   = y2      ' y endpos of undo/redo action
  ed.@undo[ed.undoCount].uActionLen    = Len(txt)
  ed.@undo[ed.undoCount].uDescription  = iType   ' for description text table

  Incr ed.undoCount
  ed.undoPos = ed.undoCount

End Sub

'
' UNDO
'
Sub edmUndoDoUndo(ed As EDM32DATA)
  If ed.undoPos > 0 Then
     Local txt As String, oldStart As Long, oldEnd As Long, ii As Long

     If ed.@undo[ed.undoPos - 1].uDescription = 23 Then 'Undo Replace all
        Local tmpX As Long, tmpY As Long
        tmpX = ed.xCaret : tmpY = ed.yCaret             'store current position
     End If

     For ii = 0 To ed.@undo[ed.undoPos - 1].uActionCount
        Decr ed.undoPos
        Select Case ed.@undo[ed.undoPos].uAction
           Case 1                                                ' delete inserted text
              ed.SelStartX = ed.@undo[ed.undoPos].uActionStartX  ' x startpos of undo/redo action
              ed.SelStartY = ed.@undo[ed.undoPos].uActionStartY  ' y startpos of undo/redo action
              ed.SelEndX   = ed.@undo[ed.undoPos].uActionEndX    ' x endpos of undo/redo action
              ed.SelEndY   = ed.@undo[ed.undoPos].uActionEndY    ' y endpos of undo/redo action
              ed.SelLen    = ed.@undo[ed.undoPos].uActionLen
              If Abs(ed.SelEndY - ed.SelStartY) > 0 Then
                 edmBookMarkDelete ed, Min&(ed.SelStartY, ed.SelEndY), Abs(ed.SelEndY - ed.SelStartY)
              End If
              edmTxtDeleteSel ed, 0, 0, 0
              SendMessage ed.hWnd, %EDM_RESETSEL, 0, 0

           Case 2                                             ' insert deleted text
              ed.xCaret = ed.@undo[ed.undoPos].uActionStartX  ' x startpos of undo/redo action
              ed.yCaret = ed.@undo[ed.undoPos].uActionStartY  ' y startpos of undo/redo action

              If ed.undoPos Then oldStart = ed.@undo[ed.undoPos - 1].uTxtBlockEnd 'else it's 0
              oldEnd = ed.@undo[ed.undoPos].uTxtBlockEnd
              txt = Peek$(ed.Undotxt + oldStart, oldEnd - oldStart)
              edmTxtPutString ed, txt, ed.xCaret, ed.yCaret, -1, 0, 0

              ed.xCaret = ed.@undo[ed.undoPos].uActionEndX
              ed.yCaret = ed.@undo[ed.undoPos].uActionEndY
              If ed.@undo[ed.undoPos].uDescription = 23 Then Exit Select      'no selection for Replace all undo
              If ed.@undo[ed.undoPos + 1].uDescription = 23 Then Exit Select  'same here
              ed.SelStartX = ed.@undo[ed.undoPos].uActionStartX  ' x startpos of undo/redo action
              ed.SelStartY = ed.@undo[ed.undoPos].uActionStartY  ' y startpos of undo/redo action
              ed.SelEndX   = ed.@undo[ed.undoPos].uActionEndX    ' x endpos of undo/redo action
              ed.SelEndY   = ed.@undo[ed.undoPos].uActionEndY    ' y endpos of undo/redo action
              ed.SelLen    = ed.@undo[ed.undoPos].uActionLen
        End Select
     Next ii

     If ii Then
        Decr ii
        ed.@undo[ed.undoPos].uActionCount = ed.@undo[ed.undoPos + ii].uActionCount
        ed.@undo[ed.undoPos].uDescription = ed.@undo[ed.undoPos + ii].uDescription
        If ed.@undo[ed.undoPos].uDescription = 23 Then ' Undo Replace all - keep position
           ed.xCaret = Min&(tmpX, edmTxtGetLineLen(ed, tmpY))
           ed.yCaret = tmpY
        End If
     End If
     SendMessage ed.hWnd, %EDM_CHECKBOUNDS, 0, 0

  Else
     MessageBeep &HFFFFFFFF
     SetFocus ed.hWnd
  End If
End Sub

'
' REDO
'
Sub edmUndoDoRedo(ed As EDM32DATA)
  If ed.undoPos < ed.undoCount Then
     Local txt As String, oldStart As Long, oldEnd As Long, ii As Long

     If ed.@undo[ed.undoPos].uDescription = 23 Then ' Redo Replace all
        Local tmpX As Long, tmpY As Long            ' store current pos
        tmpX = ed.xCaret : tmpY = ed.yCaret
     End If

     For ii = 0 To ed.@undo[ed.undoPos].uActionCount
        Select Case ed.@undo[ed.undoPos].uAction
           Case 1                                                 ' insert deleted text
              ed.xCaret = ed.@undo[ed.undoPos].uActionStartX  ' x startpos of undo/redo action
              ed.yCaret = ed.@undo[ed.undoPos].uActionStartY  ' y startpos of undo/redo action

              If ed.undoPos > 0 Then oldStart = ed.@undo[ed.undoPos - 1].uTxtBlockEnd 'else it's 0
              oldEnd = ed.@undo[ed.undoPos].uTxtBlockEnd
              txt = Peek$(ed.Undotxt + oldStart, oldEnd - oldStart)
              edmTxtPutString ed, txt, ed.xCaret, ed.yCaret, -1, 0, 0

              ed.xCaret = ed.@undo[ed.undoPos].uActionEndX
              ed.yCaret = ed.@undo[ed.undoPos].uActionEndY
              SendMessage ed.hWnd, %EDM_RESETSEL, 0, 0

           Case 2
              ed.SelStartX = ed.@undo[ed.undoPos].uActionStartX  ' x startpos of undo/redo action
              ed.SelStartY = ed.@undo[ed.undoPos].uActionStartY  ' y startpos of undo/redo action
              ed.SelEndX   = ed.@undo[ed.undoPos].uActionEndX    ' x endpos of undo/redo action
              ed.SelEndY   = ed.@undo[ed.undoPos].uActionEndY    ' y endpos of undo/redo action
              ed.SelLen    = ed.@undo[ed.undoPos].uActionLen
              If Abs(ed.SelEndY - ed.SelStartY) > 0 Then
                 edmBookMarkDelete ed, Min&(ed.SelStartY, ed.SelEndY), Abs(ed.SelEndY - ed.SelStartY)
              End If
              edmTxtDeleteSel ed, 0, 0, 0
              SendMessage ed.hWnd, %EDM_RESETSEL, 0, 0

        End Select
        Incr ed.undoPos
     Next ii

     If ed.@undo[ed.undoPos - 1].uDescription = 23 Then ' Redo Replace all - looks better to keep pos
        ed.xCaret = Min&(tmpX, edmTxtGetLineLen(ed, tmpY))
        ed.yCaret = tmpY
     End If
     SendMessage ed.hWnd, %EDM_CHECKBOUNDS, 0, 0

  Else
     MessageBeep &HFFFFFFFF
     SetFocus ed.hWnd
  End If

End Sub

'
' Get undo/redo action's description text
'
Function edmUndoGetDescription(ByVal hWnd As Dword, ByVal UndoOrRedo As Long) As String
  Local ed As EDM32DATA Ptr, txt As String, iType As Long
  ed = GetWindowLong(hWnd, 0)

  If UndoOrRedo Then 'undo text
     If @ed.undoPos > 0 Then
        iType = @ed.@undo[@ed.undoPos - 1].uDescription
        txt = "Undo"
     Else
        iType = 30
     End If
  Else               'redo text
     If @ed.undoPos < @ed.undoCount Then
        iType = @ed.@undo[@ed.undoPos].uDescription
        txt = "Redo"
     Else
        iType = 31
     End If
  End If
  Function = txt + "  " + edmUndoGetString(iType)
End Function

'
' Fill a list box with Undo/Redo descriptions
'
Sub edmUndoRedoFillList(ByVal lsthWnd As Dword, ByVal edhWnd As Dword, ByVal UndoOrRedo As Long)
  Local ed As EDM32DATA Ptr, txt As String, ii As Long
  ed = GetWindowLong(edhWnd, 0)
  SendMessage lsthWnd, %LB_RESETCONTENT, 0, 0

  If UndoOrRedo Then 'undo text
     If @ed.undoPos > 0 Then
        For ii = @ed.undoPos - 1 To 0 Step - 1
           txt = " Undo " & edmUndoGetString(@ed.@undo[ii].uDescription)
           SendMessage lsthWnd, %LB_ADDSTRING, 0, ByVal StrPtr(txt)
           If @ed.@undo[ii].uActionCount > 0 Then ii = ii - @ed.@undo[ii].uActionCount
        Next
     End If
  Else               'redo text
     If @ed.undoPos < @ed.undoCount Then
        For ii = @ed.undoPos To @ed.undoCount - 1
           If @ed.@undo[ii].uActionCount > 0 Then ii = ii + @ed.@undo[ii].uActionCount
           txt = " Redo " & edmUndoGetString(@ed.@undo[ii].uDescription)
           SendMessage lsthWnd, %LB_ADDSTRING, 0, ByVal StrPtr(txt)
        Next
     End If
  End If
End Sub

'
' Get proper undo/redo description text
'
Function edmUndoGetString(ByVal iType As Long) As String
   Select Case iType
      Case 1  : Function = "Input text"
      Case 2  : Function = "New line"
      Case 3  : Function = "Tab"
      Case 4  : Function = "Input and replace"  'when input replaces selection
      Case 5  : Function = "Paste"
      Case 6  : Function = "Upper case"
      Case 7  : Function = "Capitalize words"
      Case 8  : Function = "Lower case"
      Case 9  : Function = "Toggle case"
      Case 10 : Function = "Cut"
      Case 11 : Function = "Backspace"
      Case 12 : Function = "Delete"
      Case 13 : Function = "Paste"
      Case 14 : Function = "Indent"
      Case 15 : Function = "Outdent"
      Case 16 : Function = "Rem out"
      Case 17 : Function = "Remove Rem out"
      Case 18 : Function = "Sort ascending"
      Case 19 : Function = "Sort ascending"
      Case 20 : Function = "Sort descending"
      Case 21 : Function = "Sort descending"
      Case 22 : Function = "Replace"
      Case 23 : Function = "Replace all"
      Case 30 : Function = "Nothing to Undo"
      Case 31 : Function = "Nothing to Redo"
  End Select
End Function
